(* Tags interface.  Defines the Tag.T enumeration type which is used
	to keep saved programs consistent between compiler and
	run-time interpreter modules of the system.

	The tags interface also includes some version numbers which
	the Code interface uses to allow modules which load in Code
	files to check that they will be able to read the file correctly.
*)

(*
	05/12/96	Created
	09/12/96	Added a load of new tags required by the parser.
	11/12/96	Added tags for the "new" function and the "nil"
			constant value.
	14/12/96	Added a tag for the "colour" command.
	23/12/96	Added a tag for the "str" function.
	24/12/96	Added tags for the limiting commands, 
			"maxthread" and "maxfunction".
	11/01/97	Added a tag for the Mutex data-type.
	12/01/97	Added a tag for the "clearall" command.
	14/01/97	Added a tag for the "pause" command.
	18/01/97	Added tags for "div" and "mod".
	24/01/97	Added tag for in-built "time" value.
	04/02/97	Added tags for the port-related stuff.
	19/02/97	Added a tag for the special "quit" command.
			Added a tag for the special "execute" command.
	21/02/97	Added tags for the new sub-terms of the function
			declaration command.
	23/02/97	Format of "fun", "var" and "exception" commands
			changed slightly, so version tags were changed for safety.
	26/02/97	Tags added to handle new "type" command format!
	20/03/97	Added a tag for the Try...Finally command.
*)

INTERFACE Tags;

(* The Armadillo file code version is printed in the form
	<version>.<revision>
	Version values are checked first, then, if they were consistent
	then revision values are checked, too.
*)
CONST	Version = 0;
	Revision = 23;

	(* These tags specify the filename extensions for the source
		file and compiled code file names. *)
	SourceExt = ".arm";
	CodeExt = ".acf";

TYPE	T = {

	(* Parallel constructor tag. *)
	CMD_Par,

	(* Tags for exception commands. *)
	BLK_ExceptDef,		(* For defining exceptions. *)
	BLK_TryExcept,		(* For trapping exceptions. *)
	BLK_TryFinally,		(* For executing cleanup code. *)
	CMD_Raise,		(* Raising exceptions. *)

	(* Tags for program control. *)
	CMD_If,			(* All forms of If. *)
	CMD_ElsIf,		(* ElsIf tag. *)
	BLK_Loop,		(* Loop construct. *)
	CMD_Exit,		(* Exit loop command. *)
	CMD_Return,		(* All forms of Return. *)
	CMD_Quit,		(* Quit the run-time. *)
	
	(* Tags for variable operations. *)
	BLK_Var,		(* Variable declarator. *)
	CMD_LetAll,		(* Setting some vars to one expression. *)
	CMD_Let,		(* Setting some vars to equal some expressions. *)
	CMD_New,		(* Initialising storage for a ref variable. *)
	BLK_Readonly,		(* Make some vars readonly. *)
	BLK_Lock,		(* Locking some vars *)

	(* Concurrency controls. *)
	CMD_Signal,		(* Signalling a condition variable. *)
	CMD_Wait,		(* Waiting for a condition variable. *)

	(* Port commands. *)
	CMD_Send,		(* Send some data down a port. *)
	FUN_Recieve,		(* Recieve some data from a port. *)

	(* Function declaration/calling. *)
	CMD_Fun,		(* Declaring a function. *)
	CMD_Call,		(* Calling a function. *)
	
	(* Type declaration. *)
	CMD_Type,		(* Declaring a user type. *)

	(* Run-time system limiting routines. *)
	CMD_MaxThread,		(* Setting the threading limit. *)
	CMD_MaxFunction,	(* Setting the recursion limit. *)

	(* Command to execute a separate, precompiled program. *)
	CMD_Execute,		(* External program execution. *)

	(* Command to force garbage collection. *)
	CMD_Collect,		(* Garbage collector trigger. *)

	(* Tags for graphics commands. *)
	GFX_Forward,		(* Move turtle forward. *)
	GFX_Backward,		(* Move backward. *)
	GFX_TLeft,		(* Turn left. *)
	GFX_TRight,		(* Turn right. *)
	GFX_TUp,		(* Turn up. *)
	GFX_TDown,		(* Turn down. *)
	GFX_RLeft,		(* Rotate anticlockwise. *)
	GFX_RRight,		(* Rotate clockwise. *)
	GFX_Penup,		(* Stop drawing. *)
	GFX_Pendown,		(* Start drawing. *)
	GFX_Turtle,		(* Select a turtle. *)
	GFX_Home,		(* Send the turtle to home. *)
	GFX_Clear,		(* Clear this turtle's drawings and those of
					any parents or children. *)
	GFX_ClearAll,		(* Clear any graphics that this interpreter has
					produced on terminals. *)
	GFX_Attach,		(* Attach a turtle to a particular display. *)
	GFX_Colour,		(* Set the turtle pen colour. *)

	(* Debugging commands. *)
	BUG_Print,		(* Print some text. *)
	BUG_Assert,		(* Make an assertion. *)
	BUG_Pause,		(* Halt execution for a defined period. *)

	(* Tags for types. *)
	TYP_Number,		(* Numeric data type. *)
	TYP_Condition,		(* Condition variable type. *)
	TYP_String,		(* Text data type. *)
	TYP_Boolean,		(* Boolean data type. *)
	TYP_Port,		(* Data transmission port type. *)
	TYP_Record,		(* Record data type. *)
	TYP_Turtle,		(* Graphics turtle type. *)
	TYP_Ref,		(* Nasty 'orrible reference type. *)
	TYP_Function,		(* Function closure type. *)
	TYP_Mutex,		(* Mutex data-type, as in Modula-3. *)
	TYPE_Port,		(* Inter-thread comms port. *)

	(* Tags for expressions. *)
	MONO_Plus,		(* Monadic plus. *)
	MONO_Minus,		(* Monadic minus. *)

	EXP_Plus,		(* Dyadic plus. *)
	EXP_Minus,		(* Dyadic minus. *)
	EXP_Times,		(* Multiplication. *)
	EXP_Divide,		(* Division. *)

	EXP_IntDiv,		(* Integer division with remainder. *)
	EXP_IntMod,		(* Remainder of integer division. *)

	EXP_Bracket,		(* Bracketting. *)

	(* Tags for functions returning bools. *)
	EXP_LessOrEq,		(* Less than or equal to.
					Also boolean "implied by" *)
	EXP_MoreOrEq,		(* More than or equal to. *)
	EXP_MoreThan,		(* More than. *)
	EXP_LessThan,		(* Less than. *)

	(* Tests based on equality. *)
	EXP_Equal,		(* Equality. *)
	EXP_NotEqual,		(* Inequality. *)

	(* In-built functions. *)
	FUN_Rnd,		(* Random number function. *)
	FUN_Str,		(* Function for converting ANY data type to text. *)

	(* Operators for referencing and dereferencing. *)
	REF_Deref,		(* Dereference a pointer. *)
	REF_Pointer,		(* Get a pointer to something. *)

	(* Boolean operations. *)
	BOOL_Or,		(* Boolean OR. *)
	BOOL_And,		(* Boolean AND. *)
	BOOL_Implies,		(* Boolean implication. *)
	BOOL_Not,		(* Boolean negation. *)	

	(* Tags for representing constants and data values. *)
	TAG_Exception,		(* Exception name. *)
	TAG_ExceptTrap,		(* Specification of an exception trap. *)
	TAG_ExceptTrap_List,	(* Specification of a list of traps. *)
	TAG_BracketCmd,		(* For bracketing of commands. *)

	TAG_Type,		(* Name of a custom data type. *)

	TAG_VarDeclList,	(* Variable type declaration list. *)
	TAG_VarDecl,		(* Variable type declaration. *)
	TAG_Variable,		(* Name of a variable. *)
	TAG_VarList,		(* List of variable names. *)

	TAG_Display,		(* Name and address of a display. *)
	TAG_DisplayName,	(* Name of a 3D-display terminal. *)
	TAG_IPAddress,		(* IP address of the display. *)

	TAG_Boolean,		(* A boolean constant. *)
	TAG_Real,		(* A real number value. *)
	TAG_Int,		(* An integer value. *)
	TAG_String,		(* A string constant. *)

	TAG_ExprList,		(* List of expressions. *)

	TAG_RecordRef,		(* Reference to a field in a record. *)
	TAG_Nil,		(* Empty reference. *)

	TAG_Time,		(* In-built variable containing the current time. *)

	TAG_FunDecl,		(* Function declaration node. *)
	TAG_FunDeclList,	(* Function declaration list node. *)

	TAG_TypeDecl,		(* Name to type descriptor binding. *)

	(* Spurious (BUT USEFUL!) No-op tag. *)
	NO_OP
	};

END Tags.