(* The Turtle module provides basic abstract turtle functionality. *)

MODULE Turtle;

IMPORT Matrix4, Point3, Math;

REVEAL T = Public BRANDED OBJECT
			(* Current position. *)
			Position, Forward, Upward : Point3.T;
			(* Home position. *)
			Home_Pos, Home_FD, Home_Up : Point3.T;
		OVERRIDES
			init := Init;
			forward := Forward;
			backward := Backward;
			tleft := TLeft;
			tright := TRight;
			tup := TUp;
			tdown := TDown;
			rleft := RLeft;
			rright := RRight;
			home := Home;

			position := Position;
			facing_fd := Facing_Fd;
			facing_up := Facing_Up;

			(* duplicate_pos is used to copy a turtle's properties. *)
			duplicate_pos := Duplicate;

			(* The info function returns a text string describing the turtle
				for debugging purposes. *)
			info := Info;
		END;

TYPE Rotator = OBJECT
			transform, inverse : Matrix4.T;
		METHODS
			RotateInX(v : Point3.T; angle : REAL) : Point3.T := RotateInA;
			RotateInY(v : Point3.T; angle : REAL) : Point3.T := RotateInB;
			RotateInZ(v : Point3.T; angle : REAL) : Point3.T := RotateInC;
		END;

PROCEDURE Init(self : T; pos, forward, upward : Point3.T) : T=
BEGIN
	self.Position := pos;
	self.Forward := Point3.ScaleToLen(forward, 1.0);
	self.Upward := Point3.ScaleToLen(upward, 1.0);

	(* Set the home position. *)
	self.Home_Pos := self.Position;
	self.Home_FD := self.Forward;
	self.Home_Up := self.Upward;

	RETURN self;
END Init;

PROCEDURE Forward(self : T; dist : Units)=
BEGIN

	self.Position := Point3.Plus(self.Position,
					Point3.ScaleToLen(self.Forward, dist));

END Forward; 

PROCEDURE Backward(self : T; dist : Units)=
BEGIN

	self.Position := Point3.Minus(self.Position,
					Point3.ScaleToLen(self.Forward, dist));

END Backward; 

PROCEDURE TLeft(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN

	self.Forward := rotator.RotateInY(self.Forward, -DegToRad(angle));
	Fix(self);

END TLeft;

PROCEDURE TRight(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN

	self.Forward := rotator.RotateInY(self.Forward, DegToRad(angle));
	Fix(self);

END TRight;

PROCEDURE TUp(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN
	self.Forward := rotator.RotateInZ(self.Forward, DegToRad(angle));
	self.Upward := rotator.RotateInZ(self.Upward, DegToRad(angle));
	Fix(self);
END TUp;

PROCEDURE TDown(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN
	self.Forward := rotator.RotateInZ(self.Forward, -DegToRad(angle));
	self.Upward := rotator.RotateInZ(self.Upward, -DegToRad(angle));
	Fix(self);
END TDown;

PROCEDURE RLeft(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN

	self.Upward := rotator.RotateInX(self.Upward, -DegToRad(angle));
	Fix(self);

END RLeft;

PROCEDURE RRight(self : T; angle : REAL)=
VAR rotator := NewRotator(self.Forward, self.Upward);
BEGIN

	self.Upward := rotator.RotateInX(self.Upward, DegToRad(angle));
	Fix(self);

END RRight;

PROCEDURE Home(self : T)=
BEGIN
	self.Position := self.Home_Pos;
	self.Forward := self.Home_FD;
	self.Upward := self.Home_Up;
END Home;

PROCEDURE Position(self : T) : Point3.T=
BEGIN
	RETURN self.Position;
END Position;

PROCEDURE Facing_Fd(self : T) : Point3.T=
BEGIN
	RETURN self.Forward;
END Facing_Fd;

PROCEDURE Facing_Up(self : T) : Point3.T=
BEGIN
	RETURN self.Upward;
END Facing_Up;

PROCEDURE Duplicate(self : T; new : T) : T=
BEGIN

	(* Copy the relevant fields into a new turtle. *)

	(* Current turtle position. *)
	new.Position := self.Position;
	new.Forward := self.Forward;
	new.Upward := self.Upward;

	(* Home turtle position. *)
	new.Home_Pos := new.Position;
	new.Home_FD := new.Forward;
	new.Home_Up := new.Upward;

	RETURN new;

END Duplicate;

PROCEDURE Info(self : T) : TEXT=
VAR text : TEXT;
BEGIN

	text := "pos=" & Point3.ToText(self.Position);
	text := text & ", forward=" & Point3.ToText(self.Forward);
	text := text & ", upward=" & Point3.ToText(self.Upward);

	RETURN text;

END Info;

(* Internal rotation system. *)
PROCEDURE NewRotator(VAR a, b : Point3.T) : Rotator=
VAR c : Point3.T;
	rotator := NEW(Rotator);
BEGIN

	(* First get the third vector. *)
	c := Point3.CrossProduct(a, b);
	b := Point3.CrossProduct(c, a);

(* DEBUG OUTPUT:
IO.Put(Point3.ToText(a) & ", " & Point3.ToText(b) & ", " & Point3ToText(c) & "\n");
*)

	(* Now construct the transform matrix to map points to the
		coordinate frame that the three vectors define. *)
	rotator.transform := Matrix4.TransformUnitCube(Point3.T{0.0, 0.0, 0.0}, a, b, c);

	(* Construct the inverse matrix for mapping points back to
		the real coordinate frame.  Because of the type of transform,
		the transpose is also the inverse, so I use that instead. *)
	rotator.inverse := Matrix4.Transpose(rotator.transform);

	RETURN rotator;

END NewRotator;

(* The following routines take a vector, map it into the new coordinate
	frame and then perform the desired rotation. *)
PROCEDURE RotateInA(self : Rotator; v : Point3.T; angle : REAL) : Point3.T=
BEGIN

	(* Transform the vector into the coordinate frame
		and apply the rotation. *)
	v:=Matrix4.TransformPoint3(
		Matrix4.Multiply(self.transform, Matrix4.RotateX(self.inverse, angle)), v);

	RETURN v;

END RotateInA;

PROCEDURE RotateInB(self : Rotator; v : Point3.T; angle : REAL) : Point3.T=
BEGIN

	(* Transform the vector into the coordinate frame
		and apply the rotation. *)
	v:=Matrix4.TransformPoint3(
		Matrix4.Multiply(self.transform, Matrix4.RotateY(self.inverse, angle)), v);

	RETURN v;

END RotateInB;

PROCEDURE RotateInC(self : Rotator; v : Point3.T; angle : REAL) : Point3.T=
BEGIN

	(* Transform the vector into the coordinate frame
		and apply the rotation. *)
	v:=Matrix4.TransformPoint3(
		Matrix4.Multiply(self.transform, Matrix4.RotateZ(self.inverse, angle)), v);
	RETURN v;

END RotateInC;

(* Some other useful internal procedures. *)
PROCEDURE DegToRad(angle : REAL) : REAL=
BEGIN
	RETURN angle * Math.Degree;
END DegToRad;

(* Fix normalizes the direction vectors. *)
PROCEDURE Fix(self : T)=
BEGIN
	self.Forward := Point3.ScaleToLen(self.Forward, 1.0);
	self.Upward := Point3.ScaleToLen(self.Upward, 1.0);
END Fix;

BEGIN
END Turtle.
