(* Lexer module.
	Implements the operations available on the Lexer.T
	object.
*)

(*
	05/12/96	Created
	09/12/96	Removed MustBe function.
			Pumped up the count on the nextbytes function to print 60 chars.
	12/12/96	Added implementation of the new eof() method.
	15/12/96	Modified the implementation of the "string" method to
			cope with the supported in-line control codes.
			(See language definition for details.)
	31/01/97	Lexer now allows numeric characters in identifiers,
			except for the first character.
	02/02/97	New routine provided to scan for an IP address.
	16/01/97	Fixed the IP address and Identifier scanning routines,
			which failed to restore the old file position if they
			didn't succeed.
	19/02/97	Added the FromText method to the lexer object to allow
			lexers to parse text strings as well as files.
	22/02/97	Fixed the comment-parsing system so that comments are treated
			as if they don't exist.
*)

MODULE Lexer;

(* FOR DEBUGGING: *)
(* IMPORT IO; *)

(* System modules. *)
IMPORT OSError, Rd, FileRd, TextRd, Scan, Lex, Thread, Text, FloatMode;

CONST	Identifier = SET OF CHAR{'a'..'z', 'A'..'Z', '_'};
	Identifier2 = Identifier + SET OF CHAR{'0' .. '9'};

REVEAL	T = Public BRANDED OBJECT
			reader : Rd.T;
			lastpos : CARDINAL;
		OVERRIDES
			init := Init;
			fromtext := FromText;

			expect := Expect;

			ident := Ident;
			real := Real;
			integer := Int;
			string := String;
			ip_address := IPAddress;

			nextbytes := NextBytes;

			eof := Eof;
			close := Close;
		END;

(* Init opens a file for reading. *)
PROCEDURE Init(self : T; filename : TEXT) : T RAISES {OSError.E}=
BEGIN

	(* Open the file up. *)
	self.reader := FileRd.Open(filename);
	RETURN self;

END Init;

(* FromText makes a lexer for a text string. *)
PROCEDURE FromText(self : T; t: TEXT) : T=
BEGIN

	(* Set up the reader. *)
	self.reader := TextRd.New(t);
	RETURN self;

END FromText;

(* Expect tests for the next token being the one specified. *)
PROCEDURE Expect(self : T; token : TEXT) : BOOLEAN =
BEGIN

	KillBlanks(self);

(*
IO.Put("-Expect '" & token & "'\n");
*)

	RETURN CheckFor(self, token);

END Expect;

PROCEDURE CheckFor(self : T; token : TEXT) : BOOLEAN =
BEGIN

	SavePos(self);

	TRY
		FOR x:=0 TO Text.Length(token)-1 DO

			IF Text.GetChar(token, x) # Char(self.reader) THEN

				(* Restore the old reader position. *)
				RestorePos(self);

				(* Return failure. *)
				RETURN FALSE;

			END;

		END;

	EXCEPT

		(* If a GetChar fails, then fail the Expect. *)
		Rd.EndOfFile, Rd.Failure, Thread.Alerted=>
		RestorePos(self);
		RETURN FALSE;

	END;

	RETURN TRUE;

END CheckFor;

(* The Ident procedure returns a valid identifier name
	if the next sequence of bytes evaluates to one,
	or fails with the BadToken exception. *)
PROCEDURE Ident(self : T) : TEXT RAISES {BadToken}=
VAR 	ident : TEXT;
	char1 : CHAR;
BEGIN
	KillBlanks(self);

	SavePos(self);

	TRY	
		(* Check that the first character is not a digit. *)
		char1 := Char(self.reader);
		IF char1 IN Identifier THEN
			ident := Text.FromChar(char1) & Lex.Scan(self.reader, Identifier2);
			RETURN ident;
		ELSE
			RestorePos(self);
			RAISE BadToken;
		END;
	EXCEPT
		Rd.EndOfFile, Rd.Failure, Thread.Alerted=>RAISE BadToken;
	END;

END Ident;

(* Real returns a longreal made up from the next few characters. *)
PROCEDURE Real(self : T) : LONGREAL RAISES {BadToken}=
VAR	token := "";
	c : CHAR;
	pos : CARDINAL;
BEGIN

	SavePos(self);

	TRY

		LOOP
			c := Char(self.reader);

			IF (c >= '0') AND (c <= '9') THEN
				token := token & Text.FromChar(c);
			ELSIF c = '.' THEN
				token := token & ".";
				EXIT;
			ELSE
				RestorePos(self);
				RAISE BadToken;
			END;

		END;

		LOOP
			pos := Rd.Index(self.reader);
			c := Char(self.reader);

			IF (c >= '0') AND (c <= '9') THEN
				token := token & Text.FromChar(c);
			ELSE
				Rd.Seek(self.reader, pos);
				EXIT;
			END;

		END;

		RETURN Scan.LongReal(token);

	EXCEPT
		Rd.Failure, Rd.EndOfFile, Thread.Alerted,
		Lex.Error, FloatMode.Trap=>
			RestorePos(self);
			RAISE BadToken;
	END;

END Real;

PROCEDURE Int(self : T) : INTEGER RAISES {BadToken}=
VAR	token := "";
	c : CHAR;
	pos : CARDINAL;
BEGIN

	SavePos(self);

	TRY

		LOOP
			pos := Rd.Index(self.reader);
			c := Char(self.reader);

			IF (c >= '0') AND (c <= '9') THEN
				token := token & Text.FromChar(c);
			ELSE
				Rd.Seek(self.reader, pos);
				EXIT;
			END;

		END;

		RETURN Scan.Int(token);

	EXCEPT
		Rd.Failure, Rd.EndOfFile, Thread.Alerted,
		Lex.Error, FloatMode.Trap=>
			RestorePos(self);
			RAISE BadToken;
	END;

END Int;

PROCEDURE String(self : T) : TEXT RAISES {BadToken}=
VAR	token := "";
	c : CHAR;
BEGIN

	IF NOT(self.expect("\"")) THEN
		RAISE BadToken;
	END;

	SavePos(self);

	TRY

		LOOP
			(* Use Rd.GetChar here - strings cannot contain comments! *)
			c := Rd.GetChar(self.reader);
			IF c = '"' THEN
				EXIT;
			ELSIF c = '\\' THEN

				(* Deal with in-line control codes. *)
				c := Char(self.reader);
				IF c = 'n' THEN
					token := token & "\n";
				ELSIF c = 't' THEN
					token := token & "\t";
				ELSIF c = '\\' THEN
					token := token & "\\";
				ELSIF c = '"' THEN
					token := token & "\"";
				ELSE
					token := token & "\\" & Text.FromChar(c);
				END;

			ELSE
				token := token & Text.FromChar(c);
			END;

		END;
		RETURN token;

	EXCEPT
		Rd.Failure, Rd.EndOfFile, Thread.Alerted=>
			RestorePos(self);
			RAISE BadToken;
	END;

END String;

(* The IPAddress routine returns an IP address from the text,
	or raises BadToken if an input error occurs. *)
PROCEDURE IPAddress(self : T) : TEXT RAISES {BadToken}=
VAR 	ident : TEXT;
	char1 : CHAR;
BEGIN
	KillBlanks(self);

	SavePos(self);

	TRY	
		LOOP

			(* Get the next character. *)
			char1 := Char(self.reader);

			IF char1 IN Lex.Blanks THEN
				(* End of IP address. *)
				IF ident = NIL THEN
					RestorePos(self);
					RAISE BadToken;
				END;
				RETURN ident;
			ELSE
				IF ident = NIL THEN
					ident := Text.FromChar(char1);
				ELSE
					ident := ident & Text.FromChar(char1);
				END;
			END;

		END;
	EXCEPT
		Rd.EndOfFile, Rd.Failure, Thread.Alerted=>
			RestorePos(self);
			RAISE BadToken;
	END;
END IPAddress;

(* NextBytes returns the next 20 bytes of input as a string. *)
PROCEDURE NextBytes(self : T) : TEXT=
VAR	text : TEXT;
	c : CHAR;
	count := 0;
BEGIN

	text := "";

	TRY

		WHILE (count < 60) DO

			c := Char(self.reader);
			IF c = '\n' THEN
				EXIT;
			END;
			text := text & Text.FromChar(c);
			INC(count);

		END;

	EXCEPT
		Rd.Failure, Rd.EndOfFile, Thread.Alerted=>RETURN text;
	END;

	RETURN text;

END NextBytes;

(* Eof returns a boolean indicating whether or not we are at the end of the file. *)
PROCEDURE Eof(self : T) : BOOLEAN=
BEGIN
	TRY
		RETURN Rd.EOF(self.reader);
	EXCEPT
		Rd.Failure, Thread.Alerted=>RETURN TRUE;
	END;
END Eof;

(* Close closes the reader in the specified lexer object. *)
PROCEDURE Close(self : T)=
BEGIN
	TRY
		Rd.Close(self.reader);
	EXCEPT
		Rd.Failure, Thread.Alerted=>
	END;
END Close;

(* INTERNAL procedures. *)

(* SavePos saves the current position in the file. *)
PROCEDURE SavePos(self : T) =
BEGIN
	self.lastpos := Rd.Index(self.reader);
END SavePos; 

(* Restore pos moves the reader back the last position saved. *)
PROCEDURE RestorePos(self : T) =
BEGIN

	TRY
		Rd.Seek(self.reader, self.lastpos);
	EXCEPT
		(* In case of reader error, just return. *)
		Rd.Failure, Thread.Alerted=>
	END;

END RestorePos;

(* KillBlanks eats blank characters from a lexer.
	It also removes remarks. *)
PROCEDURE KillBlanks(self : T) =
VAR	pos : CARDINAL;
BEGIN

	TRY
		LOOP
			pos := Rd.Index(self.reader);
			IF NOT(Char(self.reader) IN Lex.Blanks) THEN
				Rd.Seek(self.reader, pos);
				EXIT;
			END;
		END;
	EXCEPT
		(* If an exception is raised then just return. *)
		Rd.EndOfFile, Rd.Failure, Thread.Alerted=>
	END;

END KillBlanks;

(* Internal function to read a character from a reader while
	ignoring any comments placed in the text. *)
PROCEDURE Char(rd : Rd.T) : CHAR RAISES {Rd.EndOfFile, Rd.Failure, Thread.Alerted}=
VAR	c, c2 : CHAR;
BEGIN

(* IO.Put("-"); *)
	LOOP

		(* Read the next character from the reader. *)
		c := Rd.GetChar(rd);
(* IO.Put(Text.FromChar(c)); *)

		(* If the character is an opening bracket then check for comments. *)
		IF c = '(' THEN

			(* Save the position and read in another character. *)
			c2 := Rd.GetChar(rd);

			(* Is this a comment? *)
			IF c2 = '*' THEN

				(* Eat characters until the EOF (error), or we
					encounter the corresponding end bracket. *)
				VAR	nesting := 0;
				BEGIN
					WHILE nesting >= 0 DO
						(* Get the next character. *)
						c2 := Rd.GetChar(rd);
						IF c2 = '(' THEN
							c2 := Rd.GetChar(rd);
							IF c2 = '*' THEN
								(* Opening a new comment! *)
								INC(nesting);
							ELSE
								Rd.UnGetChar(rd);
							END;
						ELSIF c2 = '*' THEN
							c2 := Rd.GetChar(rd);
							IF c2 = ')' THEN
								(* Closing this comment! *)
								DEC(nesting);
							ELSE
								Rd.UnGetChar(rd);
							END;
						END;

					END;
				END;
			ELSE

				(* Push the character back on and return the bracket! *)
				Rd.UnGetChar(rd);
				RETURN c;

			END;

		ELSE

			(* Return the character. *)
			RETURN c;

		END;

	END;

END Char;

BEGIN

END Lexer.