(* Main module for the Armadillo compiler.
	This module provides a simple interface to the functionality
	of the parser/compiler system, which compiles a program from
	a text file into a special record-code which the interpreter
	or optimiser can then load in and work with.
	The compiler may not run as a local machine server, connected
	to via a well-known name, to allow interpreters to request
	on-the-fly compilation of small code fragments.
*)

(*
	19/02/97	Modified the Main module to allow server-mode operation.
*)

MODULE Main;

(* System libraries. *)
IMPORT IO, Fmt, Text, Params;

(* The armacode library. *)
IMPORT Tags;

(* The parser/compiler. *)
IMPORT Parse;

(* Procedure to cope with special command line tokens. *)
PROCEDURE Special(token : TEXT)=
BEGIN

	IF Text.Equal(token, "/SERVER") THEN
		server_mode := TRUE;
	ELSE
		IO.Put("I don't understand the tag " & token & "\n");
		Usage();
	END;

END Special;

(* Procedure to print Usage information. *)
PROCEDURE Usage()=
BEGIN

	IO.Put("Usage : " & Params.Get(0) & " [<sourcefile-name>] [/SERVER]\n");

END Usage;

VAR	progname : TEXT;	(* The program name. *)
	sourcename : TEXT;	(* Source file name. *)
	compiledname : TEXT;	(* Compiled code file name. *)
	parser := NEW(Parse.T);	(* Parser object. *)
	server_mode := FALSE;	(* Flag indicating whether or not to run the compiler as
					the local compilation server. *)

BEGIN

	IO.Put("\nArmadillo 3D Compiler version "
		& Fmt.Int(Tags.Version) & "."
		& Fmt.Int(Tags.Revision) & "\n");

	(* Examine the parameter line for interesting stuff. *)
	IF Params.Count > 1 THEN
		FOR x:=1 TO Params.Count-1 DO
			IF Text.GetChar(Params.Get(x), 0) = '/' THEN

				(* This is a command-line parameter, so do
					something special with it. *)
				Special(Params.Get(x));

			ELSE

				(* Have we already got a program name? *)
				IF progname = NIL THEN
					(* No, so use this one. *)
					progname := Params.Get(x);
				ELSE
					(* Yes, so print an error. *)
					progname := NIL;
					EXIT;
				END;	
			END;
		END;
	END;

	TRY

		IF server_mode THEN

			(* Run the compiler in server mode, exporting it with a known name. *)
			(* Interpreters may then connect and use it to allow command-line
				operation.  I think this is quite neat... *)

			parser.export();

		ELSE

			(* Compile the named file. *)
			IF progname = NIL THEN
				Usage();
			ELSE

				sourcename := progname & Tags.SourceExt;
				compiledname := progname & Tags.CodeExt;

				(* Try to compile the source file. *)
				parser.go(sourcename, compiledname);

			END;
		END;

	EXCEPT
		(* Print up any error message that is generated. *)
		Parse.Error=>IO.Put("\n" & parser.lastErr() & "\n");
	END;

END Main.