(* Parse module. *)

(*
	05/12/96	Created.
	11/12/96	Added clauses to the Atom function to
			cope with the "new" function and the
			"nil" reference constant.
	12/12/96	Added a new error code so that if the program
			is basically unparsable then the compiler can say
			so, rather than just saying "nothing to do!".
	14/12/96	Added a term to deal with the "colour" command.
	19/12/96	Made the "new" function into a command.
	23/12/96	Added the "str" function for producing text
			to represent data-type values.
	24/12/96	Added the "Command_RunTime" set of commands,
			to deal with system-related constraints.
	11/01/97	Added support for the "mutex" data-type.
	12/01/97	Added support for the "clearall" command.
	14/01/97	Added the "pause" command to the Debug suite.
	16/01/97	Updated the "function" type to include the function's
			signature as part of it.
	18/01/97	Added the "IntTerm" term to the syntax.
			Changed "nil" from a constant to a function.
	24/01/97	Added the new in-built value "time".
	02/02/97	Changed the IP address routine to use the Lexer object
			method "ip_address" since this is more general.
	04/02/97	Added clauses for the "port" data-type.
	19/02/97	Added the new "export" method to the parser object to
			allow interpreter units to connect to the compiler to
			perform on-the-fly compilations.
			Added the new "quit" command to the Control clause.
			Added the new "execute" command to the Debug clause.
	21/02/97	Modified the block variable and exception declarations to the
			new syntax.
			Also modified the function definition syntax.
	24/02/97	Modified the binding of arithmetic operators from right to left.
	26/02/97	Changed the structure of the "type" command.
	27/02/97	Modified the clause for the "exception" command to allow the
			non-block form.
	20/03/97	Added a clause for the Try...Finally command.
*)

MODULE Parse;

(* System modules. *)
IMPORT IO, OSError, FileWr, Wr, Pickle;
IMPORT Thread, NetObj;

(* Custom modules. *)
IMPORT Lexer;
IMPORT Code, Tags, NetParser;

TYPE ErrCode=	{
		FileOpen,		(* File failed to open. *)
		FileOutput,		(* Output file failed. *)
		BadPickle,		(* The pickler failed. *)
		NoProgram,		(* Parser returns NIL. *)
		NetFailed,		(* The parser couldn't be exported. *)

		Expected,		(* Didn't get what we expected. *)
		Syntax,			(* Couldn't match anything. *)
		BadIdent,		(* Got a bad identifier. *)
		ExceptList,		(* Was expecting an exception list. *)
		NoExceptTrap,		(* No exceptions in a TRY block. *)
		NotExpr,		(* Wanted an expression. *)
		BadLetParams,		(* Wrong number of expressions given. *)
		BadDispName,		(* Invalid name for a 3D display. *)
		BadVarList,		(* Was expecting a list of variable names. *)
		BadTypeDecl,		(* Wanted a type declaration. *)
		BadVarDecl,		(* Wanted a variable type declaration. *)
		BadFunDecl,		(* Wanted a function closure declaration. *)
		NotCommand,		(* Expected a command and didn't get one!. *)

		Dummy			(* Dummy error. *)
		};

REVEAL	T = Public BRANDED OBJECT
			lex : Lexer.T;
 			lasterr : ErrCode;
			lasttoken : TEXT;
			lasttext : TEXT;
		OVERRIDES
			go := Go;
			export := Export;
			lastErr := LastErr;
		END;

(* The Go procedure opens the input file and starts the parser,
	then saves out the object code to disk. *)
PROCEDURE Go(s : T; in_file, out_file : TEXT) RAISES {Error}=
VAR	program : Code.T;
	writer : FileWr.T;
	version := NEW(Code.Version);
BEGIN

	TRY

		IO.Put("Opening source file '" & in_file & "'\n");
		TRY
			s.lex := NEW(Lexer.T).init(in_file);	
		EXCEPT
			(* Couldn't open file, so fail. *)
			OSError.E=>
			s.lasterr := ErrCode.FileOpen;
			RAISE Error;
		END;

		TRY
			(* Do the compilation. *)
			program := Command(s);

			IF NOT(s.lex.eof()) THEN
				s.lasterr := ErrCode.NotCommand;
				RAISE Error;
			ELSIF program = NIL THEN
				s.lasterr := ErrCode.NoProgram;
				RAISE Error;
			END;

		EXCEPT
			(* If the parse fails, then save some text from it. *)
			Error=>	s.lasttext := s.lex.nextbytes();
				RAISE Error;
		END;

		IO.Put("Writing to output file '" & out_file & "'\n");
		TRY
			writer := FileWr.Open(out_file);

			(* Save the version stamp. *)
			Pickle.Write(writer, version);

			(* Now save the file. *)
			Pickle.Write(writer, program);

			Wr.Close(writer);

		EXCEPT
			Pickle.Error(t)=>
				s.lasterr := ErrCode.BadPickle;
				s.lasttoken := t;
				RAISE Error;
		|	OSError.E, Wr.Failure, Thread.Alerted=>
				s.lasterr := ErrCode.FileOutput;
				RAISE Error;
		END;

		IO.Put("Done compiling.\n");

	FINALLY

		(* Once we've finished parsing, let's close the file. *)
		IF s.lex # NIL THEN
			s.lex.close();
		END;

	END;

END Go;

(* This is the local sub-type of the network-parser object that will be
	exported if required, by this compiler. *)
TYPE	CustomSrvr = NetParser.T OBJECT
		OVERRIDES
			compile := NetCompile;
			test := NetTest;
		END;

(* Compile method for the custom server object. *)
PROCEDURE NetCompile(<* UNUSED *> self : CustomSrvr; t : TEXT) : Code.T RAISES {NetParser.Error}=
VAR	program : Code.T;
	s := NEW(T);
BEGIN

	TRY

		(* Initalise the lexer. *)
		s.lex := NEW(Lexer.T).fromtext(t);	

		TRY
			(* Do the compilation. *)
			program := Command(s);

			IF NOT(s.lex.eof()) THEN
				s.lasterr := ErrCode.NotCommand;
				RAISE Error;
			ELSIF program = NIL THEN
				s.lasterr := ErrCode.NoProgram;
				RAISE Error;
			END;

		EXCEPT
			(* If the parse fails, then save some text from it. *)
			Error=>	s.lasttext := s.lex.nextbytes();
				RAISE NetParser.Error(s.lastErr());
		END;

	FINALLY

		(* Once we've finished parsing, let's close the lexer. *)
		IF s.lex # NIL THEN
			s.lex.close();
		END;

	END;

	(* Return the compiled program to the caller. *)
	RETURN program;

END NetCompile;

(* This method does nothing except provide a way for the network to be tested. *)
PROCEDURE NetTest(<* UNUSED *> self : CustomSrvr)=
BEGIN
	IO.Put("Interpreter connected.\n");
END NetTest;

(* The Export procedure exports a network parser object with a well-known
	name, for interpreters to connect to. *)
PROCEDURE Export(s : T) RAISES {Error}=
VAR	server := NEW(CustomSrvr);
	dummy : TEXT;
BEGIN

	IO.Put("Exporting local parser object...\n");

	TRY
		NetObj.Export(NetParser.Name, server);
	EXCEPT
		NetObj.Error, Thread.Alerted=>
			s.lasterr := ErrCode.NetFailed;
			RAISE Error;
	END;

	(* If we get here then it worked, so wait for the user to tell us to quit! *)
	IO.Put("Press return to quit compilation server.\n");
	TRY
		dummy := IO.GetLine();
	EXCEPT
		IO.Error=>
	END;

END Export;

(* Returns the last error as explanatory text. *)
PROCEDURE LastErr(s : T) : TEXT=
VAR e : TEXT;
BEGIN

	IF s.lasttext # NIL THEN
		e := s.lasttext & "\n^  ";
	ELSE
		e := "\n";
	END;

	CASE s.lasterr OF
		ErrCode.FileOpen=>e:=e & "Could not open source file.";
	|	ErrCode.FileOutput=>e:=e & "Output to disk failed.";
	|	ErrCode.BadPickle=>e:=e & "Pickler error :" & s.lasttoken;
	|	ErrCode.NoProgram=>e:=e & "Source file contains no code!";

	|	ErrCode.Expected=>e:=e & "I was expecting '" & s.lasttoken & "'";
	|	ErrCode.Syntax=>e:=e & "Syntax error.";
	|	ErrCode.BadIdent=>e:=e & "I was expecting a valid identifier. ";
	|	ErrCode.ExceptList=>e:=e & "I was expecting an exception list.";
	|	ErrCode.NoExceptTrap=>e:=e & "No exceptions were specified in a TRY block.";
	|	ErrCode.NotExpr=>e:=e & "An expression was expected.";
	|	ErrCode.BadLetParams=>e:=e & "A let command must have 1 or N expressions.";
	|	ErrCode.BadDispName=>e:=e & "I was expecting a standard display name.";
	|	ErrCode.BadVarList=>e:=e & "I was expecting a list of variable names.";
	|	ErrCode.BadTypeDecl=>e:=e & "I wanted a type declaration."
	|	ErrCode.BadVarDecl=>e:=e & "I wanted a variable type declaration."
	|	ErrCode.BadFunDecl=>e:=e & "I wanted a function declaration."
	|	ErrCode.NotCommand=>e:=e & "Not a valid command.";

	ELSE
		e:=e & "An unknown error occurred.";
	END;

	RETURN e;

END LastErr;

(* MAIN PARSER ROUTINES! *)

PROCEDURE Command(s : T) : Code.T RAISES {Error}=
VAR cmd : Code.T;
BEGIN

	(* Get a command. *)	
	cmd := Command_Par(s);

	(* Check for a sequential command separator. *)
	IF s.lex.expect(";") THEN
		cmd.next := Command(s);
	END;

	RETURN cmd;

END Command;

PROCEDURE Command_Par(s : T) : Code.T RAISES {Error}=
VAR	cmd : Code.T;
BEGIN

	(* Get the first command. *)
	cmd := Command_Atom(s);

	IF s.lex.expect("&") THEN

		(* Put the command in a parallel construct. *)
		cmd := NEW(Code.T,
			type := Tags.T.CMD_Par,
			params := NEW(Code.ParamR, param := cmd)
			);

		(* Get the second command. *)
		cmd.params.next := NEW(Code.ParamR, param := Command_Par(s));

	END;	

	RETURN cmd;

END Command_Par;

PROCEDURE Command_Atom(s : T) : Code.T RAISES {Error}=
VAR cmd : Code.T := NIL;
BEGIN

	(* Check first for the atomiciser construct! *)
	IF s.lex.expect("[") THEN
		(* Bracketed commands act as atomic ones. *)
		cmd := NEW(Code.T, type := Tags.T.TAG_BracketCmd);
		cmd.params := NEW(Code.ParamR, param := Command(s));

		(* Match the brackets.*)
		MustBe(s,"]");
		RETURN cmd;
	END;

	(* If a particular parse fails, then it will return NIL. *)
	cmd := Command_Exceptions(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Control(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Variables(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Concurrency(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Types(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Graphics(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_Debug(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	cmd := Command_RunTime(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	(* The function clause MUST be last in the list, otherwise commands
		may be mistaken for user function calls. *)
	cmd := Command_Functions(s);
	IF cmd # NIL THEN
		RETURN cmd;
	END;

	RETURN NIL;

END Command_Atom;

PROCEDURE Command_Exceptions(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Check for an exception block. *)
	IF s.lex.expect("exception") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_ExceptDef);

		(* The first parameter is the list of exceptions. *)
		cmd.params := NEW(Code.ParamR, param := Exception_List(s));

		(* The exception list must not be empty! *)
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.ExceptList;
			RAISE Error;
		END;

		IF s.lex.expect("in") THEN
			(* The second parameter is the code to execute. *)
			cmd.params.next := NEW(Code.ParamR, param := Command_Atom(s));
		END;

	(* Check for the raise exception command. *)
	ELSIF s.lex.expect("raise") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Raise);

		(* Get list of exceptions to raise. *)
		cmd.params := NEW(Code.ParamR, param := Exception_List(s));

		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.ExceptList;
			RAISE Error;
		END;

		(* Check for the optional extra information string. *)
		IF s.lex.expect("with") THEN
			cmd.params.next := NEW(Code.ParamR, param := Expression(s));
			IF cmd.params.next.param = NIL THEN
				s.lasterr := ErrCode.NotExpr;
				RAISE Error;
			END;
		END;

	(* Check for an exception trapping block. *)
	ELSIF s.lex.expect("try") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_TryExcept);

		(* Get the code to be executed! *)
		cmd.params := NEW(Code.ParamR, param := Command(s));

		(* The block could be Try...Except or Try...Finally! *)
		IF s.lex.expect("except") THEN

			(* Next, exceptions to trap and the code to execute for each. *)
			cmd.params.next := NEW(Code.ParamR, param := ExceptTrap_List(s));

			(* Having no trapped exceptions in a trap block is silly. *)
			IF cmd.params.next.param = NIL THEN
				s.lasterr := ErrCode.NoExceptTrap;
				RAISE Error;
			END;

		ELSIF s.lex.expect("finally") THEN

			(* Change the command code tag. *)
			cmd.type := Tags.T.BLK_TryFinally;

			(* Get the second command. *)
			cmd.params.next := NEW(Code.ParamR, param := Command(s));
			
		ELSE
			s.lasterr := ErrCode.Expected;
			s.lasttoken := "except/finally";
			RAISE Error;
		END;

		MustBe(s, "end");

	ELSE
		RETURN NIL;
	END;		
		
	RETURN cmd;

END Command_Exceptions;

PROCEDURE Command_Control(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
	tail : Code.ParamR;
BEGIN

	(* Check for conditional execution. *)
	IF s.lex.expect("if") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_If);

		(* Get the expression to test. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		(* There MUST be an expression to test! *)
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

		MustBe(s,"then");

		(* Get the code to execute for TRUE. *)
		cmd.params.next := NEW(Code.ParamR, param := Command(s));

		(* Now look for any number of Elsif clauses. *)
		cmd.params.next.next := NEW(Code.ParamR);
		tail := cmd.params.next.next;
		WHILE s.lex.expect("elsif") DO

			tail.param := NEW(Code.T, type := Tags.T.CMD_ElsIf);

			(* Get an expression. *)
			tail.param.params := NEW(Code.ParamR, param := Expression(s));
			IF tail.param.params.param = NIL THEN
				s.lasterr := ErrCode.NotExpr;
				RAISE Error;
			END;

			MustBe(s, "then");

			(* Get a command. *)
			tail.param.params.next := NEW(Code.ParamR, param := Command(s));

			(* Set the tail to the new value. *)
			tail.param.params.next.next := NEW(Code.ParamR);
			tail := tail.param.params.next.next;
		END;

		(* Now check for an Else clause! *)
		IF s.lex.expect("else") THEN
			(* Get the code to execute for FALSE. *)
			cmd.params.next.next.next := NEW(Code.ParamR, param := Command(s));
		END;

		MustBe(s,"end");

	(* Loop construct. *)
	ELSIF s.lex.expect("loop") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_Loop);

		(* Get the code to execute. *)
		cmd.params := NEW(Code.ParamR, param := Command(s));

		MustBe(s,"end");

	(* Check for exit command. *)
	ELSIF s.lex.expect("exit") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Exit);

	(* Return command. *)
	ELSIF s.lex.expect("return") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Return);

		(* Check for an expression to return. *)
		VAR	expr : Code.T;
		BEGIN
			expr := Expression(s);
			IF expr # NIL THEN
				cmd.params := NEW(Code.ParamR, param := expr);
			END;
		END;

	(* Check for quit command. *)
	ELSIF s.lex.expect("quit") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Quit);

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Control;

PROCEDURE Command_Variables(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Check for variable declaration. *)
	IF s.lex.expect("var") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_Var);

		(* Get the list of variables to define. *)
		cmd.params := NEW(Code.ParamR, param := VarDecl_List(s));

		IF s.lex.expect("in") THEN

			(* Get the command to execute. *)
			cmd.params.next := NEW(Code.ParamR, param := Command_Atom(s));

		END;

	ELSIF s.lex.expect("let") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Let);
		VAR	n, m : CARDINAL;
		BEGIN
			(* The following gets a list of variables and
				returns how many it got. *)
			cmd.params := NEW(Code.ParamR, param := VariableF_NList(s, n));

			MustBe(s, "=");

			(* The following gets a list of expressions and
				returns how many it got. *)
			cmd.params.next := NEW(Code.ParamR, param := Expression_NList(s, m));

			(* Check that only one expression was given, or the same
				number as there were variable names. *)
			IF (m # 1) AND (m # n) THEN
				s.lasterr := ErrCode.BadLetParams;
				RAISE Error;
			END;

			IF (m = 1) AND (m # n) THEN
				(* Change the type of the root node. *)
				cmd.type := Tags.T.CMD_LetAll;
			END;

		END;

	ELSIF s.lex.expect("new") THEN
		cmd := NEW(Code.T,
				type := Tags.T.CMD_New,
				params := NEW(Code.ParamR, param := VariableF(s))
			);
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Variables;

PROCEDURE Command_Concurrency(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Check for the readonly command construct. *)
	IF s.lex.expect("readonly") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_Readonly);

		(* Get the list of variable names. *)
		cmd.params := NEW(Code.ParamR, param := VariableF_List(s));

		MustBe(s, "in");

		(* Get the command to execute. *)
		cmd.params.next := NEW(Code.ParamR, param := Command(s));

		MustBe(s, "end");

	(* Check for the locking construct. *)
	ELSIF s.lex.expect("lock") THEN
		cmd := NEW(Code.T, type := Tags.T.BLK_Lock);

		(* Get the names of variables to lock. *)
		cmd.params := NEW(Code.ParamR, param := VariableF_List(s));

		MustBe(s, "in");

		(* Get the code to execute with them locked. *)
		cmd.params.next := NEW(Code.ParamR, param := Command(s));

		MustBe(s, "end");

	(* The signalling command. *)
	ELSIF s.lex.expect("signal") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Signal);

		(* Get the name of the variable to signal. *)
		cmd.params := NEW(Code.ParamR, param := VariableF(s));

		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

	(* The waiting command. *)
	ELSIF s.lex.expect("wait") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Wait);

		(* Get the name of the variable to wait on. *)
		cmd.params := NEW(Code.ParamR, param := VariableF(s));

		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		(* Check for a With clause to the command. *)
		IF s.lex.expect("with") THEN
			(* Get a list of variables to unlock while waiting. *)
			cmd.params.next := NEW(Code.ParamR, param := VariableF_List(s));

			IF cmd.params.next.param = NIL THEN
				s.lasterr := ErrCode.BadVarList;
				RAISE Error;
			END;

		END;

	(* The port send command. *)
	ELSIF s.lex.expect("send") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Send);

		(* Get the data to send. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));

		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		MustBe(s, "on");

		(* Get the name of the port to send on. *)
		cmd.params.next := NEW(Code.ParamR, param := VariableF(s));

		IF cmd.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		(* Check for a With clause to the command. *)
		IF s.lex.expect("with") THEN
			(* Get a list of variables to unlock while waiting. *)
			cmd.params.next.next := NEW(Code.ParamR, param := VariableF_List(s));

			IF cmd.params.next.next.param = NIL THEN
				s.lasterr := ErrCode.BadVarList;
				RAISE Error;
			END;

		END;

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Concurrency;

PROCEDURE Command_Functions(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Check for a function definition. *)
	IF s.lex.expect("fun") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Fun);

		(* Get the function declarations. *)
		cmd.params := NEW(Code.ParamR, param := FunDecl_List(s));

		IF s.lex.expect("in") THEN

			(* Get the code to execute. *)
			cmd.params.next := NEW(Code.ParamR, param := Command_Atom(s));

		END;

	(* Function call. *)
	ELSIF s.lex.expect("call") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Call);

		(* Get the name of the function. *)
		cmd.params := NEW(Code.ParamR, param := VariableF(s));
		(* A name MUST be specified. *)
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		MustBe(s, "(");

		(* Get the parameters to the call. *)
		cmd.params.next := NEW(Code.ParamR, param := Expression_List(s));

		MustBe(s, ")");

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Functions;

PROCEDURE Command_Types(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Type declaration. *)
	IF s.lex.expect("type") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Type);

		(* Get the type declaration list. *)
		cmd.params := NEW(Code.ParamR, param := TypeNameDecl(s));

		(* Check for a command to execute. *)
		IF s.lex.expect("in") THEN
			cmd.params.next := NEW(Code.ParamR, param := Command_Atom(s));
		END;

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Types;

PROCEDURE Command_Graphics(s : T) : Code.T RAISES{Error}=
VAR cmd : Code.T;
BEGIN

	(* All the bestest graphics commands! *)

	IF s.lex.expect("turtle") THEN
		(* Note that "turtle" must be parsed before "tup", or
			problems occur due to the abbreviation "tu". *)
		cmd := NEW(Code.T, type := Tags.T.GFX_Turtle);

		(* Get the variable from which to select the turtle. *)
		cmd.params := NEW(Code.ParamR, param := VariableF(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

	ELSIF s.lex.expect("forward") OR s.lex.expect("fd") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Forward);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("backward") OR s.lex.expect("bk") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Backward);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("tleft") OR s.lex.expect("tl") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_TLeft);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("tright") OR s.lex.expect("tr") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_TRight);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("tup") OR s.lex.expect("tu") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_TUp);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("tdown") OR s.lex.expect("td") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_TDown);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("rleft") OR s.lex.expect("rl") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_RLeft);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("rright") OR s.lex.expect("rr") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_RRight);
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("penup") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Penup);

	ELSIF s.lex.expect("pendown") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Pendown);

	ELSIF s.lex.expect("home") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Home);

	ELSIF s.lex.expect("clearall") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_ClearAll);

	ELSIF s.lex.expect("clear") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Clear);

	ELSIF s.lex.expect("attach") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Attach);

		(* Get the address of the display to attach to. *)
		cmd.params := NEW(Code.ParamR, param := Display(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadDispName;
			RAISE Error;
		END;

	ELSIF s.lex.expect("colour") THEN
		cmd := NEW(Code.T, type := Tags.T.GFX_Colour);

		(* Get the three expressions. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

		MustBe(s, ",");

		cmd.params.next := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.next.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

		MustBe(s, ",");

		cmd.params.next.next := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.next.next.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;
	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Graphics;

PROCEDURE Command_Debug(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Useful output routines. *)

	(* The standard print routine. *)
	IF s.lex.expect("print") THEN
		cmd := NEW(Code.T, type := Tags.T.BUG_Print);

		(* Get the expression to print. *)
		cmd.params := NEW(Code.ParamR, param := Expression_List(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("assert") THEN
		cmd := NEW(Code.T, type := Tags.T.BUG_Assert);

		(* Get the expression to be tested. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("pause") THEN
		cmd := NEW(Code.T, type := Tags.T.BUG_Pause);

		(* Get the expression to be tested. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("execute") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Execute);

		(* Get the name of the file to load. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSIF s.lex.expect("collect") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_Collect);

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_Debug;

(* The following function parses some optional language features for
	modifying and monitoring system performance. *)
PROCEDURE Command_RunTime(s : T) : Code.T RAISES{Error}=
VAR	cmd : Code.T;
BEGIN

	(* Set max thread limit. *)
	IF s.lex.expect("maxthread") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_MaxThread);

		(* Get the max value. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	(* Set max recursion limit. *)
	ELSIF s.lex.expect("maxfunction") THEN
		cmd := NEW(Code.T, type := Tags.T.CMD_MaxFunction);

		(* Get the expression to be tested. *)
		cmd.params := NEW(Code.ParamR, param := Expression(s));
		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

	ELSE
		RETURN NIL;
	END;

	RETURN cmd;

END Command_RunTime;

(* Stuff to do with exceptions. *)

PROCEDURE Exception(s : T) : Code.T=
VAR	name : TEXT;
	code : Code.T;
BEGIN

	TRY

		(* Get the lexer to get a token. *)
		name := s.lex.ident();
		code := NEW(Code.T, type := Tags.T.TAG_Exception, extra := name);
		RETURN code;

	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;

END Exception;

PROCEDURE Exception_List(s : T) : Code.T RAISES{Error}=
VAR code : Code.T;
BEGIN

	(* Get the first exception name. *)
	code := Exception(s);

	(* Check that we got a valid exception. *)
	IF code = NIL THEN
		(* Not a valid exception, so return NIL. *)
		RETURN NIL;
	END;

	(* Check for more exceptions. *)
	IF s.lex.expect(",") THEN
		code.params := NEW(Code.ParamR);
		code.params.param := Exception_List(s);
		IF code.params.param = NIL THEN
			(* We got one valid exception name, but not
				the second, so raise an error. *)
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;
	END;

	RETURN code;

END Exception_List;

PROCEDURE ExceptTrap(s : T) : Code.T RAISES {Error}=
VAR	names : Code.T;
	code : Code.T;
	cmd : Code.T;
BEGIN

	(* Get the names of the exceptions to trap. *)
	names := Exception_List(s);
	IF names = NIL THEN
		(* No exception names given! *)
		RETURN NIL;
	END;

	MustBe(s, "=>");

	(* Get the code to execute. *)
	code := Command(s);

	(* Put the names and code in an ExceptTrap. *)
	cmd := NEW(Code.T,
			type := Tags.T.TAG_ExceptTrap
			);

	(* Put name. *)
	cmd.params := NEW(Code.ParamR, param := names);

	(* Put code. *)
	cmd.params.next := NEW(Code.ParamR, param := code);

	RETURN cmd;

END ExceptTrap;

PROCEDURE ExceptTrap_List(s : T) : Code.T RAISES {Error}=
VAR	cmd : Code.T;
BEGIN

	cmd := NEW(Code.T, type := Tags.T.TAG_ExceptTrap_List);

	(* Get the first ExceptTrap. *)
	cmd.params := NEW(Code.ParamR, param := ExceptTrap(s));

	(* Check for more traps. *)
	IF s.lex.expect("|") THEN
		(* Add more ExceptTraps! *)
		cmd.params.next := NEW(Code.ParamR, param := ExceptTrap_List(s));
	END;

	RETURN cmd;

END ExceptTrap_List;

(* Variable related functions. *)
PROCEDURE VarDecl(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
	varnames : Code.T;
	type : Code.T;
BEGIN

	(* Get the names of the variables. *)
	varnames := Variable_List(s);
	IF varnames = NIL THEN
		RETURN NIL;
	END;

	MustBe(s, ":");

	(* Get the type that they will be of. *)
	type := TypeDecl(s);
	IF type = NIL THEN
		s.lasterr := ErrCode.BadTypeDecl;
		RAISE Error;
	END;

	(* Create the VarDecl node. *)
	code := NEW(Code.T, type := Tags.T.TAG_VarDecl);
	code.params := NEW(Code.ParamR, param := varnames);
	code.params.next := NEW(Code.ParamR, param := type);

	RETURN code;

END VarDecl;

PROCEDURE VarDecl_List(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	code := NEW(Code.T, type := Tags.T.TAG_VarDeclList);

	(* Get the first variable declaration. *)
	code.params := NEW(Code.ParamR, param := VarDecl(s));
	IF code.params.param = NIL THEN
		RETURN NIL;
	END;

	(* Check for more declarations. *)
	IF s.lex.expect(",") THEN
		(* Get the rest of the list. *)
		code.params.next := NEW(Code.ParamR, param := VarDecl_List(s));
		IF code.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadVarDecl;
			RAISE Error;
		END;
	END;

	RETURN code;

END VarDecl_List;

PROCEDURE Variable(s : T) : Code.T=
VAR	name : TEXT;
	code : Code.T;
BEGIN
	TRY

		(* Get the lexer to get a token. *)
		name := s.lex.ident();
		code := NEW(Code.T, type := Tags.T.TAG_Variable, extra := name);
		RETURN code;

	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;
END Variable;

PROCEDURE VariableFD(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	(* Check for the dereference operator. *)
	IF s.lex.expect("^") THEN

		(* Create a deref operator to do the stuff. *)
		code := NEW(Code.T,
			type := Tags.T.REF_Deref,
			params := NEW(Code.ParamR, param := VariableFD(s))
			);
		RETURN code;

	ELSE

		(* Just return the result of Variable. *)
		RETURN Variable(s);

	END;

END VariableFD;

PROCEDURE VariableF(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	(* Get a variable name. *)
	code := VariableFD(s);
	IF code = NIL THEN RETURN NIL; END;

	(* Check for the "." dereference character. *)
	WHILE s.lex.expect(".") DO

		code := NEW(Code.T,
				type := Tags.T.TAG_RecordRef,
				params := NEW(Code.ParamR,
						param := code
					)
		);
		code.params.next := NEW(Code.ParamR,
					param := VariableFD(s)
				);
		IF code.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

	END;

	RETURN code;

END VariableF;

PROCEDURE Variable_List(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	code := NEW(Code.T, type := Tags.T.TAG_VarList);

	(* Get the first variable name. *)
	code.params := NEW(Code.ParamR, param := Variable(s));
	IF code.params.param = NIL THEN
		RETURN NIL;
	END;

	(* Check for more items. *)
	IF s.lex.expect(",") THEN
		code.params.next := NEW(Code.ParamR, param := Variable_List(s));
		IF code.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadVarList;
			RAISE Error;
		END;
	END;

	RETURN code;

END Variable_List;

PROCEDURE VariableF_List(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	code := NEW(Code.T, type := Tags.T.TAG_VarList);

	(* Get the first variable name. *)
	code.params := NEW(Code.ParamR, param := VariableF(s));
	IF code.params.param = NIL THEN
		RETURN NIL;
	END;

	(* Check for more items. *)
	IF s.lex.expect(",") THEN
		code.params.next := NEW(Code.ParamR, param := VariableF_List(s));
		IF code.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadVarList;
			RAISE Error;
		END;
	END;

	RETURN code;

END VariableF_List;

PROCEDURE VariableF_NList(s : T; VAR n : CARDINAL) : Code.T RAISES {Error}=
VAR	code : Code.T;
	node : Code.T;
BEGIN

	(* Get the variable list. *)
	code := VariableF_List(s);

	(* And count the entries. *)
	n := 0;
	node := code;
	WHILE node # NIL DO
		INC(n);
		IF node.params.next = NIL THEN
			EXIT;
		ELSE
			node := node.params.next.param;
		END;
	END;

	RETURN code;

END VariableF_NList;

(* Functions related functions. *)

PROCEDURE FunDecl(s : T) : Code.T RAISES {Error}=
VAR	cmd : Code.T;
BEGIN

	cmd := NEW(Code.T, type := Tags.T.TAG_FunDecl);

	(* Get the name of the new function. *)
	cmd.params := NEW(Code.ParamR, param := Variable(s));
	IF cmd.params.param = NIL THEN
		s.lasterr := ErrCode.BadIdent;
		RAISE Error;
	END;

	MustBe(s, "(");

	(* Get the names of the parameters. *)
	cmd.params.next := NEW(Code.ParamR, param := VarDecl_List(s));

	MustBe(s, ")");

	IF s.lex.expect(":") THEN
		(* Get the return type of the function. *)
		cmd.params.next.next := NEW(Code.ParamR, param := TypeDecl(s));
	ELSE
		(* Put in an empty placeholder. *)
		cmd.params.next.next := NEW(Code.ParamR, param := NIL);
	END;

	MustBe(s, "=");

	(* Get the code for the function's body. *)
	cmd.params.next.next.next := NEW(Code.ParamR, param := Command_Atom(s));

	RETURN cmd;

END FunDecl;

PROCEDURE FunDecl_List(s : T) : Code.T RAISES {Error}=
VAR	cmd : Code.T;
BEGIN

	cmd := NEW(Code.T, type := Tags.T.TAG_FunDeclList);

	(* Get the name of the first function. *)
	cmd.params := NEW(Code.ParamR, param := FunDecl(s));
	IF cmd.params.param = NIL THEN
		s.lasterr := ErrCode.BadFunDecl;
		RAISE Error;
	END;

	(* Check for more function declarations. *)
	IF s.lex.expect(",") THEN
		cmd.params.next := NEW(Code.ParamR, param := FunDecl_List(s));
	END;

	RETURN cmd;

END FunDecl_List;

(* Graphics related functions. *)

PROCEDURE Display(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	code := NEW(Code.T, type := Tags.T.TAG_Display);

	(* Get the actual name of the display. *)
	code.params := NEW(Code.ParamR, param := DisplayName(s));
	IF code.params.param = NIL THEN
		s.lasterr := ErrCode.BadDispName;
		RAISE Error;
	END;

	(* Check for an IP address as well. *)
	(* Not sure what the definition of an IP address format is... *)
	IF s.lex.expect("@") THEN
		code.params.next := NEW(Code.ParamR, param := IPAddress(s));
		IF code.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadDispName;
			RAISE Error;
		END;
	END;

	RETURN code;

END Display;

PROCEDURE DisplayName(s : T) : Code.T=
VAR	name : TEXT;
	code : Code.T;
BEGIN
	TRY

		(* Get the lexer to get a token. *)
		name := s.lex.ident();
		code := NEW(Code.T, type := Tags.T.TAG_DisplayName, extra := name);
		RETURN code;
	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;
END DisplayName;

PROCEDURE IPAddress(s : T) : Code.T=
BEGIN

	(* Rough and ready "get an ip address" routine. *)

	TRY
		RETURN NEW(Code.T,
			type := Tags.T.TAG_IPAddress,
			extra := s.lex.ip_address()
			);
	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;

END IPAddress;

(* Types related functions. *)

PROCEDURE TypeName(s : T) : Code.T=
VAR	name : TEXT;
	code : Code.T;
BEGIN
	TRY
		(* Get the lexer to get a token. *)
		name := s.lex.ident();
		code := NEW(Code.T, type := Tags.T.TAG_Type, extra := name);
		RETURN code;

	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;
END TypeName;

PROCEDURE TypeDecl(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	(* Get an atomic type declarator. *)
	code := TypeAtom(s);
	IF code = NIL THEN RETURN NIL; END;

	(* Check for it being followed by "port". *)
	WHILE s.lex.expect("port") DO
		(* Update the code. *)
		code := NEW(Code.T,
				type := Tags.T.TYP_Port,
				params := NEW(Code.ParamR, param := code)
			);
	END;

	RETURN code;
END TypeDecl;

PROCEDURE TypeNameDecl(s : T) : Code.T RAISES {Error}=
VAR	cmd : Code.T;
BEGIN

		cmd := NEW(Code.T, type := Tags.T.TAG_TypeDecl);

		(* Get the name of the new type. *)
		cmd.params := NEW(Code.ParamR, param := TypeName(s));

		IF cmd.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		MustBe(s, "=");

		(* Get the description of the type. *)
		cmd.params.next := NEW(Code.ParamR, param := TypeDecl(s));

		IF cmd.params.next.param = NIL THEN
			s.lasterr := ErrCode.BadTypeDecl;
			RAISE Error;
		END;

		IF s.lex.expect(",") THEN
			cmd.params.next.next := NEW(Code.ParamR, param := TypeNameDecl(s));
		END;

		RETURN cmd;

END TypeNameDecl;

PROCEDURE TypeAtom(s : T) : Code.T RAISES {Error}=
VAR	code : Code.T;
BEGIN

	(* Check for bracketting. *)
	IF s.lex.expect("(") THEN
		code := TypeDecl(s);
		MustBe(s, ")");

	(* Check for a number type. *)
	ELSIF s.lex.expect("number") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Number);

	ELSIF s.lex.expect("condition") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Condition);

	ELSIF s.lex.expect("string") THEN
		code := NEW(Code.T, type := Tags.T.TYP_String);

	ELSIF s.lex.expect("boolean") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Boolean);

	ELSIF s.lex.expect("record") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Record);

		(* Get the sub-parameters. *)
		code.params := NEW(Code.ParamR, param := VarDecl_List(s));

		MustBe(s, "end");

	ELSIF s.lex.expect("turtle") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Turtle);

	ELSIF s.lex.expect("ref") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Ref);

		(* Get the type of thing it's a reference to. *)
		code.params := NEW(Code.ParamR, param := TypeDecl(s));
		IF code.params.param = NIL THEN
			s.lasterr := ErrCode.BadTypeDecl;
			RAISE Error;
		END;

	ELSIF s.lex.expect("function") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Function);

		(* Should also have a signature. *)
		MustBe(s, "(");

		(* Get the list of parameters - could be NIL. *)
		code.params := NEW(Code.ParamR, param := VarDecl_List(s));

		MustBe(s, ")");

		(* Check for a return-value type. *)
		IF s.lex.expect(":") THEN
			code.params.next := NEW(Code.ParamR, param := TypeDecl(s));
			(* The return type cannot be NIL. *)
			IF code.params.next.param = NIL THEN
				s.lasterr := ErrCode.BadTypeDecl;
				RAISE Error;
			END;
		END;

	ELSIF s.lex.expect("mutex") THEN
		code := NEW(Code.T, type := Tags.T.TYP_Mutex);

	ELSE
		RETURN TypeName(s);
	END;

	RETURN code;

END TypeAtom;

(* Expression functions: *)

PROCEDURE Expression(s : T) : Code.T RAISES {Error}=
VAR	expr : Code.T;
	type : Tags.T;
BEGIN

	(* Get an expression term. *)
	expr := IntTerm(s);
	IF expr = NIL THEN RETURN NIL; END;
 
	(* Enter a loop checking for extra connective operators. *)
	LOOP

		(* Check for dyadic plus. *)
		IF s.lex.expect("+") THEN
			type := Tags.T.EXP_Plus;

		(* Minus. *)
		ELSIF s.lex.expect("-") THEN
			type := Tags.T.EXP_Minus;

		(* Boolean Or.*)
		ELSIF s.lex.expect("or") THEN
			type := Tags.T.BOOL_Or;

		ELSE
			type := Tags.T.NO_OP;
		END;

		IF type # Tags.T.NO_OP THEN
			expr := NEW(Code.T,
						type := type,
						params := NEW(Code.ParamR, param := expr)
					);
			expr.params.next := NEW(Code.ParamR, param := IntTerm(s));
		ELSE
			RETURN expr;
		END;

	END;

END Expression;

PROCEDURE IntTerm(s : T) : Code.T RAISES {Error}=
VAR	iterm : Code.T;
	type : Tags.T;
BEGIN
	
	iterm := Term(s);
	IF iterm = NIL THEN RETURN NIL; END;

	LOOP

		(* Check for integer division. *)
		IF s.lex.expect("div") THEN
			type := Tags.T.EXP_IntDiv;

		(* Integer remainder. *)
		ELSIF s.lex.expect("mod") THEN
			type := Tags.T.EXP_IntMod;

		ELSE
			type := Tags.T.NO_OP;
		END;

		IF type # Tags.T.NO_OP THEN
			iterm := NEW(Code.T,
						type := type,
						params := NEW(Code.ParamR, param := iterm)
					);
			iterm.params.next := NEW(Code.ParamR, param := Term(s));
		ELSE
			RETURN iterm;
		END;		

	END;

END IntTerm;

PROCEDURE Term(s : T) : Code.T RAISES {Error}=
VAR	term : Code.T;
	type : Tags.T;
BEGIN

	term := BoolOp(s);
	IF term = NIL THEN RETURN NIL; END;

	LOOP

		(* Check for multiplication. *)
		IF s.lex.expect("*") THEN
			type := Tags.T.EXP_Times;

		(* Division. *)
		ELSIF s.lex.expect("/") THEN
			type := Tags.T.EXP_Divide;

		ELSIF s.lex.expect("and") THEN
			type := Tags.T.BOOL_And;

		ELSE
			type := Tags.T.NO_OP;
		END;

		IF type # Tags.T.NO_OP THEN
			term := NEW(Code.T,
						type := type,
						params := NEW(Code.ParamR, param := term)
					);
			term.params.next := NEW(Code.ParamR, param := Term(s));
		ELSE
			RETURN term;
		END;		

	END;

END Term;

PROCEDURE BoolOp(s : T) : Code.T RAISES {Error}=
VAR	bool : Code.T;
	type : Tags.T;
BEGIN

	bool := Atom(s);
	IF bool = NIL THEN RETURN NIL; END;

	LOOP

		(* Implies. *)
		IF s.lex.expect("=>") THEN
			type := Tags.T.BOOL_Implies;

		ELSIF s.lex.expect(">=") THEN
			type := Tags.T.EXP_MoreOrEq;

		ELSIF s.lex.expect("<=") THEN
			type := Tags.T.EXP_LessOrEq;

		ELSIF s.lex.expect("<") THEN
			type := Tags.T.EXP_LessThan;

		ELSIF s.lex.expect(">") THEN
			type := Tags.T.EXP_MoreThan;

		ELSIF s.lex.expect("=") THEN
			type := Tags.T.EXP_Equal;

		ELSIF s.lex.expect("#") THEN
			type := Tags.T.EXP_NotEqual;

		ELSE
			type := Tags.T.NO_OP;
		END;

		IF type # Tags.T.NO_OP THEN
			bool := NEW(Code.T,
						type := type,
						params := NEW(Code.ParamR, param := bool)
					);
			bool.params.next := NEW(Code.ParamR, param := BoolOp(s));
		ELSE
			RETURN bool;
		END;

	END;

END BoolOp;

PROCEDURE Atom(s : T) : Code.T RAISES {Error}=
VAR	atom : Code.T;
	type := Tags.T.NO_OP;
BEGIN

	(* Check for brackets. *)
	IF s.lex.expect("(") THEN
		atom := NEW(Code.T, type := Tags.T.EXP_Bracket);
		atom.params := NEW(Code.ParamR, param := Expression(s));
		IF atom.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

		MustBe(s, ")");

		RETURN atom;

	(* The following functions are monadic and take Atoms. *)

	ELSIF s.lex.expect("not") THEN
		type := Tags.T.BOOL_Not;

	ELSIF s.lex.expect("rnd") THEN
		type := Tags.T.FUN_Rnd;

	ELSIF s.lex.expect("nil") THEN
		RETURN NEW(Code.T, type := Tags.T.TAG_Nil);

	ELSIF s.lex.expect("receive") THEN
		(* The "recieve" function for ports. *)
		atom := NEW(Code.T, type := Tags.T.FUN_Recieve);

		MustBe(s, "(");

		(* Get the name of the port. *)
		atom.params := NEW(Code.ParamR,
					param := VariableF(s)
				);
		IF atom.params.param = NIL THEN
			s.lasterr := ErrCode.BadIdent;
			RAISE Error;
		END;

		(* Check for a list of variables to unlock. *)
		IF s.lex.expect("with") THEN
			atom.params.next := NEW(Code.ParamR,
						param := VariableF_List(s)
					);
			IF atom.params.next.param = NIL THEN
				s.lasterr := ErrCode.BadVarList;
				RAISE Error;
			END;
		END;

		MustBe(s, ")");
		RETURN atom;

	ELSIF s.lex.expect("time") THEN
		RETURN NEW(Code.T, type := Tags.T.TAG_Time);

	ELSIF s.lex.expect("str") THEN
		type := Tags.T.FUN_Str;

	ELSIF s.lex.expect("!") THEN
		type := Tags.T.REF_Pointer;

	ELSIF s.lex.expect("+") THEN
		type := Tags.T.MONO_Plus;

	ELSIF s.lex.expect("-") THEN
		type := Tags.T.MONO_Minus;

	ELSE
		(* Check for constants. *)
		atom := Const(s);
		IF atom # NIL THEN
			RETURN atom;
		END;

		(* Check for variable name. *)
		atom := VariableF(s);
		IF atom # NIL THEN
			(* Check for parameters, which would make this a function call. *)
			IF s.lex.expect("(") THEN
				atom := NEW(Code.T,
						type := Tags.T.CMD_Call,
						params := NEW(Code.ParamR,
								param := atom)
					);

				(* Get the parameters to the call. *)
				atom.params.next := NEW(Code.ParamR,
							param := Expression_List(s)
						);

				MustBe(s, ")");
			END;
			RETURN atom;
		END;

	END;

	IF type # Tags.T.NO_OP THEN
		atom := NEW(Code.T, type := type);
		atom.params := NEW(Code.ParamR, param := Atom(s));
		IF atom.params.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;
		RETURN atom;
	ELSE
		RETURN NIL;
	END;

END Atom;

PROCEDURE Const(s : T) : Code.T=
VAR	code : Code.T;
BEGIN

	(* Check for constant values! *)

	(* I check for booleans here, because it's easy. *)
	IF s.lex.expect("true") THEN
		code := NEW(Code.T, type := Tags.T.TAG_Boolean);
		code.extra := NEW(Code.BoolR, value := TRUE);

	ELSIF s.lex.expect("false") THEN
		code := NEW(Code.T, type := Tags.T.TAG_Boolean);
		code.extra := NEW(Code.BoolR, value := FALSE);

	ELSE

		(* Check for some other types. *)

		code := Number(s);
		IF code # NIL THEN
			RETURN code;
		END;

		code := String(s);
		IF code # NIL THEN
			RETURN code;
		END;

		RETURN NIL;
	END;

	RETURN code;

END Const;

PROCEDURE Number(s : T) : Code.T=
BEGIN

	TRY
		RETURN NEW(Code.T,
			type := Tags.T.TAG_Real,
			extra := NEW(Code.RealR, value := s.lex.real())
			);
	EXCEPT
		(* If not a real, then parse as an integer. *)
		Lexer.BadToken=>
	END;

	TRY
		RETURN NEW(Code.T,
			type := Tags.T.TAG_Int,
			extra := NEW(Code.IntR, value := s.lex.integer())
			);
	EXCEPT
		(* If not an integer either then fail! *)
		Lexer.BadToken=>RETURN NIL;
	END;

END Number;

PROCEDURE String(s : T) : Code.T=
BEGIN

	TRY
		RETURN NEW(Code.T,
			type := Tags.T.TAG_String,
			extra := s.lex.string()
			);
	EXCEPT
		Lexer.BadToken=>RETURN NIL;
	END;

END String;

PROCEDURE Expression_List(s : T) : Code.T RAISES {Error}=
VAR	list : Code.T;
BEGIN

	(* Create a list entry. *)
	list := NEW(Code.T, type := Tags.T.TAG_ExprList);
	list.params := NEW(Code.ParamR, param := Expression(s));
	IF list.params.param = NIL THEN
		RETURN NIL;
	END;

	(* Check for more expressions. *)
	IF s.lex.expect(",") THEN
		(* Get the rest of the list. *)
		list.params.next := NEW(Code.ParamR, param := Expression_List(s));
		IF list.params.next.param = NIL THEN
			s.lasterr := ErrCode.NotExpr;
			RAISE Error;
		END;

		RETURN list;
	END;

	RETURN list;

END Expression_List;

PROCEDURE Expression_NList(s : T; VAR n : CARDINAL) : Code.T RAISES {Error}=
VAR	list : Code.T;
	pos : Code.T;
BEGIN

	(* Get the expression list. *)
	list := Expression_List(s);

	(* Now count its elements. *)
	n := 0;
	pos := list;
	WHILE pos # NIL DO
		INC(n);
		IF pos.params.next = NIL THEN
			EXIT;
		ELSE
			pos := pos.params.next.param;
		END;
	END;

	RETURN list;

END Expression_NList;

(* INTERNAL functions. *)

(* The MustBe function checks for a token and if it is not
	found next then an error is raised. *)
PROCEDURE MustBe(s : T; token : TEXT) RAISES {Error}=
BEGIN

	IF NOT(s.lex.expect(token)) THEN
		s.lasterr := ErrCode.Expected;
		s.lasttoken := token;
		RAISE Error;
	END;

END MustBe;

BEGIN

END Parse.