(* Armadillo 3D Display Program. 
	This is the implementation of the Window module used
	in the 3D terminal part of the Armadillo project.
*)

(*
	25/11/96	Created.
	04/02/97	Modified the Init routine to automatically use
			whatever graphics system is available.
			The Init routine now takes account of the "prefBase"
			environment variable when trying to get a display
			base.
*)

MODULE Window;

(* The three possible base graphics libraries. *)
IMPORT Win_OpenGL_Base, X_PEX_Base, X_OpenGL_Base;

(* General *)
IMPORT Thread, Env, Text;

(* Anim3d display libraries. *)
IMPORT GraphicsBase, ColorProp;
IMPORT GO, GroupGO, RootGO, LineGO, CylinderGO, SurfaceGO, Point3;

(* Networking. *)
IMPORT NetObj;

(* Custom network interface. *)
IMPORT Display;

REVEAL	T=Public BRANDED OBJECT
			scene : RootGO.T;
			display : GraphicsBase.T;
			count : CARDINAL;
			mode : ModeSet;
		OVERRIDES
			init := Init;
			awaitDelete := AwaitDelete;
		END;

	(* Object that client docks with. *)
TYPE	NetServer=Display.T OBJECT
			window : T;
		OVERRIDES
			init := GetHandle;
		END;

	(* Object that client then uses for communicating. *)
	NetDisplay=Display.Conn OBJECT
			group : GroupGO.T;
			window : T;
		OVERRIDES
			join := AddEdge;
			clear := Clear;
			close := Close;
		END;

(* Procedures to implement the network interface. *)

(* Procedure to return a network object interfacing a display window. *)
PROCEDURE GetHandle(self : NetServer) : Display.Conn=
VAR handle : NetDisplay;
BEGIN

	(* Create a new display handle and return it. *)
	handle := NEW(NetDisplay, window := self.window);
	handle.group := NEW(GroupGO.T).init();
	self.window.scene.add(handle.group);
	RETURN handle;	

END GetHandle;

(* Procedure to add an edge to the graph. *)
PROCEDURE AddEdge(self : NetDisplay; pt1, pt2 : Point3.T; colour : ColorProp.Base)
							RAISES {Display.NotOpen}=
VAR edge : GO.T;
BEGIN

	IF self.group = NIL THEN
		RAISE Display.NotOpen;
	END;

	(* Check the type of edge to add: *)
	IF Mode.Solid IN self.window.mode THEN
		edge := CylinderGO.New(pt1, pt2, 0.04, 3);
			(* point1, point2, width of cylinder, precision of cylinder. *)

		IF Mode.Colour IN self.window.mode THEN
			SurfaceGO.SetColour(edge, colour);
		END;
	ELSE
		edge := LineGO.New(pt1, pt2);

		IF Mode.Colour IN self.window.mode THEN
			LineGO.SetColour(edge, colour);
		END;
	END;

	(* Add it into the graph. *)
	self.group.add(edge);

END AddEdge;

(* Procedure to clear all the edges of the graph: *)
PROCEDURE Clear(self : NetDisplay) RAISES {Display.NotOpen}=
BEGIN

	IF self.group = NIL THEN
		RAISE Display.NotOpen;
	END;
	self.group.flush();

END Clear;

(* Procedure to remove the entire group from the display. *)
PROCEDURE Close(self : NetDisplay) RAISES {Display.NotOpen}=
BEGIN

	TRY
		self.window.scene.remove(self.group);
		self.group := NIL;
	EXCEPT
		GroupGO.BadElement=>
		RAISE Display.NotOpen;
	END;

END Close;

(* Procedure to initialise the window. *)
PROCEDURE Init(self : T; name : TEXT; mode : ModeSet) : T 
				RAISES {Failed, NetFailed}=

	(* ** THIS PROCEDURE LIFTED FROM RootGO IN anim3d ** *)
  PROCEDURE NewBase (title: TEXT) : GraphicsBase.T 
      RAISES {GraphicsBase.Failure} =
    TYPE 
      Pref = {None, X_PEX, X_OpenGL};
    VAR
      pref := Pref.None;
    BEGIN

      (*** Check for environment variable $prefPref ***)
      WITH var = Env.Get ("prefBase") DO
        IF var # NIL THEN
          IF Text.Equal (var, "X_PEX") THEN
            pref := Pref.X_PEX;
          ELSIF Text.Equal (var, "X_OpenGL") THEN
            pref := Pref.X_OpenGL;
          END;
        END;
      END;

      (* I don't bother checking the command line for preferences. *)

      (* Try to create the preferred base. 
         If this does not succeed, create any base. *)
      TRY 
        CASE pref OF
        | Pref.None     => 
        | Pref.X_PEX    => RETURN NEW (X_PEX_Base.T).init (title);
        | Pref.X_OpenGL => RETURN NEW (X_OpenGL_Base.T).init (title);
        END;
      EXCEPT
        GraphicsBase.Failure =>
      END;

      TRY 
        RETURN NEW (X_PEX_Base.T).init (title);
      EXCEPT
        GraphicsBase.Failure =>
      END;
      TRY 
        RETURN NEW (X_OpenGL_Base.T).init (title);
      EXCEPT
        GraphicsBase.Failure =>
      END;
      TRY 
        RETURN NEW (Win_OpenGL_Base.T).init (title);
      EXCEPT
        GraphicsBase.Failure =>
      END;
      
      RAISE GraphicsBase.Failure;
    END NewBase;

VAR 	server := NEW(NetServer);
BEGIN

	TRY

		(* Open up a window to display into. *)
		self.display := NewBase("Armadillo [" & name & "]");
		self.mode := mode;

		(* Create an empty scene and add it to the display. *)
		self.scene := RootGO.NewStd(self.display);

	EXCEPT
		GraphicsBase.Failure=>
			(* Failed because of graphics. *)
			RAISE Failed;
	END;

	(* Create a server object and export it. *)
	server.window := self;
	TRY
		NetObj.Export(Display.DisplayNameHdr & name, server);
	EXCEPT
		NetObj.Error, Thread.Alerted=>
			(* Couldn't export the object, so delete it and raise
				NetFailed for kicks! *)
			self.display.destroy();
			RAISE NetFailed;
	END;

	(* Return this display object. *)
	RETURN self;

END Init;

PROCEDURE AwaitDelete(self : T)=
BEGIN

	(* Wait for the 3D display to be killed. *)
	self.display.awaitDelete();

END AwaitDelete;

BEGIN

END Window.