(* Debug module.
	Provides implementations for the "print" and "assert"
	command objects.
*)

(*
	13/12/96	Created.
	14/01/97	Added the Pause command.
	15/01/97	Renamed module so as not to clash with the
			Cambridge Module-3 Debug module.
	25/02/97	Added implementation of the Collect object.
*)

MODULE ArmDebug;

IMPORT IO, Thread, Rd, Stdio, RTCollector;

IMPORT Code, Tags;

IMPORT Node, Loader, Except, ArmRT;

(* Object for the print command. *)
REVEAL	Print = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			execute := Print_Execute;
			print := Print_Print;
		END;

PROCEDURE Print_Execute(self : Print; env : Node.E) RAISES {Node.Abort}=
VAR	pos := self.expr;
	text := "";
BEGIN

	WHILE pos # NIL DO

		text := text & pos.get_expr(env).get_value(env).printval(env);
		pos := pos.next;

	END;

	IO.Put(text);

END Print_Execute;

PROCEDURE Print_Print(self : Print; i : TEXT) : TEXT=
BEGIN
	RETURN "print " & self.expr.print(i);
END Print_Print;

(* Object to deal with assertions. *)
REVEAL	Assert = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			execute := Assert_Execute;
			print := Assert_Print;
		END;

PROCEDURE Assert_Execute(self : Assert; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* This command should produce debug information. *)
	IF NOT(self.expr.get_value(env).get_bool(env)) THEN
		IO.Put(self.print("\t") & " : ASSERT FAILED!\n");
		RAISE Node.Abort(env.set_raised(Except.Unsupported, self.print("")));
	END;

END Assert_Execute;

PROCEDURE Assert_Print(self : Assert; i : TEXT) : TEXT=
BEGIN
	RETURN "assert " & self.expr.print(i);
END Assert_Print;

(* Object for the pause command. *)
REVEAL	Pause = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			execute := Pause_Execute;
			print := Pause_Print;
		END;

PROCEDURE Pause_Execute(self : Pause; env : Node.E) RAISES {Node.Abort}=
VAR	duration : LONGREAL;
BEGIN

	(* See how long we're to wait for. *)
	duration := self.expr.get_value(env).get_real(env);

	IF duration <= 0.0d0 THEN
		(* Wait for a key-press.
			** This code is crap. ** *)
		VAR	char : CHAR;
		BEGIN
			TRY
				char := Rd.GetChar(Stdio.stdin);
			EXCEPT
				Rd.EndOfFile, Rd.Failure, Thread.Alerted=>
					RAISE Node.Abort(env.set_raised(Except.IOError, NIL));
			END;
		END;
	ELSE
		(* Wait for that length of time. *)
		Thread.Pause(duration);
	END;

END Pause_Execute;

PROCEDURE Pause_Print(self : Pause; i : TEXT) : TEXT=
BEGIN
	RETURN "pause " & self.expr.print(i);
END Pause_Print;

REVEAL	Execute = Node.T BRANDED OBJECT
			program : Node.T;
		OVERRIDES
			execute := Execute_Execute;
			print := Execute_Print;
		END;

(* Procedure to run a pre-compiled program from the disk. *)
PROCEDURE Execute_Execute(self : Execute; env : Node.E) RAISES {Node.Abort}=
VAR	compiledname, progname : TEXT;	(* Compiled code file name. *)
	program : Node.T;	(* Compiled program code. *)
BEGIN

	(* Get the file name. *)
	progname := self.program.get_value(env).get_string(env);

	(* Work out the compiled name. *)
	compiledname := progname & Tags.CodeExt;

	(* Try to load and run the source file. *)
	TRY

		program := Loader.Go(compiledname, FALSE);
		IF program # NIL THEN
			(* Run the program. *)
			ArmRT.Execute(program, env);
		END;

	EXCEPT
		Node.Abort(t)=>
			t := t.unraise(Except.Quit);
			IF t # NIL THEN
				RAISE Node.Abort(t);
			END;
	|	Loader.FileFail, Loader.BadVersion=>
			RAISE Node.Abort(
				env.set_raised(Except.IOError, "while loading :" & progname)
				);
	|	Loader.Failure=>
			RAISE Node.Abort(
				env.set_raised(Except.Unsupported, "while loading :" & progname)
				);
	END;

END Execute_Execute;

PROCEDURE Execute_Print(self : Execute; i : TEXT) : TEXT=
BEGIN
	RETURN "execute " & self.program.print(i & "\t");
END Execute_Print;

REVEAL	Collect = Node.T BRANDED OBJECT
		OVERRIDES
			execute := Collect_Execute;
			print := Collect_Print;
		END;

(* Procedure to force the garbage collector to do the business. *)
PROCEDURE Collect_Execute(<* UNUSED *> self : Collect; <* UNUSED *> env : Node.E)=
BEGIN
	RTCollector.Collect();
END Collect_Execute;

PROCEDURE Collect_Print(<* UNUSED *> self : Collect; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "collect";
END Collect_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.BUG_Print=>
			RETURN NEW(Print, expr := Loader.Convert(tree.params.param));
	|	Tags.T.BUG_Assert=>
			RETURN NEW(Assert, expr := Loader.Convert(tree.params.param));
	|	Tags.T.BUG_Pause=>
			RETURN NEW(Pause, expr := Loader.Convert(tree.params.param));
	|	Tags.T.CMD_Execute=>
			RETURN NEW(Execute, program := Loader.Convert(tree.params.param));
	|	Tags.T.CMD_Collect=>
			RETURN NEW(Collect);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END ArmDebug.