(* ArmRT module. *)

MODULE ArmRT;

IMPORT Fmt;

(* Interpreter modules. *)
IMPORT Node, Except, Loader;

IMPORT Code, Tags;

(* Lock for the run-time system. *)
VAR	RunTimeLock := NEW(MUTEX);

(* This procedure simpy handles the sequential execution of commands. *)
PROCEDURE Execute(self : Node.T; env : Node.E) RAISES {Node.Abort}=
BEGIN

	WHILE self # NIL DO

		(* Execute the command. *)
		self.execute(env);

		(* Now get the next command. *)
		self := self.next;

	END;

END Execute;

(* Run-time limiting procedures. *)

VAR	ThreadCount := 1;
	FunCount := 0;
	MaxThreadCount := 160;
	MaxFunCount := 2000;

PROCEDURE IncThreadCount(env : Node.E) RAISES {Node.Abort}=
BEGIN

	LOCK RunTimeLock DO

		(* Check the ThreadCount. *)
		IF ThreadCount < MaxThreadCount THEN
			INC(ThreadCount);
		ELSE
			RAISE Node.Abort(
				env.set_raised(Except.TooManyThreads, Fmt.Int(ThreadCount))
				);
		END;

	END;

END IncThreadCount;

PROCEDURE DecThreadCount()=
BEGIN

	LOCK RunTimeLock DO
		DEC(ThreadCount);
	END;

END DecThreadCount;

PROCEDURE IncFunCount(env : Node.E) RAISES {Node.Abort}=
BEGIN

	LOCK RunTimeLock DO

		(* Check the function count. *)
		IF FunCount < MaxFunCount THEN
			INC(FunCount);
		ELSE
			RAISE Node.Abort(
				env.set_raised(Except.TooManyFunctions, Fmt.Int(FunCount))
				);
		END;

	END;

END IncFunCount;

PROCEDURE DecFunCount()=
BEGIN

	LOCK RunTimeLock DO
		DEC(FunCount);
	END;

END DecFunCount;

(* OBJECTS for the run-time commands. *)

(* Thread limiter. *)
REVEAL MaxThread = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			execute := MaxThread_Execute;
			print := MaxThread_Print;
		END;

PROCEDURE MaxThread_Execute(self : MaxThread; env : Node.E) RAISES {Node.Abort}=
VAR	newmax : INTEGER;
BEGIN

	(* Get the new count. *)
	newmax := self.expr.get_value(env).get_integer(env);

	(* Check it... *)
	IF newmax < ThreadCount THEN
		RAISE Node.Abort(env.set_raised(Except.TooManyThreads, NIL));
	END;

	MaxThreadCount := newmax;

END MaxThread_Execute;

PROCEDURE MaxThread_Print(self : MaxThread; i : TEXT) : TEXT=
BEGIN
	RETURN "maxthread " & self.expr.print(i & "\t");
END MaxThread_Print;

(* Recursion limiter. *)
REVEAL MaxFunction = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			execute := MaxFunction_Execute;
			print := MaxFunction_Print;
		END;

PROCEDURE MaxFunction_Execute(self : MaxFunction; env : Node.E) RAISES {Node.Abort}=
VAR	newmax : INTEGER;
BEGIN

	(* Get the new count. *)
	newmax := self.expr.get_value(env).get_integer(env);

	(* Check it... *)
	IF newmax < FunCount THEN
		RAISE Node.Abort(env.set_raised(Except.TooManyFunctions, NIL));
	END;

	MaxFunCount := newmax;

END MaxFunction_Execute;

PROCEDURE MaxFunction_Print(self : MaxFunction; i : TEXT) : TEXT=
BEGIN
	RETURN "maxfunction " & self.expr.print(i & "\t");
END MaxFunction_Print;

(* Loader for the command objects described above. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.CMD_MaxThread=>
			RETURN NEW(MaxThread, expr := Loader.Convert(tree.params.param));
	|	Tags.T.CMD_MaxFunction=>
			RETURN NEW(MaxFunction, expr := Loader.Convert(tree.params.param));
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END ArmRT.