(* BoolOp module.
	Implements objects for basic expression
	constructs.
*)

(*
	15/12/96	Created.
	13/01/97	MAJOR RE-WRITE!  SEE NODE MODULE.
*)

MODULE BoolOp;

IMPORT Node, Loader;
IMPORT Bool;

IMPORT Code, Tags;

(* Implication object. *)
REVEAL	Implies = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Implies_GetValue;
			print := Implies_Print;
		END;

PROCEDURE Implies_GetValue(self : Implies; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		NOT(self.left.get_value(env).get_bool(env)) OR
		self.right.get_value(env).get_bool(env)
		);
END Implies_GetValue;

PROCEDURE Implies_Print(self : Implies; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " => " & self.right.print(i);
END Implies_Print;

(* More than or equals to object. *)
REVEAL	MoreEqual = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := MoreEqual_GetValue;
			print := MoreEqual_Print;
		END;

PROCEDURE MoreEqual_GetValue(self : MoreEqual; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		self.left.get_value(env).compare(env, self.right.get_value(env)) # -1
		);
END MoreEqual_GetValue;

PROCEDURE MoreEqual_Print(self : MoreEqual; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " >= " & self.right.print(i);
END MoreEqual_Print;

(* Less than or equal to object. *)
REVEAL	LessEqual = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := LessEqual_GetValue;
			print := LessEqual_Print;
		END;

PROCEDURE LessEqual_GetValue(self : LessEqual; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	left_val : Node.V;
	type : Node.TYP;
BEGIN

	(* Evaluate the left operand. *)
	left_val := self.left.get_value(env);
	type := left_val.get_type().resolve(env);

	IF ISTYPE(type, Bool.BoolT) THEN
		RETURN Bool.NewBool(
			left_val.get_bool(env) OR
			NOT(self.right.get_value(env).get_bool(env))
			);
	ELSE
		RETURN Bool.NewBool(
			self.left.get_value(env).compare(env,
					self.right.get_value(env)
					) # 1
			);
	END;

END LessEqual_GetValue;

PROCEDURE LessEqual_Print(self : LessEqual; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " <= " & self.right.print(i);
END LessEqual_Print;

(* Less than object. *)
REVEAL	Less = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Less_GetValue;
			print := Less_Print;
		END;

PROCEDURE Less_GetValue(self : Less; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		self.left.get_value(env).compare(env, self.right.get_value(env)) = -1
		);
END Less_GetValue;

PROCEDURE Less_Print(self : Less; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " < " & self.right.print(i);
END Less_Print;

(* More than object. *)
REVEAL	More = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := More_GetValue;
			print := More_Print;
		END;

PROCEDURE More_GetValue(self : More; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		self.left.get_value(env).compare(env, self.right.get_value(env)) = 1
		);
END More_GetValue;

PROCEDURE More_Print(self : More; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " > " & self.right.print(i);
END More_Print;

(* Equals to object. *)
REVEAL	Equal = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Equal_GetValue;
			print := Equal_Print;
		END;

PROCEDURE Equal_GetValue(self : Equal; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		self.left.get_value(env).equal(env, self.right.get_value(env))
		);
END Equal_GetValue;

PROCEDURE Equal_Print(self : Equal; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " = " & self.right.print(i);
END Equal_Print;

(* Not equals to object. *)
REVEAL	NotEqual = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := NotEqual_GetValue;
			print := NotEqual_Print;
		END;

PROCEDURE NotEqual_GetValue(self : NotEqual; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		NOT(self.left.get_value(env).equal(env, self.right.get_value(env)))
		);
END NotEqual_GetValue;

PROCEDURE NotEqual_Print(self : NotEqual; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " # " & self.right.print(i);
END NotEqual_Print;

(* Routine to convert Code trees into the above. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.BOOL_Implies=>
			RETURN NEW(Implies,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_MoreOrEq=>
			RETURN NEW(MoreEqual,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_LessOrEq=>
			RETURN NEW(LessEqual,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_MoreThan=>
			RETURN NEW(More,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_LessThan=>
			RETURN NEW(Less,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_Equal=>
			RETURN NEW(Equal,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_NotEqual=>
			RETURN NEW(NotEqual,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END BoolOp.