(* Concurrency module.
	Implements the mutex datatype used for locking critical
	sections of code and exports the concurrency control commands.
*)

(*
	11/01/97	Created.
	13/01/97	MAJOR RE-WRITE. SEE NODE MODULE.
	04/02/97	Added the Send and Recieve commands used by the
			port system.
*)

MODULE Concurrency;

(* IMPORT IO; *)

IMPORT Thread;

IMPORT Code, Tags;

IMPORT Node, Loader, ArmRT, Except, TypeSys;

(* The mutex data-type. *)
REVEAL	Mutex = Node.K BRANDED OBJECT
			(* Signal sent when unlocked by a thread. *)
			unlocked : Thread.Condition;
			(* ID of locking thread. *)
			held_by : Node.E := NIL;
		OVERRIDES
			lock := Mutex_Lock;
			unlock := Mutex_UnLock;
			wait_unlocked := Mutex_WaitUnLocked;
			print := Mutex_Print;
		END;

PROCEDURE Mutex_Lock(self : Mutex; env : Node.E) : BOOLEAN RAISES {Node.Abort}=
BEGIN

	IF self.held_by = env.get_thread_base() THEN

		(* The mutex is already locked by this thread. *)
		RAISE Node.Abort(env.set_raised(Except.AlreadyLocked, NIL));

	ELSIF self.held_by = NIL THEN

		(* Give the mutex to this thread. *)
		self.held_by := env.get_thread_base();
		RETURN TRUE;

	ELSE

		(* Another thread has already locked it. *)
		RETURN FALSE;

	END;

END Mutex_Lock;	

PROCEDURE Mutex_UnLock(self : Mutex; env : Node.E) RAISES {Node.Abort}=
BEGIN

	IF self.held_by = env.get_thread_base() THEN

		(* Clear the lock. *)
		self.held_by := NIL;

		(* Signal any waiting threads. *)
		Thread.Signal(self.unlocked);

	ELSE

		(* The lock does not belong to this thread. *)
		RAISE Node.Abort(env.set_raised(Except.UnlockedVar, NIL));

	END;

END Mutex_UnLock;

PROCEDURE Mutex_WaitUnLocked(self : Mutex; <* UNUSED *> env : Node.E; lock : Thread.Mutex)=
BEGIN

	(* Wait on the unlocked signal of the var. *)
	Thread.Wait(lock, self.unlocked);

END Mutex_WaitUnLocked;

PROCEDURE Mutex_Print(<* UNUSED *> self : Mutex; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "mutex";
END Mutex_Print;

(* Mutex datatype allocator object. *)
REVEAL	MutexT = Node.TYP BRANDED OBJECT
		OVERRIDES
			allocate := MutexT_Allocate;
			check_type := MutexT_CheckType;
			print := MutexT_Print;
		END;

PROCEDURE MutexT_Allocate(self : MutexT; <* UNUSED *> env : Node.E) : Node.K=
BEGIN
	RETURN NEW(Mutex,
			unlocked := NEW(Thread.Condition),
			held_by := NIL
		);
END MutexT_Allocate;

PROCEDURE MutexT_CheckType(<* UNUSED *> s : MutexT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Mutexes cannot be recursive. *)
	RETURN TRUE;
END MutexT_CheckType;

PROCEDURE MutexT_Print(<* UNUSED *> self : MutexT; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "mutex";
END MutexT_Print;

(* The signal command. *)
REVEAL	Signal = Node.T BRANDED OBJECT
			var : Node.T;
		OVERRIDES
			execute := Signal_Execute;
			print := Signal_Print;
		END;

PROCEDURE Signal_Execute(self : Signal; env : Node.E) RAISES {Node.Abort}=
BEGIN
	self.var.get_value(env).signal(env);
END Signal_Execute;

PROCEDURE Signal_Print(self : Signal; i : TEXT) : TEXT=
BEGIN
	RETURN "signal " & self.var.print(i & "\t");
END Signal_Print;

(* The wait command. *)
REVEAL	Wait = Node.T BRANDED OBJECT
			var : Node.T;
			with : Node.T := NIL;
		OVERRIDES
			execute := Wait_Execute;
			print := Wait_Print;
		END;

PROCEDURE Wait_Execute(self : Wait; env : Node.E) RAISES {Node.Abort}=
VAR	varnames : Node.T;
	varlist : Node.VarList;
BEGIN

	(* Get the global locking-control mutex. *)
	LOCK GlobalLock DO

		TRY

			(* Now work through the list of "with" variables, unlocking them. *)
			(* At the same time, a VarList is generated, to be used when re-locking
				all the unlocked variables. *)
			varnames := self.with;
			varlist := NIL;
			WHILE varnames # NIL DO

				(* Try to unlock the variable. *)
				varnames.get_value(env).unlock(env);

				(* Add an entry to the varlist. *)
				varlist := NEW(Node.VarList).init(varnames.get_name(env), NIL).setnext(varlist);

				varnames := varnames.next;

			END;

			(* The variables are unlocked, so now wait on the specified variable. *)
			self.var.get_value(env).wait(env, GlobalLock);

		FINALLY

			(* Re-Lock the unlocked variables. *)
			(* This should be done regardless of the success of the command. *)
			LockVars(varlist, env, GlobalLock);

		END;

	END;

END Wait_Execute;

PROCEDURE Wait_Print(self : Wait; i : TEXT) : TEXT=
BEGIN
	RETURN "wait " & self.var.print(i & "\t") & " with " & self.with.print(i & "\t");
END Wait_Print;

(* The send command. *)
REVEAL	Send = Node.T BRANDED OBJECT
			data : Node.T;
			port : Node.T;
			with : Node.T := NIL;
		OVERRIDES
			execute := Send_Execute;
			print := Send_Print;
		END;

PROCEDURE Send_Execute(self : Send; env : Node.E) RAISES {Node.Abort}=
VAR	varnames : Node.T;
	varlist : Node.VarList;
	data, port : Node.V;
BEGIN

	(* Get the global locking-control mutex. *)
	LOCK GlobalLock DO

		TRY

			(* Now work through the list of "with" variables, unlocking them. *)
			(* At the same time, a VarList is generated, to be used when re-locking
				all the unlocked variables. *)
			varnames := self.with;
			varlist := NIL;
			WHILE varnames # NIL DO

				(* Try to unlock the variable. *)
				varnames.get_value(env).unlock(env);

				(* Add an entry to the varlist. *)
				varlist := NEW(Node.VarList).init(varnames.get_name(env), NIL).setnext(varlist);

				varnames := varnames.next;

			END;

			(* The variables are unlocked, so now send the data down the specified port. *)
			data := self.data.get_value(env);
			port := self.port.get_value(env);
			port.send(env, data, GlobalLock);

		FINALLY

			(* Re-Lock the unlocked variables. *)
			LockVars(varlist, env, GlobalLock);

		END;

	END;

END Send_Execute;

PROCEDURE Send_Print(self : Send; i : TEXT) : TEXT=
BEGIN
	RETURN "send " & self.data.print(i & "\t") & " on " &
		self.port.print(i & "\t") & " with " & self.with.print(i & "\t");
END Send_Print;

(* The recieve command. *)
REVEAL	Recieve = Node.T BRANDED OBJECT
			port : Node.T;
			with : Node.T := NIL;
		OVERRIDES
			get_value := Recieve_GetValue;
			print := Recieve_Print;
		END;

PROCEDURE Recieve_GetValue(self : Recieve; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	varnames : Node.T;
	varlist : Node.VarList;
	value : Node.V;
BEGIN

	(* Get the global locking-control mutex. *)
	LOCK GlobalLock DO

		TRY

			(* Now work through the list of "with" variables, unlocking them. *)
			(* At the same time, a VarList is generated, to be used when re-locking
				all the unlocked variables. *)
			varnames := self.with;
			varlist := NIL;
			WHILE varnames # NIL DO

				(* Try to unlock the variable. *)
				varnames.get_value(env).unlock(env);

				(* Add an entry to the varlist. *)
				varlist := NEW(Node.VarList).init(varnames.get_name(env), NIL).setnext(varlist);

				varnames := varnames.next;

			END;

			(* The variables are unlocked, so now recieve from the specified port. *)
			value := self.port.get_value(env).recieve(env, GlobalLock);

		FINALLY

			(* Re-Lock the unlocked variables. *)
			LockVars(varlist, env, GlobalLock);

		END;

	END;

	RETURN value;

END Recieve_GetValue;

PROCEDURE Recieve_Print(self : Recieve; i : TEXT) : TEXT=
BEGIN
	RETURN "receive(" & self.port.print(i & "\t") & " with " & self.with.print(i & "\t") & ")";
END Recieve_Print;

(* The lock command object. *)
REVEAL	Lock = Node.T BRANDED OBJECT
			varnames : Node.T;
			code : Node.T;
		OVERRIDES
			execute := Lock_Execute;
			print := Lock_Print;
		END;

PROCEDURE Lock_Execute(self : Lock; env : Node.E) RAISES {Node.Abort}=
VAR	varlist : Node.VarList;
	pos := self.varnames;
BEGIN

	(* Make a varlist for the lock. *)
	WHILE pos # NIL DO
		varlist := NEW(Node.VarList).init(pos.get_name(env), NIL).setnext(varlist);
		pos := pos.next;
	END;

	(* Lock the listed variables. *)
	LOCK GlobalLock DO
		LockVars(varlist, env, GlobalLock);
	END;

	TRY
		(* Execute the given code. *)
		ArmRT.Execute(self.code, env);
	FINALLY
		(* Unlock the listed variables. *)
		LOCK GlobalLock DO
			WHILE varlist # NIL DO
				env.get_var(varlist.getname()).unlock(env);
				varlist := varlist.getnext();
			END;
		END;
	END;

END Lock_Execute;

PROCEDURE Lock_Print(self : Lock; i : TEXT) : TEXT=
BEGIN
	RETURN "lock " & self.varnames.print(i) & " in\n\t" & i & self.code.print(i & "\t") & "\n" & i & "end";
END Lock_Print;

(* Internal procedure for locking a list of variables. *)
PROCEDURE LockVars(VAR varnames : Node.VarList; env : Node.E; lock : Thread.Mutex) RAISES {Node.Abort}=
VAR	locked : Node.VarList := NIL;
	temp : Node.VarList;
	unlockable : Node.V;
BEGIN

	LOOP
		(* Work through the varnames list, locking the vars. *)
		WHILE varnames # NIL DO
			IF env.get_var(varnames.getname()).lock(env) THEN
				temp := varnames.getnext();
				locked := varnames.setnext(locked);
				varnames := temp;
			ELSE
				EXIT;
			END;
		END;

		IF varnames = NIL THEN
			(* All the given variables are now locked!! *)
			EXIT;
		END;

		(* One of the variables is locked by another thread. *)
		(* Store the variable so that it's unlocked signal can
			be waited upon. *)
		unlockable := env.get_var(varnames.getname());

		(* Now unlock all the variables we just locked... *)
		WHILE locked # NIL DO
			env.get_var(locked.getname()).unlock(env);
			temp := locked.getnext();
			varnames := locked.setnext(varnames);
			locked := temp;
		END;

		(* Now wait for the unlockable thingy to become lockable: *)
		unlockable.wait_unlocked(env, lock);

	END;

	(* On exit from this routine, leave the "varnames" list containing
		the list of locked variables. *)
	varnames := locked;

END LockVars;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Mutex=>
			RETURN MutexType;
	|	Tags.T.CMD_Signal=>
			RETURN NEW(Signal, var := Loader.Convert(tree.params.param));
	|	Tags.T.CMD_Wait=>
			token := NEW(Wait, var := Loader.Convert(tree.params.param));
			IF tree.params.next # NIL THEN
				NARROW(token, Wait).with := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.CMD_Send=>
			token := NEW(Send,
					data := Loader.Convert(tree.params.param),
					port := Loader.Convert(tree.params.next.param)
				);
			IF tree.params.next.next # NIL THEN
				NARROW(token, Send).with := Loader.Convert(tree.params.next.next.param);
			END;
			RETURN token;
	|	Tags.T.FUN_Recieve=>
			token := NEW(Recieve,
					port := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, Recieve).with := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.BLK_Lock=>
			RETURN NEW(Lock,
					varnames := Loader.Convert(tree.params.param),
					code := Loader.ConvertCommand(tree.params.next.param)
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

(* Routine to allocate a new mutex. *)
PROCEDURE NewMutex() : Node.V=
BEGIN
	RETURN NEW(Node.V).init(MutexType,
				NEW(Mutex,
					unlocked := NEW(Thread.Condition)
				)
			);
END NewMutex;

VAR	GlobalLock : MUTEX;
BEGIN
	MutexType := NEW(MutexT, code := TypeSys.NewTypeCode());
	GlobalLock := NEW(MUTEX);
END Concurrency.