(* Condition module.
	Implements the condition variable datatype used for signalling
	as well as.
*)

(*
	19/12/96	Created.
	10/01/97	Added the get_typecode method.
	11/01/97	Moved the Signal and Wait command objects into the
			Concurrency module for simplicity.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
*)

MODULE Condition;

IMPORT Thread;

IMPORT Code, Tags;

IMPORT Node, Loader, TypeSys, Except;

(* The condition variable data-type. *)
REVEAL	Condition = Node.K BRANDED OBJECT
			condition : Thread.Condition;
		OVERRIDES
			signal := Condition_Signal;
			wait := Condition_Wait;
			get_condition := Condition_GetCondition;
			set_value := Condition_SetValue;
			printval := Condition_PrintVal;
			equal := Condition_Equal;
		END;

PROCEDURE Condition_Signal(self : Condition; env : Node.E) RAISES {Node.Abort}=
BEGIN
	(* Signal the condition variable. *)
	Thread.Signal(self.condition);
END Condition_Signal;

PROCEDURE Condition_Wait(self : Condition; env : Node.E; lock : Thread.Mutex) RAISES {Node.Abort}=
BEGIN
	(* Wait on the condition variable. *)
	Thread.Wait(lock, self.condition);
END Condition_Wait;

PROCEDURE Condition_GetCondition(self : Condition; <* UNUSED *> env : Node.E) : Thread.Condition=
BEGIN
	RETURN self.condition;
END Condition_GetCondition;

PROCEDURE Condition_SetValue(self : Condition; env : Node.E; val : Node.K) RAISES {Node.Abort}=
BEGIN
	self.condition := val.get_condition(env);
END Condition_SetValue;

PROCEDURE Condition_PrintVal(self : Condition; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	RAISE Node.Abort(env.set_raised(Except.Unsupported, "condition.printval"));
END Condition_PrintVal;

PROCEDURE Condition_Equal(self : Condition; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN self.condition = val.get_condition(env);
END Condition_Equal;

(* Condition datatype allocator object. *)
REVEAL	ConditionT = Node.TYP BRANDED OBJECT
		OVERRIDES
			allocate := ConditionT_Allocate;
			check_type := ConditionT_CheckType;
			print := ConditionT_Print;
		END;

PROCEDURE ConditionT_Allocate(self : ConditionT; <* UNUSED *> env : Node.E) : Node.K=
BEGIN
	RETURN NEW(Condition, condition := NEW(Thread.Condition));
END ConditionT_Allocate;

PROCEDURE ConditionT_CheckType(<* UNUSED *> s : ConditionT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Condition variables cannot be recursive. *)
	RETURN TRUE;
END ConditionT_CheckType;

PROCEDURE ConditionT_Print(<* UNUSED *> self : ConditionT; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "condition";
END ConditionT_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Condition=>
			RETURN ConditionType;
	ELSE
		RETURN NIL;
	END;

END Convert;

(* Routine used to allocate new condition variables. *)
PROCEDURE NewCond() : Node.V=
BEGIN
	RETURN NEW(Node.V).init(ConditionType,
				NEW(Condition,
					condition := NEW(Thread.Condition)
				)
			);
END NewCond;

BEGIN
	ConditionType := NEW(ConditionT, code := TypeSys.NewTypeCode());
END Condition.