(* Control module.
	Implements the program flow-control objects.
*)

(*
	14/12/96	Created skeleton, implementing only the loop
			construct, to use when testing concurrency.
	17/12/96	Added Return object.
	20/12/96	Implemented the "elsif" language extension.
	19/02/97	Implemented the "quit" command.
*)

MODULE Control;

IMPORT Node, Loader, ArmRT, Except;

IMPORT Code, Tags;

(* The conditional execution object. *)
REVEAL	If = Node.T BRANDED OBJECT
			expr : Node.T;
			command1 : Node.T;
			command2 : Node.T := NIL;
			link : Node.T;
		OVERRIDES
			execute := If_Execute;
			print := If_Print;
		END;

PROCEDURE If_Execute(self : If; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* Check the expression. *)
	IF self.expr.get_value(env).get_bool(env) THEN

		(* Run the first command. *)
		ArmRT.Execute(self.command1, env);

	ELSE

		(* NEW Check for Elsif. *)
		IF self.link # NIL THEN
			self.link.execute(env);
		END;

		(* Check that we have a second command. *)
		IF self.command2 # NIL THEN
			(* Run it. *)
			ArmRT.Execute(self.command2, env);
		END;

	END;

END If_Execute;

PROCEDURE If_Print(self : If; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := "if " & self.expr.print(i) & " then\n" & i & "\t";
	text := text & Node.ToText(self.command1, i & "\t") & "\n" & i;
	IF self.command2 # NIL THEN
		text := text & "else\n\t" & i & Node.ToText(self.command2, i & "\t") & "\n" & i;
	END;
	text := text & "end";
	RETURN text;
END If_Print;

(* The loop object.  The only way to exit from a loop is by exception. *)
REVEAL	Loop = Node.T BRANDED OBJECT
			command : Node.T;
		OVERRIDES
			execute := Loop_Execute;
			print := Loop_Print;
		END;

PROCEDURE Loop_Execute(self : Loop; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* Continually execute the given code until an exception is
		raised. *)

	TRY
		LOOP
			ArmRT.Execute(self.command, env);
		END;
	EXCEPT
		Node.Abort(t)=>
			(* The "exit" exception is not propagated by this command. *)
			t := t.unraise(Except.Exit);
			IF t # NIL THEN
				RAISE Node.Abort(t);
			END;
	END;

END Loop_Execute;

PROCEDURE Loop_Print(self : Loop; i : TEXT) : TEXT=
BEGIN
	RETURN "loop\n" & i & "\t" & Node.ToText(self.command, i & "\t") & "\n" & i & "end";
END Loop_Print;

(* The exit command. Just raises the "exit" exception. *)
REVEAL	Exit = Node.T BRANDED OBJECT
		OVERRIDES
			execute := Exit_Execute;
			print := Exit_Print;
		END;

PROCEDURE Exit_Execute(<* UNUSED *> self : Exit; env : Node.E) RAISES {Node.Abort}=
BEGIN
	RAISE Node.Abort(env.set_raised(Except.Exit, NIL));
END Exit_Execute;

PROCEDURE Exit_Print(<* UNUSED *> self : Exit; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "exit";
END Exit_Print;

(* The return command. Raises the "return" exception and stores a value. *)
REVEAL	Return = Node.T BRANDED OBJECT
			expr : Node.T := NIL;
		OVERRIDES
			execute := Return_Execute;
			print := Return_Print;
		END;

PROCEDURE Return_Execute(self : Return; env : Node.E) RAISES {Node.Abort}=
BEGIN
	(* Firstly, store the return value if there is one. *)
	IF self.expr # NIL THEN
		env.set_return(self.expr.get_value(env));
	END;

	(* Now raise the return exception and bomb. *)
	RAISE Node.Abort(env.set_raised(Except.Return, NIL));
END Return_Execute;

PROCEDURE Return_Print(self : Return; i : TEXT) : TEXT=
VAR 	text : TEXT;
BEGIN
	text := "return";
	IF self.expr # NIL THEN
		text := text & " " & self.expr.print(i);
	END;
	RETURN text;
END Return_Print;

(* The quit command. Just raises the "quit" exception. *)
REVEAL	Quit = Node.T BRANDED OBJECT
		OVERRIDES
			execute := Quit_Execute;
			print := Quit_Print;
		END;

PROCEDURE Quit_Execute(<* UNUSED *> self : Quit; env : Node.E) RAISES {Node.Abort}=
BEGIN
	RAISE Node.Abort(env.set_raised(Except.Quit, NIL));
END Quit_Execute;

PROCEDURE Quit_Print(<* UNUSED *> self : Quit; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "quit";
END Quit_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.CMD_If=>
			token := NEW(If,
				expr := Loader.Convert(tree.params.param),
				command1 := Loader.ConvertCommand(tree.params.next.param),
				link := Loader.ConvertCommand(tree.params.next.next.param)
				);
			IF tree.params.next.next.next # NIL THEN
				NARROW(token, If).command2 :=
					Loader.ConvertCommand(tree.params.next.next.next.param);
			END;
			RETURN token;

		(* Elsif is implemented as a nested If. *)
	|	Tags.T.CMD_ElsIf=>
			RETURN NEW(If,
				expr := Loader.Convert(tree.params.param),
				command1 := Loader.ConvertCommand(tree.params.next.param),
				link := Loader.ConvertCommand(tree.params.next.next.param)
				);
	|	Tags.T.BLK_Loop=>
			RETURN NEW(Loop, command := Loader.ConvertCommand(tree.params.param));
	|	Tags.T.CMD_Exit=>
			RETURN NEW(Exit);
	|	Tags.T.CMD_Return=>
			token := NEW(Return);
			IF tree.params # NIL THEN
				NARROW(token, Return).expr := Loader.Convert(tree.params.param);
			END;
			RETURN token;
	|	Tags.T.CMD_Quit=>
			RETURN NEW(Quit);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Control.