(* Core module.
	Reveals the details of the core program node
	objects and gives implementations for their
	methods, as well as routines for loading them
	in from disc.
*)

(*
	11/12/96	Created.
	13/12/96	Implemented the parallel execution constructor.
	20/12/96	Modified the parallel execution system to cope
			with "return"ing from threads.
	16/02/97	Fixed raised-exception list handling.
*)

MODULE Core;

(* IMPORT IO; *)

(* System modules. *)
IMPORT Thread;

(* Armacode library modules. *)
IMPORT Code, Tags;

(* Interpreter internal modules. *)
IMPORT Loader, ArmRT;
IMPORT Node, Except;

(* Parallel execution node. *)
REVEAL	Parallel = Node.T BRANDED OBJECT
			command1, command2 : Node.T;
		OVERRIDES
			execute := Parallel_Execute;
			print := Parallel_Print;
		END;

(* Special parallel execution closure, used internally. *)
TYPE	Parallel_Closure = Thread.Closure OBJECT
			code : Node.T;
			env : Node.E;
		OVERRIDES
			apply := Parallel_apply;
		END;

(* The apply procedure runs the specified code in the environment given. *)
PROCEDURE Parallel_apply(self : Parallel_Closure) : REFANY=
BEGIN

	TRY
		(* Execute the command. *)
		ArmRT.Execute(self.code, self.env);
	EXCEPT
		(* Any exceptions just terminate the thread. *)
		Node.Abort(t)=>
			(* Remove the return exception, if present. *)
			RETURN t.unraise(Except.Return);
	END;

	RETURN NIL;

END Parallel_apply;

(* Implementations for the parallel execution object. *)
PROCEDURE Parallel_Execute(self : Parallel; env : Node.E) RAISES {Node.Abort}=
VAR	env1 : Node.E;
	env2 : Node.E;
	closure : Parallel_Closure;
	thread : Thread.T;
	excepts1, excepts2 : Node.ExceptList := NIL;
BEGIN

	(* Routine creates two empty stack frames and spawns a new thread.
		The new thread executes one thread of commands, the original
		thread does the other set of commands.
		The new stack frame will have no turtle selected.  It will
		be capable of creating a special "temporary" turtle if required,
		which inherit's its properties from the parent stack frame.
	*)

	(* Increment and check the thread count. *)
	ArmRT.IncThreadCount(env);

	(* Create a pair of "base" environments (one for each thread). *)
	env1 := env.make_base();
	env1.add_return(NIL);
	env1.make_thread_base();
	env2 := env.make_base();
	env2.add_return(NIL);
	env2.make_thread_base();

	(* Make the closure for the second thread. *)
	closure := NEW(Parallel_Closure, code := self.command2, env := env2);

	(* Start the second thread going. *)
	thread := Thread.Fork(closure);

	TRY
		TRY
			(* Execute the first command now. *)
			ArmRT.Execute(self.command1, env1);
		FINALLY
			(* Once this thread terminates, join with the others. *)
			excepts2 := Thread.Join(thread);

			(* Decrement the thread count. *)
			ArmRT.DecThreadCount();
		END;
	EXCEPT
		(* If any exceptions occur then terminate. *)
		Node.Abort(t)=>
			(* The "return" exception is used to exit from
				a thread prematurely.  If a return value is specified
				then since no space has been allocated to store it in
				the stack, a "type mismatch" exception will occur. *)
			excepts1 := t.unraise(Except.Return);
	END;

	(* Resolve exceptions. This is done by forming the union of the two
		raised exception lists. *)
	IF excepts1 # NIL THEN
		excepts1 := excepts1.union(excepts2);
	ELSE
		excepts1 := excepts2;
	END;

	(* If exceptions were raised then raise Abort. *)
	IF excepts1 # NIL THEN
		RAISE Node.Abort(excepts1);
	END;

END Parallel_Execute;

PROCEDURE Parallel_Print(self : Parallel; i : TEXT) : TEXT=
BEGIN
	RETURN self.command1.print(i) & " &\n" & i & self.command2.print(i);
END Parallel_Print;

(* Bracketting node. *)
REVEAL	Bracket = Node.T BRANDED OBJECT
			command : Node.T;
		OVERRIDES
			execute := Bracket_Execute;
			print := Bracket_Print;
		END;

(* Bracket method implementations. *)
PROCEDURE Bracket_Execute(self : Bracket; env : Node.E) RAISES {Node.Abort}=
BEGIN
	ArmRT.Execute(self.command, env);
END Bracket_Execute;

PROCEDURE Bracket_Print(self : Bracket; i : TEXT) : TEXT=
BEGIN
	RETURN "[\n" & i & "\t" & Node.ToText(self.command, i & "\t") & "\n" & i & "]";
END Bracket_Print;

(* Routine for converting a saved tree into objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.CMD_Par=>
			RETURN NEW(Parallel,
				command1 := Loader.ConvertCommand(tree.params.param),
				command2 := Loader.ConvertCommand(tree.params.next.param)
				);
	|	Tags.T.TAG_BracketCmd=>
			RETURN NEW(Bracket,
				command := Loader.ConvertCommand(tree.params.param)
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Core.