(* Exceptions module.
	Implements the exception handling commands of
	the armadillo language.

	16/01/97	Added the "with" functionality to the "raise" command.
	27/02/97	Modified the code for "exception" command to allow it
			to be called in a non-block form.
	20/03/97	Added the Try...Finally command to the interpreter.

*)

MODULE Exceptions;

(* IMPORT IO; *)

IMPORT Node, Loader, ArmRT, Except;

IMPORT Code, Tags;

(* The exception name type. *)
REVEAL	Exception = Node.T BRANDED OBJECT
			name : TEXT;
		OVERRIDES
			get_name := Exception_GetName;
			print := Exception_Print;
		END;

PROCEDURE Exception_GetName(self : Exception; <* UNUSED *> env : Node.E) : TEXT=
BEGIN
	RETURN self.name;
END Exception_GetName;

PROCEDURE Exception_Print(self : Exception; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := self.name;
	IF self.next # NIL THEN
		text := text & ", " & self.next.print(i);
	END;
	RETURN text;
END Exception_Print;

(* The exception-defining block. *)
REVEAL	ExceptionCMD = Node.T BRANDED OBJECT
			names : Node.T;
			code : Node.T;
			is_block := FALSE;
		OVERRIDES
			execute := ExceptionCMD_Execute;
			print := ExceptionCMD_Print;
		END;

PROCEDURE ExceptionCMD_Execute(self : ExceptionCMD; env : Node.E) RAISES {Node.Abort}=
VAR	excepts, this_except, new_except : Node.ExceptList;
	pos : Node.T;
	new_env : Node.E;
	infotext : TEXT;
BEGIN

	(* Construct the ExceptList for the exceptions named. *)
	pos := self.names;
	WHILE pos # NIL DO

		excepts := NEW(Node.ExceptList).init(pos.get_name(env)).setnext(excepts);
		pos := pos.next;

	END;

	IF self.is_block THEN
		(* Create a new stack frame and add the exceptions. *)
		new_env := env.make_child().add_exceptions(excepts);
	ELSE
		(* Add the exceptions to this stack frame. *)
		new_env := env.add_exceptions(excepts);
	END;

	TRY

		(* Run the command in this stack frame. *)
		IF self.is_block THEN
			ArmRT.Execute(self.code, new_env);
		END;

	EXCEPT

		(* If the code raises any exceptions then we must check
			that none are from the above list. *)
		Node.Abort(t)=>
			WHILE excepts # NIL DO

				(* Was it raised? *)
				this_except := t.find(excepts.getname());
				IF this_except # NIL THEN

					(* Create the info string. *)
					infotext := this_except.getname();
					IF this_except.getinfo() # NIL THEN
						infotext := infotext & "{" & this_except.getinfo() & "}";
					END;

					(* Unraise the exception. *)
					t := t.unraise(this_except.getname());

					(* Add in the untrapped exception. *)
					new_except := env.set_raised(
									Except.Untrapped,
									infotext
									);
					IF t = NIL THEN
						t := new_except;
					ELSE
						t := t.union(new_except);
					END;

				END;

				excepts := excepts.getnext();

			END;

			(* If exceptions were raised then bomb out. *)
			IF t # NIL THEN
				RAISE Node.Abort(t);
			END;

	END;

END ExceptionCMD_Execute;

PROCEDURE ExceptionCMD_Print(self : ExceptionCMD; i : TEXT) : TEXT=
BEGIN
	RETURN "exception " & self.names.print(i) & " in\n\t" & i & Node.ToText(self.code, i & "\t");
END ExceptionCMD_Print;

(* Object used to raise general exceptions. *)
REVEAL	Raise = Node.T BRANDED OBJECT
			names : Node.T;
			info : Node.T := NIL;
		OVERRIDES
			execute := Raise_Execute;
			print := Raise_Print;
		END;

PROCEDURE Raise_Execute(self : Raise; env : Node.E) RAISES {Node.Abort}=
VAR	pos : Node.T;
	infotext : TEXT := NIL;
	excepts : Node.ExceptList := NIL;
BEGIN

	(* Is there an information string? *)
	IF self.info # NIL THEN
		infotext := self.info.get_value(env).get_string(env);
	END;

	(* Run through the list of exceptions, setting them raised. *)
	pos := self.names;
	WHILE pos # NIL DO

		(* Raise the exception. *)
		TRY
			excepts := env.set_raised(pos.get_name(env), infotext).union(excepts);
		EXCEPT
			Node.Abort(t)=>
				(* The exception didn't exist, so instead return the error exception. *)
				excepts := t.union(excepts);
		END;

		pos := pos.next;

	END;

	(* Now cause execution to bomb out. *)
	RAISE Node.Abort(excepts);

END Raise_Execute;

PROCEDURE Raise_Print(self : Raise; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := "raise " & self.names.print(i);
	IF self.info # NIL THEN
		text := text & " with " & self.info.print(i & "\t");
	END;
	RETURN text;
END Raise_Print;

(* Try..Except exception trap block. *)
REVEAL	Try = Node.T BRANDED OBJECT
			code : Node.T;
			traps : Node.T;
		OVERRIDES
			execute := Try_Execute;
			print := Try_Print;
		END;

PROCEDURE Try_Execute(self : Try; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY

		(* Execute the given code. *)
		ArmRT.Execute(self.code, env);

	EXCEPT

		(* Trap any exceptions that occur. *)
		Node.Abort(t)=>
			t := self.traps.handle(env, t);
			(* If any exceptions are still raised then bomb out. *)
			IF t # NIL THEN
				RAISE Node.Abort(t);
			END;

	END;

END Try_Execute;

PROCEDURE Try_Print(self : Try; i : TEXT) : TEXT=
BEGIN
	RETURN "try\n\t" & i & Node.ToText(self.code, i & "\t") & "\n" & i &
		"except\n\t" & i & self.traps.print(i & "\t") & "\n" & i & "end"
END Try_Print;

(* Try..Except exception trap block. *)
REVEAL	TryFinally = Node.T BRANDED OBJECT
			main_code : Node.T;
			finally_code : Node.T;
		OVERRIDES
			execute := TryFinally_Execute;
			print := TryFinally_Print;
		END;

PROCEDURE TryFinally_Execute(self : TryFinally; env : Node.E) RAISES {Node.Abort}=
VAR	excepts : Node.ExceptList := NIL;
BEGIN

	TRY

		(* Execute the main code. *)
		ArmRT.Execute(self.main_code, env);

	EXCEPT

		(* If any exceptions occur then save them for later. *)
		Node.Abort(t)=>
			excepts := t;

	END;

	TRY

		(* Execute the finally code. *)
		ArmRT.Execute(self.finally_code, env);

	EXCEPT

		(* If any exceptions occur then add them to the existing set. *)
		Node.Abort(t)=>
			excepts := t.union(excepts);

	END;

	(* If any exceptions were raised then bomb out! *)
	IF excepts # NIL THEN
		RAISE Node.Abort(excepts);
	END;

END TryFinally_Execute;

PROCEDURE TryFinally_Print(self : TryFinally; i : TEXT) : TEXT=
BEGIN
	RETURN "try\n\t" & i & Node.ToText(self.main_code, i & "\t") & "\n" & i &
		"except\n\t" & i & Node.ToText(self.finally_code, i & "\t") & "\n" & i & "end"
END TryFinally_Print;

(* Exception trap object. *)
REVEAL	ExceptTrap = Node.T BRANDED OBJECT
				names : Node.T;
				code : Node.T
			OVERRIDES
				handle := ExceptTrap_Handle;
				print := ExceptTrap_Print;
			END;

PROCEDURE ExceptTrap_Handle(self : ExceptTrap;
				env : Node.E;
				t : Node.ExceptList) : Node.ExceptList RAISES {Node.Abort}=
VAR	pos : Node.T;
	run_code : BOOLEAN := FALSE;
BEGIN

	(* Runs through the listed exception names, checking to see
		whether each is raised. *)

	pos := self.names;
	WHILE pos # NIL DO

		(* Was exception raised? *)
		IF t.is_raised(pos.get_name(env)) THEN
			(* Reset the exception. *)
			t := t.unraise(pos.get_name(env));

			(* Set flag to make code be executed. *)
			run_code := TRUE;

		END;

		pos := pos.next;

	END;

	(* If code was marked to run, then do it! *)
	IF run_code THEN
		TRY
			ArmRT.Execute(self.code, env);
		EXCEPT
			(* If the handler code raises exceptions then add them to the
				list of other raise exceptions and bomb out of the handlers! *)
			Node.Abort(t2)=>
				t2 := t2.union(t);
				RAISE Node.Abort(t2);
		END;
	END;

	(* Return any unhandled exceptions. *)
	RETURN t;

END ExceptTrap_Handle;

PROCEDURE ExceptTrap_Print(self : ExceptTrap; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := self.names.print(i) & "=>\n\t";
	text := text & i & Node.ToText(self.code, i & "\t") & "\n" & i;
	RETURN text;
END ExceptTrap_Print;

(* List of exception trap-handlers. *)
REVEAL	ExceptTrapList = Node.T BRANDED OBJECT
				trap : Node.T;
			OVERRIDES
				handle := ExceptTrapList_Handle;
				print := ExceptTrapList_Print;
			END;

PROCEDURE ExceptTrapList_Handle(self : ExceptTrapList;
					env : Node.E;
					t : Node.ExceptList) : Node.ExceptList RAISES {Node.Abort}=
BEGIN

	(* Call this node's handler. *)
	t := self.trap.handle(env, t);

	(* Check for more handlers. *)
	IF self.next # NIL THEN
		RETURN self.next.handle(env, t);
	ELSE
		RETURN t;
	END;

END ExceptTrapList_Handle;

PROCEDURE ExceptTrapList_Print(self : ExceptTrapList; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := self.trap.print(i);
	IF self.next # NIL THEN
		text := text & "| " & self.next.print(i);
	END;
	RETURN text;
END ExceptTrapList_Print;

(* Routine to produce exception related objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TAG_Exception=>
			token := NEW(Exception, name := tree.extra);
			IF tree.params # NIL THEN
				token.next := Loader.Convert(tree.params.param);
			END;
			RETURN token;
	|	Tags.T.BLK_ExceptDef=>
			token := NEW(ExceptionCMD,
					names := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, ExceptionCMD).code :=
					Loader.ConvertCommand(tree.params.next.param);
				NARROW(token, ExceptionCMD).is_block := TRUE;
			END;
			RETURN token;
	|	Tags.T.CMD_Raise=>
			token := NEW(Raise,
					names := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, Raise).info := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.BLK_TryExcept=>
			RETURN NEW(Try,
					code := Loader.ConvertCommand(tree.params.param),
					traps := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.BLK_TryFinally=>
			RETURN NEW(TryFinally,
					main_code := Loader.ConvertCommand(tree.params.param),
					finally_code := Loader.ConvertCommand(tree.params.next.param)
				);
	|	Tags.T.TAG_ExceptTrap=>
			RETURN NEW(ExceptTrap,
					names := Loader.Convert(tree.params.param),
					code := Loader.ConvertCommand(tree.params.next.param)
				);
	|	Tags.T.TAG_ExceptTrap_List=>
			token := NEW(ExceptTrapList,
					trap := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Exceptions.