(* Expression module.
	Implements objects for basic expression
	constructs.
*)

(*
	13/12/96	Created skeleton with ExprList in
			it.
	17/12/96	Added the "printval" method for the expression
			list, to allow the new "print" command syntax.
	13/01/96	MAJOR RE-WRITE.  SEE NODE MODULE.
*)

MODULE Expression;

IMPORT Node, Loader, Except;
IMPORT Strings, Number, Bool;

IMPORT Code, Tags;

(* Expression list object. *)
REVEAL	ExprList = Node.T BRANDED OBJECT
			expr : Node.T;
		OVERRIDES
			get_expr := ExprList_GetExpr;
			print := ExprList_Print;
		END;

PROCEDURE ExprList_GetExpr(self : ExprList; <* UNUSED *> env : Node.E) : Node.T=
BEGIN
	RETURN self.expr;
END ExprList_GetExpr;

PROCEDURE ExprList_Print(self : ExprList; i : TEXT) : TEXT=
VAR	pos := self;
	text : TEXT := NIL;
BEGIN

	WHILE pos # NIL DO

		IF text = NIL THEN
			text := "";
		ELSE
			text := text & ", ";
		END;

		text := text & pos.expr.print(i);
		pos := pos.next;

	END;

	IF text = NIL THEN RETURN ""; END;
	RETURN text;

END ExprList_Print;

(* Addition object. *)
REVEAL	Plus = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Plus_GetValue;
			print := Plus_Print;
		END;

PROCEDURE Plus_GetValue(self : Plus; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	left_val := self.left.get_value(env);
	type := left_val.get_type().resolve(env);
BEGIN

	(* Now switch dependent upon type of left arg: *)
	IF ISTYPE(type, Number.NumberT) THEN
			RETURN Number.NewNumber(
				left_val.get_real(env) +
				self.right.get_value(env).get_real(env)
				);
	ELSIF ISTYPE(type, Strings.StringT) THEN
			RETURN Strings.NewString(
				left_val.get_string(env) &
				self.right.get_value(env).get_string(env)
				);
	ELSE
		RAISE Node.Abort(env.set_raised(Except.BadType, self.print("")));
	END;

END Plus_GetValue;

PROCEDURE Plus_Print(self : Plus; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " + " & self.right.print(i);
END Plus_Print;

(* Subtraction object. *)
REVEAL	Minus = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Minus_GetValue;
			print := Minus_Print;
		END;

PROCEDURE Minus_GetValue(self : Minus; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Number.NewNumber(
		self.left.get_value(env).get_real(env) -
		self.right.get_value(env).get_real(env)
		);
END Minus_GetValue;

PROCEDURE Minus_Print(self : Minus; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " - " & self.right.print(i);
END Minus_Print;

(* Logical OR. *)
REVEAL	Or = Node.T BRANDED OBJECT
			left, right : Node.T;
		OVERRIDES
			get_value := Or_GetValue;
			print := Or_Print;
		END;

PROCEDURE Or_GetValue(self : Or; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN Bool.NewBool(
		self.left.get_value(env).get_bool(env) OR
		self.right.get_value(env).get_bool(env)
		);
END Or_GetValue;

PROCEDURE Or_Print(self : Or; i : TEXT) : TEXT=
BEGIN
	RETURN self.left.print(i) & " or " & self.right.print(i);
END Or_Print;

(* Routine to convert Code trees into the above. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TAG_ExprList=>
			token := NEW(ExprList, expr := Loader.Convert(tree.params.param));
			IF tree.params.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.EXP_Plus=>
			RETURN NEW(Plus,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.EXP_Minus=>
			RETURN NEW(Minus,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.BOOL_Or=>
			RETURN NEW(Or,
				left := Loader.Convert(tree.params.param),
				right := Loader.Convert(tree.params.next.param)
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Expression.