(* Functions module.
	Implements the function closure type and the function
	related commands.
*)

(*
	16/12/96	Created.
	10/01/97	Added get_typecode method.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
	16/01/97	A signature must now be given for the function
			type.  This is now implemented.
*)

MODULE Functions;

(* IMPORT IO; *)

IMPORT Code, Tags;

IMPORT Node, Loader, Except, ArmRT, TypeSys;

(* The function closure datatype. *)
REVEAL	Closure = Node.K BRANDED OBJECT
			base_env : Node.E := NIL;
			params : Node.T := NIL;
			code : Node.T := NIL;
			returntype : Node.TYP := NIL;
		OVERRIDES
			get_closure := Closure_GetClosure;
			set_value := Closure_SetValue;
			print := Closure_Print;
			printval := Closure_PrintVal;
		END;

PROCEDURE Closure_GetClosure(self : Closure; env : Node.E; VAR base_env : Node.E; VAR params, code : Node.T; VAR returntype : Node.TYP)=
BEGIN
	base_env := self.base_env;
	params := self.params;
	code := self.code;
	returntype := self.returntype;
END Closure_GetClosure;

PROCEDURE Closure_SetValue(self : Closure; env : Node.E; val : Node.K) RAISES {Node.Abort}=
BEGIN
	val.get_closure(env, self.base_env, self.params, self.code, self.returntype);
END Closure_SetValue;

PROCEDURE Closure_Print(<* UNUSED *> self : Closure; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "fun";
END Closure_Print;

PROCEDURE Closure_PrintVal(self : Closure; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	RETURN "fun";
END Closure_PrintVal;

(* Closure allocator function. *)
REVEAL	ClosureT = Node.TYP BRANDED OBJECT
			params : Node.TYP := NIL;
			returntype : Node.TYP := NIL;
			resolved := FALSE;
		OVERRIDES
			allocate := ClosureT_Allocate;
			type_of_fun := ClosureT_TypeOfFun;
			check_type := ClosureT_CheckType;
			resolve := ClosureT_Resolve;
			print := ClosureT_Print;
		END;

PROCEDURE ClosureT_Allocate(self : ClosureT; env : Node.E) : Node.K=
BEGIN
	RETURN NEW(Closure);
END ClosureT_Allocate;

PROCEDURE ClosureT_TypeOfFun(self : ClosureT; env : Node.E;
				VAR params, return : Node.TYP)=
BEGIN
	params := self.params;
	return := self.returntype;
END ClosureT_TypeOfFun;

PROCEDURE ClosureT_CheckType(<* UNUSED *> s : ClosureT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Closures cannot be recursive. *)
	RETURN TRUE;
END ClosureT_CheckType;

PROCEDURE ClosureT_Resolve(s : ClosureT; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	IF NOT(s.resolved) THEN
		s.resolved := TRUE;
		IF s.params # NIL THEN s.params := s.params.resolve(env) END;
		IF s.returntype # NIL THEN s.returntype := s.returntype.resolve(env) END;
	END;
	RETURN s;
END ClosureT_Resolve;

PROCEDURE ClosureT_Print(self : ClosureT; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	RETURN "function (...) : ...";
END ClosureT_Print;

(* The function declaration block. *)
REVEAL	FunDef = Node.T BRANDED OBJECT
			fundecls : Node.T;
			code : Node.T;
			is_block := FALSE;
		OVERRIDES
			execute := FunDef_Execute;
			print := FunDef_Print;
		END;

(* This routine adds a new stack frame containing some variables. *)
PROCEDURE FunDef_Execute(self : FunDef; env : Node.E) RAISES {Node.Abort}=
VAR	vars : Node.VarList;
	new_env : Node.E;
BEGIN

	IF self.is_block THEN
		(* Create a new stack frame. *)
		new_env := env.make_child();
	ELSE
		(* Use the current frame. *)
		new_env := env;
	END;

	(* Create the varlist IN THE NEW FRAME and add it to the frame. *)
	new_env := new_env.add_variables(self.fundecls.get_varlist(new_env));

	IF self.is_block THEN
		(* Execute the specified commands in this frame. *)
		ArmRT.Execute(self.code, new_env);
	END;

END FunDef_Execute;

PROCEDURE FunDef_Print(self : FunDef; i : TEXT) : TEXT=
BEGIN
	RETURN "fun " & self.fundecls.print(i & "\t") & " in\n" & i & "\t" &
			Node.ToText(self.code, i & "\t");
END FunDef_Print;

(* Function declarator object. *)
REVEAL	FunDecl = Node.T BRANDED OBJECT
			varname : Node.T;
			params : Node.TYP;
			code : Node.T;
			returntype : Node.TYP;
			type : Node.TYP := NIL;
		OVERRIDES
			get_varlist := FunDecl_GetVarList;
			print := FunDecl_Print;
		END;

PROCEDURE FunDecl_GetVarList(self : FunDecl;
						env : Node.E;
						list : Node.VarList) : Node.VarList RAISES {Node.Abort}=
VAR	val := NEW(Closure);
BEGIN

	(* Set the value of the new closure. *)
	val.base_env := env;
	val.params := self.params;
	val.code := self.code;
	val.returntype := self.returntype;

	(* This routine caches a type-object so that we don't have to keep
		on creating them for every instance of the function that
		we produce. *)
	IF self.type = NIL THEN
		self.type := NEW(ClosureT,
				params := self.params,
				returntype := self.returntype,
				code := TypeSys.NewTypeCode()
			).resolve(env);
	END;

	(* Return the new variable with the new closure bound to it. *)
	RETURN NEW(Node.VarList).init(
				self.varname.get_name(env),
				NEW(Node.V).init(self.type, val)
			).setnext(list);

END FunDecl_GetVarList;

PROCEDURE FunDecl_Print(self : FunDecl; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := self.varname.print(i) & "(";
	IF self.params # NIL THEN
		text := text & self.params.print(i & "\t");
	END;
	text := text & ")";
	IF self.returntype # NIL THEN
		text := text & " : " & self.returntype.print(i & "\t\t");
	END;
	text := text & " = ";
	IF self.code # NIL THEN
		text := text & Node.ToText(self.code, i & "\t");
	END;
	RETURN text;
END FunDecl_Print;

(* Function declarator list. *)
REVEAL	FunDeclList = Node.T BRANDED OBJECT
			fundecl : Node.T;
		OVERRIDES
			get_varlist := FunDeclList_GetVarList;
			print := FunDeclList_Print;
		END;

PROCEDURE FunDeclList_GetVarList(self : FunDeclList;
						env : Node.E;
						list : Node.VarList) : Node.VarList RAISES {Node.Abort}=
BEGIN

	(* Get the VarList node for this fundecl. *)
	list := self.fundecl.get_varlist(env, list);

	(* If there are more fundecls then do them, too. *)
	IF self.next # NIL THEN
		list := self.next.get_varlist(env, list);
	END;

	RETURN list;

END FunDeclList_GetVarList;

PROCEDURE FunDeclList_Print(self : FunDeclList; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := self.fundecl.print(i);
	IF self.next # NIL THEN
		text := text & ", " & self.next.print(i);
	END;
	RETURN text;
END FunDeclList_Print;

(* Function call command object. *)
REVEAL	FunCall = Node.T BRANDED OBJECT
			varname : Node.T;
			params : Node.T;
		OVERRIDES
			execute := FunCall_Execute;
			get_value := FunCall_GetValue;
			print := FunCall_Print;
		END;

PROCEDURE FunCall_Execute(self : FunCall; env : Node.E) RAISES {Node.Abort}=
BEGIN
	IF FunCall_DoCall(self, env) # NIL THEN
		RAISE Node.Abort(env.set_raised(Except.BadType, self.print("")));
	END;
END FunCall_Execute;

PROCEDURE FunCall_GetValue(self : FunCall; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	val : Node.V;
BEGIN
	val := FunCall_DoCall(self, env);
	IF val = NIL THEN
		RAISE Node.Abort(env.set_raised(Except.BadType, self.print("")));
	END;
	RETURN val;
END FunCall_GetValue;

PROCEDURE FunCall_DoCall(self : FunCall; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	new_env : Node.E;
	base_env : Node.E;
	param_names : Node.T;
	returntype : Node.TYP;
	code : Node.T;
	expr : Node.T;
	num_exprs, num_params : CARDINAL;
	paramlist, pos : Node.VarList;
BEGIN

	(* Increment and test the function call count. *)
	ArmRT.IncFunCount(env);

	(* Firstly, get the function closure fields: *)
	self.varname.get_value(env).get_closure(env,
						base_env,
						param_names,
						code,
						returntype);

	IF param_names = NIL THEN
		paramlist := NIL;
	ELSE
		(* Generate a list of variable names and empty data-nodes for them. *)
		paramlist := param_names.get_varlist(env);
	END;

	(* Count the number of parameters required and expressions supplied: *)
	pos := paramlist;
	num_params := 0;
	WHILE pos # NIL DO
		INC(num_params);
		pos := pos.getnext();
	END;

	expr := self.params;
	num_exprs := 0;
	WHILE expr # NIL DO
		INC(num_exprs);
		expr := expr.next;
	END;

	IF num_params # num_exprs THEN
		RAISE Node.Abort(env.set_raised(Except.NumParams, self.print("")));
	END;

	(* Evaluate the parameters in the expression list and assign them to vars. *)
	pos := paramlist;
	expr := self.params;
	WHILE pos # NIL DO

		(* Assign the expression to the parameter. *)
		pos.getvalue().set_value(env, expr.get_expr(env).get_value(env));

		(* Move on to the next pair. *)
		pos := pos.getnext();
		expr := expr.next;

	END;

	(* Create a stack frame for the call, linked to the function's base env. *)
	new_env := env.make_child(base_env).add_variables(paramlist);

	(* Add a node of the required return type to the stack frame. *)
	new_env.add_return(returntype);

	TRY
		TRY
			IF code # NIL THEN
				(* Execute the function code in that frame. *)
				ArmRT.Execute(code, new_env);
			END;
		FINALLY
			(* Decrement the function call count. *)
			ArmRT.DecFunCount();
		END;
	EXCEPT
		(* If any exceptions were raised, then check for a return value. *)
		Node.Abort(t)=>
			IF t.is_raised(Except.Return) THEN
				t := t.unraise(Except.Return);
				IF t # NIL THEN
					(* Forget the return value - just abort. *)
					RAISE Node.Abort(t);
				ELSE
					(* Return the value. *)
					RETURN new_env.get_return();
				END;
			ELSE
				RAISE Node.Abort(t);
			END;
	END;

	(* If no exceptions occur, then return nothing! *)
	RETURN NIL;

END FunCall_DoCall;

PROCEDURE FunCall_Print(self : FunCall; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := "call " & self.varname.print(i) & "(";
	IF self.params # NIL THEN
		text := text & self.params.print(i);
	END;
	text := text & ")";
	RETURN text;
END FunCall_Print;

(* Ruotine to load in the above nodes. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Function=>
			token := NEW(ClosureT,
					params := Loader.Convert(tree.params.param),
					code := TypeSys.NewTypeCode()
					);
			IF tree.params.next # NIL THEN
				NARROW(token, ClosureT).returntype :=
					Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.CMD_Fun=>
			token := NEW(FunDef,
				fundecls := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, FunDef).code := Loader.ConvertCommand(tree.params.next.param);
				NARROW(token, FunDef).is_block := TRUE;
			END;
			RETURN token;
	|	Tags.T.TAG_FunDecl=>
			RETURN NEW(FunDecl,
				varname := Loader.Convert(tree.params.param),
				params := Loader.Convert(tree.params.next.param),
				returntype := Loader.Convert(tree.params.next.next.param),
				code := Loader.ConvertCommand(tree.params.next.next.next.param)
				);
	|	Tags.T.TAG_FunDeclList=>
			token := NEW(FunDeclList,
				fundecl := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.CMD_Call=>
			RETURN NEW(FunCall,
				varname := Loader.Convert(tree.params.param),
				params := Loader.Convert(tree.params.next.param)
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

PROCEDURE NewClosure(base_env : Node.E; code, params, return_type : Node.T) : Node.V=
BEGIN
	RETURN NEW(Node.V).init(
			NEW(ClosureT,
				params := params,
				returntype := return_type,
				code := TypeSys.NewTypeCode()
				),
			NEW(Closure,
				base_env := base_env,
				code := code,
				params := params,
				returntype := return_type
				)
			);
END NewClosure;

BEGIN
END Functions.