(* GfxTurtle module.
	This module implements the special graphics methods of
	the graphics turtle type.
*)

(*
	12/12/96	Created skeleton implementation.
	12/01/97	Added the "clear" and "set_colour" methods.
	16/01/97	Added the "ClearAll" routine, for clearing all
			active connections.
*)

MODULE GfxTurtle;

(* Custom network system. *)
IMPORT Display;

(* Custom turtle system. *)
IMPORT Turtle;

(* Netobjects. *)
IMPORT NetObj;

(* From anim3d. *)
IMPORT Point3, ColorProp;

(* System modules. *)
IMPORT Thread;

REVEAL	T = Public BRANDED OBJECT

			(* The display to which this turtle is connected. *)
			display : Display.Conn := NIL;

			(* Flag to indicate whether to draw when moving. *)
			draw : BOOLEAN := TRUE;

			(* Colour to draw in. *)
			pen := ColorProp.Base{1.0, 1.0, 1.0};

			(* Lock to protect from concurrency and flag to indicate
				whether or not this turtle is in use. *)
			lock : MUTEX;
			used : BOOLEAN;


		OVERRIDES

			(* GfxTurtle functions. *)
			attach := Attach;
			pendown := PenDown;

			(* Overrides of standard turtle functions. *)
			forward := Forward;
			backward := Backward;

			(* To set the pen colour. *)
			set_colour := SetColour;

			(* Locking routines. *)
			lock_turtle := Turtle_Lock;
			unlock_turtle := Turtle_Unlock;

			(* Duplication method. *)
			duplicate := Duplicate;

			(* Clear graphics command. *)
			clear := Clear

		END;

TYPE	ConnList = REF RECORD
			next : ConnList;
			display : Display.Conn;
		END;

	(* Storage for the list of connections to be closed when done. *)
VAR	connect_list : ConnList := NIL;

PROCEDURE Attach(self :T; display : Display.T) RAISES {NoDisplay}=
BEGIN

	TRY
		(* Now try to connect to the new display. *)
		self.display := display.init();

		(* Get the connections list. *)
		LOCK TurtleLock DO

			(* Add the new connection to the list. *)
			connect_list := NEW(ConnList,
						next := connect_list,
						display := self.display
					);

		END;

	EXCEPT
		NetObj.Error, Thread.Alerted=>
			RAISE NoDisplay;
	END;

END Attach;

PROCEDURE PenDown(self : T; draw : BOOLEAN)=
BEGIN
	self.draw := draw;
END PenDown;

PROCEDURE Forward(self : T; dist : Turtle.Units) RAISES {NoDisplay}=
VAR	old_pos : Point3.T;
	new_pos : Point3.T;
BEGIN

	(* Check that we're attached. *)
	IF self.display = NIL THEN
		RAISE NoDisplay;
	END;

	(* Move the turtle. *)
	old_pos := self.position();

	NARROW(self, Turtle.T).forward(dist);

	new_pos := self.position();

	(* Do we need to draw? *)
	IF self.draw THEN
		TRY
			(* Join the two points *)
			self.display.join(old_pos, new_pos, self.pen);
		EXCEPT
			NetObj.Error, Display.NotOpen, Thread.Alerted=>
				RAISE NoDisplay;
		END;	
	END;

END Forward;

PROCEDURE Backward(self : T; dist : Turtle.Units) RAISES {NoDisplay}=
VAR	old_pos : Point3.T;
	new_pos : Point3.T;
BEGIN

	IF self.display = NIL THEN
		RAISE NoDisplay;
	END;

	old_pos := self.position();

	NARROW(self, Turtle.T).backward(dist);

	new_pos := self.position();

	IF self.draw THEN
		TRY
			self.display.join(old_pos, new_pos, self.pen);
		EXCEPT
			NetObj.Error, Display.NotOpen, Thread.Alerted=>
				RAISE NoDisplay;
		END;	
	END;

END Backward;

PROCEDURE SetColour(self : T; r, g, b : REAL)=
BEGIN
	self.pen := ColorProp.Base{r, g, b};
END SetColour;

PROCEDURE Turtle_Lock(self : T) RAISES {CannotLock}=
BEGIN

	(* Lock the turtle. *)
	LOCK self.lock DO

		(* Check that no-one else is using it. *)
		IF self.used THEN
			RAISE CannotLock;
		ELSE
			self.used := TRUE;
		END;

	END;

END Turtle_Lock;

PROCEDURE Turtle_Unlock(self : T)=
BEGIN
	self.used := FALSE;
END Turtle_Unlock;

PROCEDURE Duplicate(self : T; new : T) : T=
BEGIN
	(* This routine is used to generate a turtle with the same properties as this one. *)

	(* Use the parent Turtle.T duplicate_pos method. *)
	new := self.duplicate_pos(new);

	(* And copy across the remaining fields. *)
	new.display := self.display;
	new.draw := self.draw;
	new.pen := self.pen;
	new.lock := NEW(MUTEX);
	new.used := FALSE;

	RETURN new;

END Duplicate;

PROCEDURE Clear(self : T) RAISES {NoDisplay}=
BEGIN

	(* This command clears any graphics due to the current turtle or
		to any turtle sharing the same display connection. *)
	IF self.display = NIL THEN
		RAISE NoDisplay;
	ELSE
		TRY
			self.display.clear();
		EXCEPT
			Display.NotOpen, NetObj.Error, Thread.Alerted=>
				RAISE NoDisplay;
		END;
	END;

END Clear;

PROCEDURE New() : T=
BEGIN
	RETURN NEW(T, lock := NEW(MUTEX)).init();
END New;

(* Procedure used when interpreter quits, to clean up any used displays.
	The GfxTurtle module keeps a list of used display connections so that
	they can be disconnected when done.
 *)
PROCEDURE Cleanup()=
VAR	pos : ConnList;
BEGIN

	(* Protect the connection list. *)
	LOCK TurtleLock DO

	(* Work through the list. *)
	pos := connect_list;
	WHILE pos # NIL DO

		TRY
			(* Try disconnecting from the display. *)
			IF pos.display # NIL THEN
				pos.display.close();
			END;
		EXCEPT
			NetObj.Error, Display.NotOpen, Thread.Alerted=>
		END;

		pos := pos.next;

	END;

	END;

END Cleanup;

(* Routine to clear all graphics from active connections. *)
PROCEDURE ClearAll()=
VAR	pos : ConnList;
BEGIN

	(* Protect the connection list from modification. *)
	LOCK TurtleLock DO

	(* Work through the list. *)
	pos := connect_list;
	WHILE pos # NIL DO

		TRY
			(* Try clearing the display. *)
			IF pos.display # NIL THEN
				pos.display.clear();
			END;
		EXCEPT
			NetObj.Error, Display.NotOpen, Thread.Alerted=>
		END;

		pos := pos.next;

	END;

	END;

END ClearAll;

VAR	TurtleLock := NEW(MUTEX);
BEGIN
END GfxTurtle.