(* Graphics module.
	This module gives implementation details for the
	Armadillo graphics commands.
*)

(*
	11/12/96	Created.
	10/01/97	Added get_typecode method to data objects.
	12/01/97	Added Colour command object.
			Added the ClearAll command and implemented
			the Clear and ClearAll commands.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
	19/02/97	Implemented equality for turtles.
	07/03/97	Finally implemented the Home command.
*)

MODULE Graphics;

(* System modules. *)
IMPORT Thread;

(* Network objects modules. *)
IMPORT NetObj;

(* Interpreter modules. *)
IMPORT Except, Loader, GfxTurtle;
IMPORT Node, TypeSys;

(* Armacode modules. *)
IMPORT Code, Tags;

(* Armanet modules. *)
IMPORT Display;

(* The Turtle data type: *)
REVEAL	Turtle = Node.K BRANDED OBJECT

			(* The graphics turtle for this turtle. *)
			turtle : GfxTurtle.T;

		OVERRIDES
			get_turtle := Turtle_GetTurtle;
			set_value := Turtle_SetValue;
			printval := Turtle_PrintVal;
			equal := Turtle_Equal;

		END;

PROCEDURE Turtle_GetTurtle(self : Turtle; env : Node.E) : GfxTurtle.T RAISES {Node.Abort}=
BEGIN
	RETURN self.turtle;
END Turtle_GetTurtle;

PROCEDURE Turtle_SetValue(self : Turtle; env : Node.E; val : Node.K) RAISES {Node.Abort}=
BEGIN
	self.turtle := val.get_turtle(env);
END Turtle_SetValue;

PROCEDURE Turtle_PrintVal(self : Turtle; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	RETURN self.turtle.info();
END Turtle_PrintVal;

PROCEDURE Turtle_Equal(self : Turtle; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN self.turtle = val.get_turtle(env);
END Turtle_Equal;

(* Turtle data type allocator object.  Generates instances of the above on request. *)
REVEAL	TurtleT = Node.TYP BRANDED OBJECT
		OVERRIDES
			allocate := TurtleT_Allocate;
			check_type := TurtleT_CheckType;
			print := TurtleT_Print;
		END;

PROCEDURE TurtleT_Allocate(self : TurtleT; env : Node.E) : Node.K RAISES {Node.Abort}=
BEGIN
	RETURN NEW(Turtle, turtle := GfxTurtle.New());
END TurtleT_Allocate;

PROCEDURE TurtleT_CheckType(<* UNUSED *> s : TurtleT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Turtles cannot be recursive. *)
	RETURN TRUE;
END TurtleT_CheckType;

PROCEDURE TurtleT_Print(self : TurtleT; i : TEXT) : TEXT=
BEGIN
	RETURN "turtle";
END TurtleT_Print;

(* The forward command. *)
REVEAL	Forward = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := Forward_Execute;
			print := Forward_Print;
		END;

PROCEDURE Forward_Execute(self : Forward; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Move the turtle forward. *)
		env.get_turtle().forward(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END Forward_Execute;

PROCEDURE Forward_Print(self : Forward; i : TEXT) : TEXT=
BEGIN
	RETURN "forward " & self.dist.print(i);
END Forward_Print;

(* The backward command. *)
REVEAL	Backward = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := Backward_Execute;
			print := Backward_Print;
		END;

PROCEDURE Backward_Execute(self : Backward; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Move the turtle backward. *)
		env.get_turtle().backward(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END Backward_Execute;

PROCEDURE Backward_Print(self : Backward; i : TEXT) : TEXT=
BEGIN
	RETURN "backward " & self.dist.print(i);
END Backward_Print;

(* The tleft command. *)
REVEAL	TLeft = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := TLeft_Execute;
			print := TLeft_Print;
		END;

PROCEDURE TLeft_Execute(self : TLeft; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Turn the turtle left. *)
		env.get_turtle().tleft(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END TLeft_Execute;

PROCEDURE TLeft_Print(self : TLeft; i : TEXT) : TEXT=
BEGIN
	RETURN "tleft " & self.dist.print(i);
END TLeft_Print;

(* The tright command. *)
REVEAL	TRight = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := TRight_Execute;
			print := TRight_Print;
		END;

PROCEDURE TRight_Execute(self : TRight; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Turn the turtle right. *)
		env.get_turtle().tright(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END TRight_Execute;

PROCEDURE TRight_Print(self : TRight; i : TEXT) : TEXT=
BEGIN
	RETURN "tright " & self.dist.print(i);
END TRight_Print;

(* The tup command. *)
REVEAL	TUp = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := TUp_Execute;
			print := TUp_Print;
		END;

PROCEDURE TUp_Execute(self : TUp; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Turn the turtle upwards. *)
		env.get_turtle().tup(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END TUp_Execute;

PROCEDURE TUp_Print(self : TUp; i : TEXT) : TEXT=
BEGIN
	RETURN "tup " & self.dist.print(i);
END TUp_Print;

(* The tdown command. *)
REVEAL	TDown = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := TDown_Execute;
			print := TDown_Print;
		END;

PROCEDURE TDown_Execute(self : TDown; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Turn the turtle downward. *)
		env.get_turtle().tdown(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END TDown_Execute;

PROCEDURE TDown_Print(self : TDown; i : TEXT) : TEXT=
BEGIN
	RETURN "tdown " & self.dist.print(i);
END TDown_Print;

(* The rleft command. *)
REVEAL	RLeft = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := RLeft_Execute;
			print := RLeft_Print;
		END;

PROCEDURE RLeft_Execute(self : RLeft; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Rotate the turtle anticlockwise. *)
		env.get_turtle().rleft(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END RLeft_Execute;

PROCEDURE RLeft_Print(self : RLeft; i : TEXT) : TEXT=
BEGIN
	RETURN "rleft " & self.dist.print(i);
END RLeft_Print;

(* The rright command. *)
REVEAL	RRight = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := RRight_Execute;
			print := RRight_Print;
		END;

PROCEDURE RRight_Execute(self : RRight; env : Node.E) RAISES {Node.Abort}=
BEGIN

	TRY
		(* Rotate the turtle right. *)
		env.get_turtle().rright(FLOAT(self.dist.get_value(env).get_real(env)));
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END RRight_Execute;

PROCEDURE RRight_Print(self : RRight; i : TEXT) : TEXT=
BEGIN
	RETURN "rright " & self.dist.print(i);
END RRight_Print;

(* The penup command. *)
REVEAL	PenUp = Node.T BRANDED OBJECT
		OVERRIDES
			execute := PenUp_Execute;
			print := PenUp_Print;
		END;

PROCEDURE PenUp_Execute(self : PenUp; env : Node.E) RAISES {Node.Abort}=
BEGIN

	env.get_turtle().pendown(FALSE);

END PenUp_Execute;

PROCEDURE PenUp_Print(self : PenUp; i : TEXT) : TEXT=
BEGIN
	RETURN "penup";
END PenUp_Print;

(* The pendown command. *)
REVEAL	PenDown = Node.T BRANDED OBJECT
			dist : Node.T;
		OVERRIDES
			execute := PenDown_Execute;
			print := PenDown_Print;
		END;

PROCEDURE PenDown_Execute(self : PenDown; env : Node.E) RAISES {Node.Abort}=
BEGIN

	env.get_turtle().pendown(TRUE);

END PenDown_Execute;

PROCEDURE PenDown_Print(self : PenDown; i : TEXT) : TEXT=
BEGIN
	RETURN "pendown";
END PenDown_Print;

(* The turtle command. *)
REVEAL	TurtleCMD = Node.T BRANDED OBJECT
			turtle : Node.T;
		OVERRIDES
			execute := TurtleCMD_Execute;
			print := TurtleCMD_Print;
		END;

PROCEDURE TurtleCMD_Execute(self : TurtleCMD; env : Node.E) RAISES {Node.Abort}=
VAR	old_turtle : GfxTurtle.T;
	new_turtle : GfxTurtle.T;
BEGIN

	TRY

		(* Get the previously selected turtle. *)
		old_turtle := env.get_turtle();

		(* Select the new turtle. *)
		new_turtle := self.turtle.get_value(env).get_turtle(env);

		new_turtle.lock_turtle();
		env.set_turtle(new_turtle);

		(* Unlock the old turtle. *)
		IF old_turtle # NIL THEN
			old_turtle.unlock_turtle();
		END;	

	EXCEPT
		GfxTurtle.CannotLock=>
			RAISE Node.Abort(env.set_raised(Except.TurtleUsed, self.turtle.print("")));
	END;

END TurtleCMD_Execute;

PROCEDURE TurtleCMD_Print(self : TurtleCMD; i :TEXT) : TEXT=
BEGIN
	RETURN "turtle " & self.turtle.print(i);
END TurtleCMD_Print;

(* The home command. *)
REVEAL	Home = Node.T BRANDED OBJECT
		OVERRIDES
			execute := Home_Execute;
			print := Home_Print;
		END;

PROCEDURE Home_Execute(self : Home; env : Node.E) RAISES {Node.Abort}=
BEGIN

	env.get_turtle().home();

END Home_Execute;

PROCEDURE Home_Print(self : Home; i : TEXT) : TEXT=
BEGIN
	RETURN "home";
END Home_Print;

(* The clear command. *)
REVEAL	Clear = Node.T BRANDED OBJECT
		OVERRIDES
			execute := Clear_Execute;
			print := Clear_Print;
		END;

PROCEDURE Clear_Execute(self : Clear; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* This command clears any graphics associated with the current
		turtle, or with any other turtle that shares it's display
		connection. *)
	TRY
		env.get_turtle().clear();
	EXCEPT
		GfxTurtle.NoDisplay=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	END;

END Clear_Execute;

PROCEDURE Clear_Print(self : Clear; i : TEXT) : TEXT=
BEGIN
	RETURN "clear";
END Clear_Print;

(* The clearall command. *)
REVEAL	ClearAll = Node.T BRANDED OBJECT
		OVERRIDES
			execute := ClearAll_Execute;
			print := ClearAll_Print;
		END;

PROCEDURE ClearAll_Execute(self : ClearAll; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* This command clears any graphics from this interpreter. *)
	GfxTurtle.ClearAll()

END ClearAll_Execute;

PROCEDURE ClearAll_Print(self : ClearAll; i : TEXT) : TEXT=
BEGIN
	RETURN "clearall";
END ClearAll_Print;

(* The attach command. *)
REVEAL	Attach = Node.T BRANDED OBJECT
			display : Node.T;
		OVERRIDES
			execute := Attach_Execute;
			print := Attach_Print;
		END;

PROCEDURE Attach_Execute(self : Attach; env : Node.E) RAISES {Node.Abort}=
VAR	connection : Display.T;
BEGIN

	TRY
		(* Make the connection to the display. *)
		connection := self.display.connect(env);

		(* Tell the current turtle to attach to this terminal. *)
		env.get_turtle().attach(connection);

	EXCEPT
		GfxTurtle.NoDisplay, GfxTurtle.NoDisconnect=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.display.print("")));
	END;

END Attach_Execute;

PROCEDURE Attach_Print(self : Attach; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN

	text := "attach " & self.display.print(i);
	RETURN text;

END Attach_Print;

(* The colour command object. *)
REVEAL Colour = Node.T BRANDED OBJECT
			r, g, b : Node.T;
		OVERRIDES
			execute := Colour_Execute;
			print := Colour_Print;
		END;

PROCEDURE Colour_Execute(self : Colour; env : Node.E) RAISES {Node.Abort}=
BEGIN
	env.get_turtle().set_colour(
		FLOAT(self.r.get_value(env).get_real(env)),
		FLOAT(self.g.get_value(env).get_real(env)),
		FLOAT(self.b.get_value(env).get_real(env))
	);
END Colour_Execute;

PROCEDURE Colour_Print(self : Colour; i : TEXT) : TEXT=
BEGIN
	RETURN "colour " & self.r.print(i & "\t") & ", " &
				self.g.print(i & "\t") & ", " &
				self.b.print(i & "\t");
END Colour_Print;

(* Objects used internally by the graphics system to represent displays. *)
REVEAL	DisplayID = Node.T BRANDED OBJECT
			display : Node.T;
			address : Node.T := NIL;
		OVERRIDES
			connect := DisplayID_Connect;
			print := DisplayID_Print;
		END;

PROCEDURE DisplayID_Connect(self : DisplayID; env : Node.E) : Display.T RAISES {Node.Abort}=
VAR	display : ROOT;
BEGIN

	(* Gets a Display.T variable from a display on a host machine. *)

	TRY

		(* Firstly, check to see if an IP address was given. *)
		IF self.address # NIL THEN
			display := NetObj.Import(
					Display.DisplayNameHdr & self.display.get_name(env),
					self.address.locate(env)
					);
		ELSE
			display := NetObj.Import(Display.DisplayNameHdr & self.display.get_name(env));
		END;

		(* Check that the imported object is ok. *)
		IF display = NIL THEN
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
		END;
		IF NOT(ISTYPE(display, Display.T)) THEN
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
		END;

	EXCEPT
		NetObj.Error=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	|	Thread.Alerted=>
			RAISE Node.Abort(env.set_raised(Except.Alerted, self.print("")));
	END;

	RETURN display;

END DisplayID_Connect;

PROCEDURE DisplayID_Print(self : DisplayID; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN

	text := self.display.print(i);
	IF self.address # NIL THEN
		text := text & "@" & self.address.print(i);
	END;
	RETURN text;

END DisplayID_Print;

REVEAL DisplayName = Node.T BRANDED OBJECT
			name : TEXT;
		OVERRIDES
			get_name := DisplayName_GetName;
			print := DisplayName_Print;
		END;

PROCEDURE DisplayName_GetName(self : DisplayName; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	RETURN self.name;
END DisplayName_GetName;

PROCEDURE DisplayName_Print(self : DisplayName; i : TEXT) : TEXT=
BEGIN
	RETURN self.name;
END DisplayName_Print;

REVEAL IPAddress = Node.T BRANDED OBJECT
			address : TEXT;
		OVERRIDES
			locate := IPAddress_Locate;
			print := IPAddress_Print;
		END;

PROCEDURE IPAddress_Locate(self : IPAddress; env : Node.E) : NetObj.Address RAISES {Node.Abort}=
BEGIN

	TRY
		(* Locate the specified address. *)
		RETURN NetObj.Locate(self.address);
	EXCEPT
		NetObj.Error, NetObj.Invalid=>
			RAISE Node.Abort(env.set_raised(Except.NoDisplay, self.print("")));
	|	Thread.Alerted=>
			RAISE Node.Abort(env.set_raised(Except.Alerted, self.print("")));
	END;
END IPAddress_Locate;

PROCEDURE IPAddress_Print(self : IPAddress; i : TEXT) : TEXT=
BEGIN
	RETURN self.address;
END IPAddress_Print;

(* Tree to objects conversion routine. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Turtle=>
			RETURN TurtleType;
	|	Tags.T.GFX_Forward=>
			RETURN NEW(Forward,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_Backward=>
			RETURN NEW(Backward,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_TLeft=>
			RETURN NEW(TLeft,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_TRight=>
			RETURN NEW(TRight,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_TUp=>
			RETURN NEW(TUp,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_TDown=>
			RETURN NEW(TDown,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_RLeft=>
			RETURN NEW(RLeft,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_RRight=>
			RETURN NEW(RRight,
				dist := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_Penup=>
			RETURN NEW(PenUp);
	|	Tags.T.GFX_Pendown=>
			RETURN NEW(PenDown);
	|	Tags.T.GFX_Turtle=>
			RETURN NEW(TurtleCMD,
				turtle := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_Home=>
			RETURN NEW(Home);
	|	Tags.T.GFX_Clear=>
			RETURN NEW(Clear);
	|	Tags.T.GFX_ClearAll=>
			RETURN NEW(ClearAll);
	|	Tags.T.GFX_Attach=>
			RETURN NEW(Attach,
				display := Loader.Convert(tree.params.param)
				);
	|	Tags.T.GFX_Colour=>
			RETURN NEW(Colour,
				r := Loader.Convert(tree.params.param),
				g := Loader.Convert(tree.params.next.param),
				b := Loader.Convert(tree.params.next.next.param)
				);
	|	Tags.T.TAG_Display=>
			token := NEW(DisplayID, display := Loader.Convert(tree.params.param));
			IF tree.params.next # NIL THEN
				NARROW(token, DisplayID).address := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.TAG_DisplayName=>
			RETURN NEW(DisplayName, name := tree.extra);
	|	Tags.T.TAG_IPAddress=>
			RETURN NEW(IPAddress, address := tree.extra);

	ELSE
		RETURN NIL;
	END;

END Convert;

(* Routine to allocate a new turtle. *)
PROCEDURE NewTurtle() : Node.V=
BEGIN
	RETURN NEW(Node.V).init(TurtleType, NEW(Turtle, turtle := GfxTurtle.New()));
END NewTurtle;

BEGIN
	TurtleType := NEW(TurtleT, code := TypeSys.NewTypeCode());
END Graphics.