(* Loader module.
	Defines a routine Go, which opens a file and loads
	a compiled program from it.  The program is converted
	into object code and this is returned.
*)

(*
	10/12/96	Created.
	04/02/97	Editted the Go function to use the verbose flag.
*)

MODULE Loader;

(* Interpreter modules. *)
IMPORT Node, ArmRT;
IMPORT Core, Graphics, Variables, Number;
IMPORT Expression, ArmDebug, Control, Exceptions;
IMPORT Strings, Term, Bool, BoolOp, Functions;
IMPORT AtomExp, Record, RefType, TypeSys;
IMPORT Condition, Concurrency, Ports;

(* Armacode library. *)
IMPORT Code, Tags;

(* System modules. *)
IMPORT Rd, FileRd, Pickle, Thread, OSError;
IMPORT IO, Fmt;

(* Routine to load in a program. *)
PROCEDURE Go(file : TEXT; verbose : BOOLEAN) : Node.T RAISES {FileFail, BadVersion, Failure}=
VAR	reader : Rd.T := NIL;
	tree : Code.T;
	ver : Code.Version;
BEGIN

	TRY
		(* Open up the file. *)
		IF verbose THEN	IO.Put("Opening file '" & file & "' for reading.\n"); END;
		reader := FileRd.Open(file);

		(* Load in the version tag. *)
		ver := Pickle.Read(reader);
		(* Check the version. *)
		IF (ver.ver # Tags.Version) OR (ver.rev # Tags.Revision) THEN
			RAISE BadVersion("(" & Fmt.Int(ver.ver) & "." & Fmt.Int(ver.rev) & ")");
		END;
		IF verbose THEN IO.Put("File version Ok.\n"); END;

		(* Load in the parse tree. *)
		tree := Pickle.Read(reader);
		IF verbose THEN IO.Put("Loaded compiled program.\n"); END;

		(* Close the file. *)
		IF verbose THEN IO.Put("Converting compiled program.\n"); END;
		Rd.Close(reader);

		(* Parse the tree. *)
		RETURN ConvertCommand(tree);

	EXCEPT
		OSError.E, Thread.Alerted, Pickle.Error,
		Rd.Failure, Rd.EndOfFile=>
			TRY
				IF reader # NIL THEN
					IF NOT(Rd.Closed(reader)) THEN
						Rd.Close(reader);
					END;
				END;
			EXCEPT
				Rd.Failure, Thread.Alerted=>
			END;
			RAISE FileFail;
	END;

END Go;

(* Routine to convert a parse tree into objects. *)
PROCEDURE ConvertCommand(tree : Code.T) : Node.T RAISES {Failure}=
VAR	obj : Node.T;
BEGIN

	obj := Convert(tree);
	IF obj = NIL THEN RETURN NIL; END;

	(* Do the next command. *)
	obj.next := ConvertCommand(tree.next);
	RETURN obj;

END ConvertCommand;

PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Failure}=
VAR	obj : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	IF tree.type = Tags.T.NO_OP THEN
		IO.Put("\twarning : no-op found!\n");
		RETURN NIL;
	END;

	obj := Core.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Graphics.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Variables.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Expression.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Term.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := BoolOp.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := AtomExp.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := ArmDebug.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Control.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Exceptions.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := TypeSys.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Concurrency.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Strings.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Number.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Bool.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Functions.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Record.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := RefType.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Condition.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := Ports.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	obj := ArmRT.Convert(tree);
	IF obj # NIL THEN RETURN obj; END;

	IO.Put("\tbad typecode : " & Fmt.Int(ORD(tree.type)) & "\n");
	RAISE Failure;

END Convert;

BEGIN
END Loader.