(* Main module for the Armadillo run-time interpreter.
	This module provides a simple interface to the functionality
	of the interpreter system, which reads a pre-compiled program
	a file and converts it into a special object-code which
	can then be interpreted.

	14/01/97	The interpreter used to wait for a key-press
			before quitting, to allow the user to view the
			results of a program.  The same effect can now
			be achieved using the "pause" command, so this
			has been removed.
	04/02/97	Most of the verbose output is now produced only if
			the /DEBUG flag is supplied, to keep output cleaner.
	19/02/97	The interpreter may now be run as a command-line
			interpreter, allowing the user direct access to the
			immense power of this amazing language, by connecting
			to an external, networked instance of the compiler.

*)

MODULE Main;

(* System libraries. *)
IMPORT IO, Fmt, Text, Params, Time;
IMPORT Thread, NetObj;

(* The armacode library. *)
IMPORT Tags, Code, NetParser;

(* The Armadillo interpreter modules. *)
IMPORT ArmRT, Loader, Node, GfxTurtle, Except;

EXCEPTION Quit;

(* Procedure to cope with special command line tokens. *)
PROCEDURE Special(token : TEXT) RAISES {Quit}=
BEGIN

	IF Text.Equal(token, "/DEBUG") THEN
		debug_mode := TRUE;

	ELSIF Text.Equal(token, "/NOCLEANUP") THEN
		cleanup := FALSE;

	ELSE
		IO.Put("I don't understand the tag " & token & "\n");
		Usage();
	END;

END Special;

(* Procedure to print Usage information. *)
PROCEDURE Usage() RAISES {Quit}=
BEGIN

	IO.Put("Usage : " & Params.Get(0) & " [<codefile-name>] [/DEBUG] [/NOCLEANUP]\n");
	RAISE Quit;

END Usage;

(* Command-Line execution system. *)
PROCEDURE CommandLine() RAISES {Quit}=
VAR	parser : NetParser.T;		(* Parser object for this instance. *)
	temp : ROOT;			(* Temporary storage in case the compiler is duff! *)
	command : TEXT := NIL;		(* The command-line. *)
	code : Code.T;			(* The code version. *)
	program : Node.T;		(* The objectised version. *)
	env : Node.E;			(* The local environment. *)
	quit_raised := FALSE;		(* Whether or not we must quit. *)
	connected := FALSE;		(* Whether or not we are connected to a server. *)
BEGIN

	IO.Put("Armadillo command line mode entered.\n\n");

	(* Create an environment to run in. *)
	env := Node.NewBaseEnv();

	LOOP

		(* Check that we are connected to the server. *)
		IF NOT(connected) THEN

			IO.Put("Contacting server...\n");

			TRY
				(* Attempt to connect to the compilation server! *)
				temp := NetObj.Import(NetParser.Name);

				IF temp = NIL THEN
					IO.Put("Unable to locate server.\n");
					RAISE Quit;
				ELSIF NOT(ISTYPE(temp, NetParser.T)) THEN
					IO.Put("Server is not compatible with this interpreter.\n");
					RAISE Quit;
				END;

				(* Narrow temp into parser object. *)
				parser := temp;
				(* Test the connection. *)
				parser.test();
				connected := TRUE;

			EXCEPT
				NetObj.Error, Thread.Alerted=>
					IO.Put("Unable to locate server.\n");
					RAISE Quit;
			END;

		END;

		TRY

			WHILE command = NIL DO

				(* Get the user to type in a command. *)
				IO.Put("\n- ");
				command := IO.GetLine();

				IF Text.Length(command)>=1 THEN

					WHILE Text.GetChar(command, Text.Length(command) - 1) = '~' DO
						IO.Put("= ");
						command := Text.Sub(command, 0, Text.Length(command)-1) & IO.GetLine();
					END;

				ELSE
					command := NIL;
				END;

			END;

			IF NOT(Text.Equal(command, "")) THEN

				(* Call the server to convert the command to code. *)
				(* The extra space is important! Leave it! *)
				code := parser.compile(command & " ");

				(* Clear the command text. *)
				command := NIL;

				(* Convert the code to executable objects. *)
				program := Loader.ConvertCommand(code);

				(* Run the command in the local environment. *)
				ArmRT.Execute(program, env);

			END;

		EXCEPT
			IO.Error=>
				IO.Put("\nAn I/O error has occurred.\n");
				RAISE Quit;
		|	NetObj.Error, Thread.Alerted=>
				(* Server has disconnected. *)
				IO.Put("Server is no longer available.\n");
				connected := FALSE;
		|	Loader.Failure=>
				IO.Put("Compiler returned garbage.\n");
		|	NetParser.Error(t)=>
				command := NIL;
				IO.Put(t & "\n");
		|	Node.Abort(list)=>
			TRY
				IO.Put("\nOne or more exceptions reached the outermost level:\n");
				VAR	text : TEXT;
				BEGIN
					WHILE list # NIL DO

						(* Print details of the raised exception. *)
						text := "\t" & list.getname();
						IF list.getinfo() # NIL THEN
							text := text & "\t" & list.getinfo();
						END;
						text := text & "\n";
						IO.Put(text);

						(* Check for the "quit" exception. *)
						IF Text.Equal(list.getname(), Except.Quit) THEN
							quit_raised := TRUE;
						END;

						list := list.getnext();
					END;
				END;

				(* Was "quit" raised? *)
				IF quit_raised THEN
					IO.Put("Execution halted by the 'quit' exception.\n");
					RAISE Quit;
				END;
			EXCEPT
				Node.Abort=>
					IO.Put("Run-time system failed!\n");
					RAISE Quit;
			END;
		END;

	END;

END CommandLine;

(* One-time execution from a compiled file. *)
PROCEDURE RunFile(progname : TEXT; debug_mode : BOOLEAN) RAISES {Quit}=
VAR	compiledname : TEXT;	(* Compiled code file name. *)
	program : Node.T;	(* Compiled program code. *)
	envir : Node.E;		(* Environment to run program in. *)
	timetaken : Time.T;	(* Time taken to run the program. *)
BEGIN
	compiledname := progname & Tags.CodeExt;

	(* Try to load and run the source file. *)
	TRY

		program := Loader.Go(compiledname, debug_mode);
		IF debug_mode THEN
			IO.Put("Program loaded & converted.\n");
		END;

		IF program # NIL THEN

			IF debug_mode THEN
				IO.Put("\n" & Node.ToText(program) & "\n\n");
				IO.Put("Running program:\n\n");
			END;

			(* Create an environment to run it in. *)
			envir := Node.NewBaseEnv();

			(* Get the current timer. *)
			timetaken := Time.Now();

			(* Run the program. *)
			ArmRT.Execute(program, envir);

			IF debug_mode THEN
				(* Find how much time was taken. *)
				timetaken := Time.Now() - timetaken;
				IO.Put("\n\ttime elapsed:" & Fmt.LongReal(timetaken, Fmt.Style.Auto, 2) & "\n");
			END;

		ELSE
			IO.Put("Nothing to do!\n");
		END;

	EXCEPT
		Node.Abort(list)=>
			IO.Put("\nOne or more exceptions reached the outermost level:\n");
			VAR	text : TEXT;
			BEGIN
				WHILE list # NIL DO
					text := "\t" & list.getname();
					IF list.getinfo() # NIL THEN
						text := text & "\t" & list.getinfo();
					END;
					text := text & "\n";
					IO.Put(text);
					list := list.getnext();
				END;
			END;
	|	Loader.FileFail=>IO.Put("\nCouldn't load program!\n");
	|	Loader.BadVersion(v)=>IO.Put("\nFile is the wrong version." & v & "\n");
	|	Loader.Failure=>
		IO.Put("The program contains invalid node tags.\n");
		IO.Put("Some compiler-supported constructs are not supported ");
		IO.Put("by this interpreter, which may lead to this problem.\n");
	END;

END RunFile;

VAR	progname : TEXT;	(* The program name. *)
	debug_mode := FALSE;	(* Debug mode flag. *)
	cleanup := TRUE;	(* Whether or not to clean up on exit. *)

BEGIN

	IO.Put("\nArmadillo 3D Interpreter version "
		& Fmt.Int(Tags.Version) & "."
		& Fmt.Int(Tags.Revision) & "\n\n");

	TRY

		(* Examine the parameter line for interesting stuff. *)
		IF Params.Count > 1 THEN
			FOR x:=1 TO Params.Count-1 DO
				IF Text.GetChar(Params.Get(x), 0) = '/' THEN

					(* This is a command-line parameter, so do
						something special with it. *)
					Special(Params.Get(x));

				ELSE

					(* Have we already got a program name? *)
					IF progname = NIL THEN
						(* No, so use this one. *)
						progname := Params.Get(x);
					ELSE
						(* Yes, so print an error. *)
						Usage();
					END;	
				END;
			END;
		END;

		IF progname = NIL THEN

			(* Not file name was specified, so run in command-line mode. *)
			CommandLine();

		ELSE

			(* File name was specified, so go for one-time file run only. *)
			RunFile(progname, debug_mode);

		END;
	EXCEPT
		Quit=>
	END;

	IF cleanup THEN
		(* Close any turtle connections made. *)
		GfxTurtle.Cleanup();
	END;

END Main.