(* Node module.
	This module defines the base node object on which
	all the command, operator etc. nodes are based.

	Node.T is the general node type.
	Node.K is the "value" node type.
	Node.V is the "value & type" node type.
	Node.E is the environment object type.

*)

(*
	10/12/96	Created.
	12/12/96	Added turtle routines to the program node
			and environment frame objects.
	14/12/96	Added the "handle" method to allow exception-handler
			objects to be created and used.
	15/12/96	Added the "compare" method to affect comparisons.
			Also added the "get_value" method to allow return values
			of any type to be obtained, which is useful when routines
			don't care about the types of their arguments.
	10/01/97	Added method for obtaining a type-code describing the
			structural composition of a value or type.
	11/01/97	Added methods to environment objects for getting a handle
			on the current thread.
			Added methods for the mutex data-type.
	13/01/97	FUNDAMENTAL RE-WRITE OF THE VALUE-PASSING SYSTEM.
			The result of get_value and value passed to set_value are
			now of type Node.V.  Data-type still export sub-types of
			the Node.K, but calculations should all return a Node.V
			containing a type constructor and a value of Node.K type.
	15/01/97	Changed the signature of the "set_raised" method of environment
			objects to allow an information string to be given when an
			exception is raised.  This also means adding a "getinfo" to
			the ExceptList object, to allow the info string to be retrieved.
	17/01/97	Added the "get_typecode" method to the environment, to get
			the typecode of a user data-type.
	18/01/97	Added the "type_of_ref" method so that it's easy to dereference
			a value.
			Added the "get_nil" method for type-constructors so that reference
			type constructors can provide an efficient(!) way to get "nil" for
			that type.
	20/01/97	Removed get_nil method - no longer necessary.
	22/01/97	Created a special sub-type of the Node.T object for datatype objects
			to be derived from.
	28/01/97	Added the "check_type" method for types so that they can check their
			own structures for infinite, illegal loops.
	04/02/97	Added methods to support ports as an in-built primitive.
	21/02/97	MAJOR RE-WRITE of the exception system.  This affects any other
			module containing exception-handling!
	26/02/97	Added the "get_typelist" method to the Node.T to support the new
			type-declaration structure.
*)

INTERFACE Node;

IMPORT Thread;

(* Network object modules. *)
IMPORT NetObj;

(* Armanet modules. *)
IMPORT Display;

(* Interpreter modules. *)
IMPORT GfxTurtle;

(* Exception used to signal failed execution to the previous layer. *)
EXCEPTION Abort(ExceptList);

(* BASIC PROGRAM NODE. *)

TYPE	T <: Public;
	Public = OBJECT
			next : T := NIL;
		METHODS

			(* Simple command execution. *)
			execute(env : E) RAISES {Abort};

			(* Routine for returning lists of variable names,
				fields or type names. *)
			get_varlist(env : E; list : VarList := NIL) : VarList RAISES {Abort};
			get_fields(env : E) : Fields RAISES {Abort};
			get_typelist(env : E; list : TypeList := NIL) : TypeList RAISES {Abort};

			(* Generic value retrieval routines.  These are the
				ones that operators should export. Node.K values
				should NOT export this! *)
			get_value(env : E) : V RAISES {Abort};

			(* Routines for working with displays. *)
			connect(env : E) : Display.T RAISES {Abort};
			locate(env : E) : NetObj.Address RAISES {Abort};

			(* Routines for debugging. *)
			print(indent : TEXT) : TEXT;

			(* Routine to get the name from a variable. *)
			get_name(env : E) : TEXT RAISES {Abort};

			(* Routine for getting expressions out of lists. *)
			get_expr(env : E) : T RAISES {Abort};

			(* Routine to get an object to handle some exceptions. *)
			handle(env : E; t : ExceptList) : ExceptList RAISES {Abort};

		END;

(* FIELD LIST.  CONTAINS A LIST OF DATA VALUES (No type info or names).
	This is used by records in constructing their data storage. *)

	Fields = OBJECT
			value : K := NIL;
			next : Fields := NIL;
		END;

(* DATA-TYPE PARENT OBJECT, FROM WHICH ALL DATA-TYPES SHOULD BE DERIVED. *)

	TYP <: PublicTYP;
	PublicTYP = T OBJECT
			code : CARDINAL := 0;			(* Type-code of the type. *)
		METHODS

			(* Routines for dealing with record types. *)
			get_field_no(env : E; name : TEXT) : INTEGER RAISES {Abort};
			get_field_type(env : E; n : CARDINAL) : TYP RAISES {Abort};

			(* Routine to create instance of type, including type-info. *)
			allocate_v(env : E) : V RAISES {Abort};
			(* Routine to generate just the storage for a data-type. *)
			allocate(env : E) : K RAISES {Abort};

			(* This routine returns a structural type-code for the value. *)
			get_typecode(env : E) : CARDINAL RAISES {Abort};
			(* Procedure to reset the type-code to a new value. *)
			set_typecode(env : E; code : CARDINAL) RAISES {Abort};
			(* Routine to resolve a type's structure. *)
			resolve(env : E) : TYP RAISES {Abort};
			(* Routine to check that a type is valid. *)
			check_type(env : E; l : InvalidTypes) : BOOLEAN RAISES {Abort};

			(* Routine to recover the type of thing pointed to by a ref. *)
			type_of_ref(env : E) : TYP RAISES {Abort};

			(* Routine to get the parameter and return types for a function. *)
			type_of_fun(env : E; VAR params, return : TYP) RAISES {Abort};

		END;

	(* Data structure used by check_type to pass lists of unnacceptable type-codes. *)
	InvalidTypes = REF RECORD
			next : InvalidTypes := NIL;
			typecode : CARDINAL;
		END;

(* ARMADILLO DATA-VALUE OBJECT. *)

	K <: PublicK;
	PublicK = T OBJECT
		METHODS
			(* Routines for getting the value.  *)

			get_real(env : E) : LONGREAL RAISES {Abort};
			get_integer(env : E) : INTEGER RAISES {Abort};
			get_string(env : E) : TEXT RAISES {Abort};
			get_bool(env : E) : BOOLEAN RAISES {Abort};
			get_ref(env : E) : K RAISES {Abort};
			get_turtle(env : E) : GfxTurtle.T RAISES {Abort};
			get_closure(env : E;
					VAR base_env : E;
					VAR params, code : T;
					VAR returntype : TYP
				) RAISES {Abort};
			get_field(env : E; n : CARDINAL) : K RAISES {Abort};
			get_condition(env : E) : Thread.Condition RAISES {Abort};

			(* The general "set value" routine.  Doesn't
				need to do type-checking - this is done by the
				parent Node.V object. *)
			set_value(env : E; val : K) RAISES {Abort};

			(* Routine to compare an object to another. *)
			compare(env : E; obj : K) : Result RAISES {Abort};
			equal(env : E; obj : K) : BOOLEAN RAISES {Abort};

			(* Print routine for the "print" command to use. *)
			printval(env : E) : TEXT RAISES {Abort};

			(* Routine used to initialise storage for "ref"s. *)
			alloc_ref(env : E; type : TYP) RAISES {Abort};

			(* Routines for signalling and waiting on condition vars. *)
			signal(env : E) RAISES {Abort};
			wait(env : E; lock : MUTEX) RAISES {Abort};

			(* Routines for sending and recieving on ports. *)
			send(env : E; val : V; lock : MUTEX) RAISES {Abort};
			recieve(env : E; lock : MUTEX) : V RAISES {Abort};

			(* Routines for locking and unlocking mutexes. *)
			lock(env : E) : BOOLEAN RAISES {Abort};
			unlock(env : E) RAISES {Abort};
			wait_unlocked(env : E; lock : MUTEX) RAISES {Abort};

		END;

(* THE ARMADILLO DATA-TYPE OBJECT. *)
	V <: PublicV;
	PublicV = T OBJECT
		METHODS
			(* Routines for getting the value.  These routines should
				only ever be implemented for actual types, not
				for operators. *)
			get_real(env : E) : LONGREAL RAISES {Abort};
			get_integer(env : E) : INTEGER RAISES {Abort};
			get_string(env : E) : TEXT RAISES {Abort};
			get_bool(env : E) : BOOLEAN RAISES {Abort};
			get_ref(env : E) : K RAISES {Abort};
			get_turtle(env : E) : GfxTurtle.T RAISES {Abort};
			get_closure(env : E;
					VAR base_env : E;
					VAR params, code: T;
					VAR returntype : TYP
				) RAISES {Abort};
			get_field_byname(env : E; name : TEXT) : V RAISES {Abort};
			get_condition(env : E) : Thread.Condition RAISES {Abort};

			(* The general "set value" routine. *)
			set_value(env : E; val : V) RAISES {Abort};

			(* Routine to compare an object to another. *)
			compare(env : E; obj : V) : Result RAISES {Abort};
			equal(env : E; obj : V) : BOOLEAN RAISES {Abort};

			(* Print routine for the "print" command to use. *)
			printval(env : E) : TEXT RAISES {Abort};

			(* Routine used to initialise storage for "ref"s. *)
			alloc_ref(env : E) RAISES {Abort};

			(* Routines for signalling and waiting on condition vars. *)
			signal(env : E) RAISES {Abort};
			wait(env : E; lock : MUTEX) RAISES {Abort};

			(* Routines for sending and recieving on ports. *)
			send(env : E; val : V; lock : MUTEX) RAISES {Abort};
			recieve(env : E; lock : MUTEX) : V RAISES {Abort};

			(* Routines for locking and unlocking mutexes. *)
			lock(env : E) : BOOLEAN RAISES {Abort};
			unlock(env : E) RAISES {Abort};
			wait_unlocked(env : E; lock : MUTEX) RAISES {Abort};

			(* Routine for initialising a Node.V. *)
			init(type : TYP; val : K) : V;

			(* Routine for obtaining the type constructor of
				a value. *)
			get_type() : TYP;
			(* Routine for obtaining data without type. *)
			get_data() : K;

		END;

(* COMPARISON RESULT TYPE: *)
	Result = [-1..1];

(* ENVIRONMENT STRUCTURES. *)

	E <: PublicEnv;
	PublicEnv = OBJECT
		METHODS
			(* ENVIRONMENTS. *)

			(* Make_child makes a new environment and links it
				either to this one, or to another. *)
			make_child(linkto : E := NIL) : E;

			(* Function to create a complete base environment from
				this one, containing all it's exceptions. *)
			make_base(linkto : E := NIL) : E;

			(* Get_parent returns the parent env frame. *)
			get_parent() : E;

			(* TURTLES *)
			add_turtle();
			set_turtle(turtle : GfxTurtle.T) RAISES {Abort};
			get_turtle() : GfxTurtle.T RAISES {Abort};

			(* VARIABLES. *)

			(* Add new vars to the current environment. *)
			add_variables(list : VarList) : E;

			get_var(name : TEXT) : V RAISES {Abort};

			(* EXCEPTIONS. *)

			(* Create a sub-env frame containing new exception names. *)
			add_exceptions(list : ExceptList) : E;

			(* Set an exception in the current env.  If not available
				in current env, then parent is checked etc. *)
			set_raised(name : TEXT; info : TEXT) : ExceptList RAISES {Abort};

			(* FUNCTION RETURN VALUES. *)

			(* Set the type of the return-value for the stack frame. *)
			add_return(type : TYP) RAISES {Abort};
			(* Set the return-value of the stack frame. *)
			set_return(val : V) RAISES {Abort};
			(* Get the return-value. *)
			get_return() : V RAISES {Abort};

			(* USER TYPES. *)

			(* Add a new user type. *)
			add_types(types : TypeList) : E RAISES {Abort};
			(* Get a user type. Returns the environment in which it was found in "env" *)
			get_type(name : TEXT) : TYP RAISES {Abort};

			(* THREADS. *)

			(* Make this environment a base thread environment. *)
			make_thread_base();
			(* Get the current base environment for this thread. *)
			get_thread_base() : E;

		END;

	(* List structure for passing to the add_variables routine. *)
	VarList <: VarListP;
	VarListP = OBJECT
		METHODS
			init(varname : TEXT; value : V) : VarList;
			setnext(list : VarList) : VarList;
			getname() : TEXT;
			getnext() : VarList;
			getvalue() : V;
		END;

	(* List structure for passing to the add_exceptions routine. *)
	ExceptList <: ExceptListP;
	ExceptListP = OBJECT
		METHODS
				(* For initalising the ExceptLists. *)
			init(exceptname : TEXT) : ExceptList;
			setnext(list : ExceptList) : ExceptList;
				(* Getting info from ExceptLists. *)
			getname() : TEXT;
			getnext() : ExceptList;
			getinfo() : TEXT;
				(* Stick together two ExceptLists. *)
			union(with : ExceptList) : ExceptList;
				(* Remove an exception from the ExceptLists. *)
			unraise(name : TEXT) : ExceptList;
				(* Check if an exception is in the ExceptList. *)
			is_raised(name : TEXT) : BOOLEAN;
				(* Find an exception in an ExceptList. *)
			find(name : TEXT) : ExceptList;
		END;

	(* TYPE LIST CONSTRUCTOR OBJECT. *)
	TypeList <: TypeListP;
	TypeListP = OBJECT
		METHODS
			init(name : TEXT; type : TYP) : TypeList;
			setnext(list : TypeList) : TypeList;
			getname() : TEXT;
			getnext() : TypeList;
			gettype() : TYP;
			resolve(env : E) RAISES {Abort};
		END;

(* Procedure to create a new "base" environment, containing various
	useful add-ons (e.g. system exceptions) and stuff. *)
PROCEDURE NewBaseEnv() : E;

(* This procedure prints out a tree of commands using their "print" methods. *)
PROCEDURE ToText(tree : T; indent := "") : TEXT;

END Node.
