(* Node module.
	Provides some simple stubs for some of the routines.
*)

(*
	10/12/96	Created.
	12/12/96	Added some extra methods for handling turtles
			and exceptions.
	15/12/96	Added the compare method for comparing objects.
			Also added the get_value method.
	10/01/97	Added the get_typecode method.
	11/01/97	Added methods to the environment object type for
			obtaining a handle to the current thread.
			Added stubs for the mutex data-type methods.
	13/01/97	MAJOR RE-WRITE OF VALUE-PASSING SYSTEM!
	14/01/97	Added an info field to the "set_raised" environment
			object method.  This can be retrieved from the raised
			exception list using the getinfo() method, which
			may return NIL.
	17/01/97	Added the "get_typecode" method to the environment
			object to get the typecode of a user type.
	18/01/97	Added "type_of_ref" method to the T object type.
	20/01/97	Removed get_nil method - no longer necessary.
	22/01/97	Added the TYPE object from which datatypes are derived.
	04/02/97	Added method stubs for the port methods.
	16/02/97	RaisedList routine may now return either all exceptions
			raised in current scope or only those in current frame.
	17/02/97	Fixed SetRaised so that the "no_such_except" exception
			is raised in the correct frame if necessary.
	18/02/97	Fixed scope of some exceptions.
	19/02/97	Fixed scope of exceptions properly!  Exceptions were
			previously raised in outer stack frames using the "chain"
			pointer in the frames.  Now they use the "parent" pointer.
	21/02/97	MAJOR RE-WRITE of exception-handling sub-system.
	05/03/97	Fixed the "printval" default method to raise the correct
			exception.
*)

MODULE Node;

(* ** DEBUGGING! ** *)
(* IMPORT IO; *)

(* System modules. *)
IMPORT Text, Thread;

(* Networking. *)
IMPORT NetObj;

(* Armanet modules. *)
IMPORT Display;

(* Interpreter run-time modules. *)
IMPORT GfxTurtle;
IMPORT Except, TypeSys;

(* THE MAIN PROGRAM NODE TYPE! *)

REVEAL	T = Public BRANDED OBJECT
		OVERRIDES
			execute := Stub_Execute;

			get_varlist := Stub_GetVarList;
			get_fields := Stub_GetFields;

			get_value := Stub_GetValue;

			connect := Stub_Connect;
			locate := Stub_Locate;

			print := Stub_Print;

			get_name := Stub_GetName;

			get_expr := Stub_GetExpr;

			handle := Stub_Handle;
		END;

(* Stub procedures. *)

PROCEDURE Stub_Execute(s : T; env : E) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Execute;

PROCEDURE Stub_GetVarList(s : T; env : E; list : VarList) : VarList RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetVarList;
PROCEDURE Stub_GetFields(s : T; env : E) : Fields RAISES {Abort}=

	PROCEDURE getfields(list : VarList) : Fields=
	BEGIN
		IF list = NIL THEN RETURN NIL; END;
		RETURN NEW(Fields,
				value := list.getvalue().value,
				next := getfields(list.getnext())
			);
	END getfields;

BEGIN
	(* Get the varlist and then convert it. *)
	RETURN getfields(s.get_varlist(env));
END Stub_GetFields;

PROCEDURE Stub_GetValue(s : T; env : E) : V RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetValue;

PROCEDURE Stub_Connect(s : T; env : E) : Display.T RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Connect;
PROCEDURE Stub_Locate(s : T; env : E) : NetObj.Address RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Locate;

PROCEDURE Stub_Print(s : T; indent : TEXT) : TEXT=BEGIN RETURN "<unknown>"; END Stub_Print;

PROCEDURE Stub_GetName(s : T; env : E) : TEXT RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetName;

PROCEDURE Stub_GetExpr(s : T; env : E) : T RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetExpr;

PROCEDURE Stub_Handle(s : T; env : E; t : ExceptList) : ExceptList RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Handle;

(* OBJECT FROM WHICH DATA-TYPES ARE DERIVED! *)
REVEAL	TYP = PublicTYP BRANDED OBJECT
		OVERRIDES

			get_field_no := Stub_GetFieldNo;
			get_field_type := Stub_GetFieldType;

			get_typecode := Stub_GetTypeCode;
			set_typecode := Stub_SetTypeCode;
			resolve := Stub_Resolve;
			check_type := Stub_CheckType;

			allocate_v := Stub_AllocateV;
			allocate := Stub_Allocate;

			type_of_ref := Stub_TypeOfRef;
			type_of_fun := Stub_TypeOfFun;

		END;

PROCEDURE Stub_GetFieldNo(s : TYP; env : E; name : TEXT) : INTEGER RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetFieldNo;
PROCEDURE Stub_GetFieldType(s : TYP; env : E; n : CARDINAL) : TYP RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetFieldType;

PROCEDURE Stub_GetTypeCode(s : TYP; env : E) : CARDINAL RAISES {Abort}=
BEGIN
	IF s.code = 0 THEN
		RAISE Abort(
			env.set_raised(Except.Unsupported, "Typecode not set:" & s.print(""))
			);
	ELSE
		RETURN s.code;
	END;
END Stub_GetTypeCode;

PROCEDURE Stub_SetTypeCode(s : TYP; env : E; code : CARDINAL) RAISES {Abort}=
BEGIN
	(* Typecodes may decrease, but never increase. *)
	IF code < s.code THEN
		s.code := code;
	END;
END Stub_SetTypeCode;

PROCEDURE Stub_Resolve(s : TYP; <* UNUSED *> env : E) : TYP=
BEGIN
	(* Any simple type should just return itself. *)
	RETURN s;
END Stub_Resolve;

PROCEDURE Stub_CheckType(s : TYP; env : E; <* UNUSED *> i : InvalidTypes) : BOOLEAN RAISES {Abort}=
BEGIN
	(* If it's not implemented for a type then fail! *)
	RAISE Abort(
		env.set_raised(Except.Unsupported, "check_type unimplemented : " & s.print("\t\t"))
		);
END Stub_CheckType;

PROCEDURE Stub_AllocateV(s : TYP; env : E) : V RAISES {Abort}=
BEGIN
	(* Check that the type being allocated is valid. *)
	IF s.check_type(env, NIL) THEN
		RETURN NEW(V).init(s.resolve(env), s.allocate(env));
	ELSE
		RAISE Abort(
			env.set_raised(Except.InfiniteType, s.print("\t\t"))
			);
	END;
END Stub_AllocateV;
PROCEDURE Stub_Allocate(s : TYP; env : E) : K RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Allocate;

PROCEDURE Stub_TypeOfRef(s : TYP; env : E) : TYP RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_TypeOfRef;

PROCEDURE Stub_TypeOfFun(s : TYP; env : E; VAR params, return : TYP) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_TypeOfFun;

(* OBJECT FROM WHICH DATA VALUES ARE DERIVED! *)

REVEAL	K = PublicK BRANDED OBJECT
		OVERRIDES
			get_real := Stub_GetReal;
			get_integer := Stub_GetInt;
			get_string := Stub_GetString;
			get_bool := Stub_GetBool;
			get_ref := Stub_GetRef;
			get_turtle := Stub_GetTurtle;
			get_closure := Stub_GetClosure;
			get_field := Stub_GetField;
			get_condition := Stub_GetCondition;

			set_value := Stub_SetValue;

			compare := Stub_Compare;

			printval := Stub_PrintVal;

			alloc_ref := Stub_AllocRef;

			signal := Stub_Signal;
			wait := Stub_Wait;

			send := Stub_Send;
			recieve := Stub_Recieve;

			lock := Stub_Lock;
			unlock := Stub_UnLock;
			wait_unlocked := Stub_WaitUnLock;

		END;

PROCEDURE Stub_GetReal(s : K; env : E) : LONGREAL RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL));END Stub_GetReal;
PROCEDURE Stub_GetInt(s : K; env : E) : INTEGER RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetInt;
PROCEDURE Stub_GetString(s : K; env : E) : TEXT RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetString;
PROCEDURE Stub_GetBool(s : K; env : E) : BOOLEAN RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetBool;
PROCEDURE Stub_GetRef(s : K; env : E) : K RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetRef;
PROCEDURE Stub_GetTurtle(s : K; env : E) : GfxTurtle.T RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetTurtle;
PROCEDURE Stub_GetClosure(s : K; env : E; VAR base_env : E; VAR params, code : T; VAR returntype : TYP) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetClosure;
PROCEDURE Stub_GetField(s : K; env : E; n : CARDINAL) : K RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetField;
PROCEDURE Stub_GetCondition(s : K; env : E) : Thread.Condition RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_GetCondition;

PROCEDURE Stub_SetValue(s : K; env : E; val : K) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_SetValue;

PROCEDURE Stub_Compare(s : K; env : E; obj : K) : Result RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Compare;

PROCEDURE Stub_PrintVal(s : K; env : E) : TEXT RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.Unsupported, s.print("") & ".printval")); END Stub_PrintVal;

PROCEDURE Stub_AllocRef(s : K; env : E; type : TYP) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_AllocRef;

PROCEDURE Stub_Signal(s : K; env : E) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Signal;
PROCEDURE Stub_Wait(s : K; env : E; lock : MUTEX) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Wait;

PROCEDURE Stub_Send(s : K; env : E; val : V; lock : MUTEX) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.Unsupported, "send")); END Stub_Send;
PROCEDURE Stub_Recieve(s : K; env : E; lock : MUTEX) : V RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.Unsupported, "recieve")); END Stub_Recieve;

PROCEDURE Stub_Lock(s : K; env : E) : BOOLEAN RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_Lock;
PROCEDURE Stub_UnLock(s : K; env : E) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_UnLock;
PROCEDURE Stub_WaitUnLock(s : K; env : E; lock : MUTEX) RAISES {Abort}=BEGIN RAISE Abort(env.set_raised(Except.BadType, NIL)); END Stub_WaitUnLock;

(* THE DATA-TYPE OBJECT AND IT'S METHODS. *)

REVEAL	V = PublicV BRANDED OBJECT
			type : TYP;
			value : K;
		OVERRIDES
			get_real := V_GetReal;
			get_integer := V_GetInt;
			get_string := V_GetString;
			get_bool := V_GetBool;
			get_ref := V_GetRef;
			get_turtle := V_GetTurtle;
			get_closure := V_GetClosure;
			get_field_byname := V_GetField;
			get_condition := V_GetCondition;

			get_value := V_GetValue;

			set_value := V_SetValue;

			compare := V_Compare;
			equal := V_Equal;

			printval := V_PrintVal;

			alloc_ref := V_AllocRef;

			signal := V_Signal;
			wait := V_Wait;

			send := V_Send;
			recieve := V_Recieve;

			lock := V_Lock;
			unlock := V_UnLock;
			wait_unlocked := V_WaitUnLocked;

			init := V_Init;

			get_type := V_GetType;
			get_data := V_GetData;

		END;

(* IMPLEMENTATIONS OF Node.V METHODS.
	Note that most of these simply call the corresponding functions
	of the value.
*)


PROCEDURE V_GetReal(self : V; env : E) : LONGREAL RAISES {Abort}=BEGIN RETURN self.value.get_real(env); END V_GetReal;
PROCEDURE V_GetInt(self : V; env : E) : INTEGER RAISES {Abort}=BEGIN RETURN self.value.get_integer(env); END V_GetInt;
PROCEDURE V_GetString(self : V; env : E) : TEXT RAISES {Abort}=BEGIN RETURN self.value.get_string(env); END V_GetString;
PROCEDURE V_GetBool(self : V; env : E) : BOOLEAN RAISES {Abort}=BEGIN RETURN self.value.get_bool(env); END V_GetBool;
PROCEDURE V_GetRef(self : V; env : E) : K RAISES {Abort}=BEGIN RETURN self.value.get_ref(env); END V_GetRef;
PROCEDURE V_GetTurtle(self : V; env : E) : GfxTurtle.T RAISES {Abort}=BEGIN RETURN self.value.get_turtle(env); END V_GetTurtle;
PROCEDURE V_GetClosure(self : V; env : E; VAR base_env : E; VAR params, code : T; VAR returntype : TYP) RAISES {Abort}=BEGIN self.value.get_closure(env, base_env, params, code, returntype); END V_GetClosure;
PROCEDURE V_GetField(self : V; env : E; name : TEXT) : V RAISES {Abort}=
VAR	field : INTEGER;
BEGIN
	(* Search the data-type definition to find which field
		has the name specified, then get that field from the value. *)
	field := self.type.get_field_no(env, name);
	IF field < 0 THEN
		RAISE Abort(env.set_raised(Except.NoSuchField, name));
	END;
	RETURN NEW(V).init(
			self.type.get_field_type(env, field),
			self.value.get_field(env, field)
			);
END V_GetField;
PROCEDURE V_GetCondition(self : V; env : E) : Thread.Condition RAISES {Abort}=BEGIN RETURN self.value.get_condition(env); END V_GetCondition;

PROCEDURE V_GetValue(self : V; <* UNUSED *> env : E) : V=
BEGIN
	RETURN self;
END V_GetValue;

PROCEDURE V_SetValue(self : V; env : E; val : V) RAISES {Abort}=
BEGIN

(*
IO.Put(self.type.get_typecode(env) & "<\n");
IO.Put(val.get_typecode(env) & ">\n");
*)

	(* Check the type-codes are the same! *)
	IF TypeSys.Equivalent(self.type, val.type, env) THEN
		(* Perform the assignment. *)
		self.value.set_value(env, val.value);

	ELSE
		(* Raise the type-mismatch exception. *)
		RAISE Abort(env.set_raised(Except.BadType,
					self.type.print("") & " # " & val.type.print("")
					)
				);
	END;

END V_SetValue;

PROCEDURE V_Compare(self : V; env : E; val : V) : Result RAISES {Abort}=
BEGIN

	(* Check the type-codes are the same! *)
	IF TypeSys.Equivalent(self.type, val.type, env) THEN
		(* Perform the comparison. *)
		RETURN self.value.compare(env, val.value);
	ELSE
		(* Raise the type-mismatch exception. *)
		RAISE Abort(env.set_raised(Except.BadType,
					self.type.print("") & " # " & val.type.print("")
					)
				);
	END;

END V_Compare;

PROCEDURE V_Equal(self : V; env : E; val : V) : BOOLEAN RAISES {Abort}=
BEGIN

	(* Check the type-codes are the same! *)
	IF TypeSys.Equivalent(self.type, val.type, env) THEN
		(* Check for equality. *)
		RETURN self.value.equal(env, val.value);
	ELSE
		(* Raise the type-mismatch exception. *)
		RAISE Abort(env.set_raised(Except.BadType,
					self.type.print("") & " # " & val.type.print("")
					)
				);
	END;

END V_Equal;

PROCEDURE V_PrintVal(self : V; env : E) : TEXT RAISES {Abort}=BEGIN RETURN self.value.printval(env); END V_PrintVal;

PROCEDURE V_AllocRef(self : V; env : E) RAISES {Abort}=
BEGIN
	self.value.alloc_ref(env, self.type);
END V_AllocRef;

PROCEDURE V_Signal(self : V; env : E) RAISES {Abort}=BEGIN self.value.signal(env); END V_Signal;
PROCEDURE V_Wait(self : V; env : E; lock : MUTEX) RAISES {Abort}=BEGIN self.value.wait(env, lock); END V_Wait;

PROCEDURE V_Send(s : V; env : E; val : V; lock : MUTEX) RAISES {Abort}=BEGIN s.value.send(env, val, lock); END V_Send;
PROCEDURE V_Recieve(s : V; env : E; lock : MUTEX) : V RAISES {Abort}=BEGIN RETURN s.value.recieve(env, lock); END V_Recieve;

PROCEDURE V_Lock(self : V; env : E) : BOOLEAN RAISES {Abort}=BEGIN RETURN self.value.lock(env); END V_Lock;
PROCEDURE V_UnLock(self : V; env : E) RAISES {Abort}=BEGIN self.value.unlock(env); END V_UnLock;
PROCEDURE V_WaitUnLocked(self : V; env : E; lock : MUTEX) RAISES {Abort}=BEGIN self.value.wait_unlocked(env, lock); END V_WaitUnLocked;

PROCEDURE V_Init(self : V; type : TYP; val : K) : V=
BEGIN
	self.type := type;
	self.value := val;
	RETURN self;
END V_Init;

PROCEDURE V_GetType(self : V) : TYP=
BEGIN
	RETURN self.type;
END V_GetType;

PROCEDURE V_GetData(self : V) : K=
BEGIN
	RETURN self.value;
END V_GetData;

(* OBJECTS FOR THE ENVIRONMENT SUB-SYSTEM. *)

REVEAL	E = PublicEnv BRANDED OBJECT

			(* Lock for protecting the frame from concurrency. *)
			lock : MUTEX;

			(* Parent stack frame, to be used when function returns. *)
			parent : E := NIL;
			(* Last visible stack frame, for variable inheritance. *)
			chain : E := NIL;

			(* A list of the available exceptions, including "raised" info. *)
			exceptions : ExceptList := NIL;

			(* A list of the variables in this frame, including functions. *)
			variables : VarList := NIL;

			(* Pointer to the currently selected turtle. *)
			turtle : GfxTurtle.T := NIL;
			(* Flag to indicate whether or not this frame contains a turtle. *)
			has_turtle : BOOLEAN := FALSE;

			(* Return value for function calls. *)
			return_value : V := NIL;
			(* Flag to indicate whether or not this frame has a returnval. *)
			has_return : BOOLEAN;

			(* Linked list of types. *)
			types : TypeList := NIL;

			(* Flag indicating that this is a thread-base frame. *)
			thread_base := FALSE;

		OVERRIDES
			make_child := MakeChild;
			make_base := MakeBase;
			get_parent := GetParent;

			add_variables := AddVars;
			get_var := GetVar;

			add_turtle := AddTurtle;
			get_turtle := GetTurtle;
			set_turtle := SetTurtle;

			add_exceptions := AddExcepts;
			set_raised := Raise;

			add_return := AddReturnType;
			set_return := SetReturnVal;
			get_return := GetReturnVal;

			add_types := NewUserTypes;
			get_type := GetUserType;

			make_thread_base := MakeThreadBase;
			get_thread_base := GetThreadBase;

		END;

	(* Variable list structure for handing parameters to add_variables. *)
	VarList = VarListP BRANDED OBJECT
			next : VarList := NIL;
			name : TEXT := NIL;
			value : V := NIL;
		OVERRIDES
			init := InitVarList;
			setnext := SetNextVar;
			getname := GetVarName;
			getvalue := GetVarValue;
			getnext := GetNextVar;
		END;

	(* Exception list structure.  Also stores info on whether raised or not. *)
	ExceptList = ExceptListP BRANDED OBJECT
			next : ExceptList := NIL;
			name : TEXT := NIL;
			info : TEXT := NIL;
		OVERRIDES
			init := InitExceptList;
			setnext := SetNextExceptList;
			getname := GetExceptName;
			getnext := GetNextExcept;
			getinfo := GetExceptInfo;
			union := GetExceptUnion;
			unraise := UnraiseExcept;
			is_raised := IsExceptRaised;
			find := FindExcept;
		END;

	(* TYPE LIST CONSTRUCTOR OBJECT. *)
	TypeList = TypeListP BRANDED OBJECT
			next : TypeList;
			name : TEXT;
			code : TYP;
		OVERRIDES
			init := InitTypeList;
			setnext := SetNextTypeList;
			getname := GetTypeName;
			getnext := GetTypeNext;
			gettype := GetTypeType;
			resolve := ResolveType;
		END;

(* ROUTINES IMPLEMENTING ENVIRONMENT METHODS. *)

PROCEDURE MakeChild(self : E; linkto : E) : E=
VAR	env := NEW(E);
BEGIN

	(* MakeChild allows a new environment to be easily created
		and linked to en existing one. *)

	(* Link the new env back to it's parents. *)
	env.parent := self;
	IF linkto # NIL THEN
		env.chain := linkto;
	ELSE
		env.chain := self;
	END;

	RETURN env;

END MakeChild;

PROCEDURE MakeBase(self : E; linkto : E) : E=
VAR	env := NEW(E);
BEGIN

	env := self.make_child(linkto);

	(* Base frames also contain turtle entries. *)
	env.add_turtle();

	RETURN env;

END MakeBase;

PROCEDURE GetParent(self : E) : E=
BEGIN
	RETURN self.parent;
END GetParent;

PROCEDURE AddVars(self : E; list : VarList) : E=
	PROCEDURE AddToList(list : VarList; new : VarList) : VarList=
	BEGIN
		IF list = NIL THEN
			RETURN new;
		ELSE
			list.next := AddToList(list.next, new);
			RETURN list;
		END;
	END AddToList;
BEGIN

	(* AddVars adds variables to the current stack frame! *)

	(* Add the variables to it. *)
	self.variables := AddToList(list, self.variables);

	RETURN self;

END AddVars;

PROCEDURE AddExcepts(self : E; list : ExceptList) : E=
	PROCEDURE AddLists(a, b : ExceptList) : ExceptList=
	BEGIN
		IF a = NIL THEN
			RETURN b;
		ELSE
			a.next := AddLists(a.next, b);
			RETURN a;
		END;
	END AddLists;
BEGIN

	(* Add the exceptions to this frame. *)
	self.exceptions := AddLists(list, self.exceptions);

	RETURN self;

END AddExcepts;

PROCEDURE AddTurtle(self : E)=
BEGIN

	(* AddTurtle sets the has_turtle flag in the environment frame. *)
	self.has_turtle := TRUE;

END AddTurtle;

PROCEDURE GetTurtle(self : E) : GfxTurtle.T RAISES {Abort}=
BEGIN

	(* GetTurtle returns the turtle currently selected, or creates a new temporary
		one in the current frame. *)

	(* IMPORTANT : Turtles are inherited from previous stack frames by
		PARENT, not by CHAIN.  CHAIN determines variable scope, but
		the scope of the current turtle works differently. *)

	IF self.has_turtle THEN

		(* This frame has a turtle, so check whether it's been set. *)
		IF self.turtle = NIL THEN

			(* Generate a dummy turtle from the parent frame. *)
			IF self.parent = NIL THEN
				(* There is no parent frame, so generate the default turtle. *)
				self.turtle := GfxTurtle.New();
			ELSE
				(* Copy the parent frame's turtle. *)
				self.turtle := self.parent.get_turtle().duplicate(NEW(GfxTurtle.T));
			END;

		END;
		RETURN self.turtle;

	ELSE

		(* This is not a turtle-holding frame, so skip it. *)
		IF self.parent = NIL THEN
			RAISE Abort(self.set_raised(Except.NoTurtle, NIL));
		END;
		RETURN self.parent.get_turtle();

	END;

END GetTurtle;

PROCEDURE SetTurtle(self : E; turtle : GfxTurtle.T) RAISES {Abort}=
BEGIN

	(* SetTurtle tries to set the turtle for the current frame. *)
	
	(* Does this frame hold a turtle? *)
	IF self.has_turtle THEN
		(* Yes, so set it. *)
		self.turtle := turtle;
	ELSE
		(* Check the previous frame. *)
		IF self.chain = NIL THEN
			RAISE Abort(self.set_raised(Except.NoTurtle, NIL));
		ELSE
			(* Try to set it in the previous frame. *)
			self.chain.set_turtle(turtle);
		END;
	END;

END SetTurtle;

PROCEDURE GetVar(self : E; name : TEXT) : V RAISES {Abort}=
VAR	pos : VarList;
BEGIN

	(* GetVar searches through the environment chain, looking for the
		named variable. *)

	(* Search the variable list of the current frame. *)
	pos := self.variables;
	WHILE pos # NIL DO

		IF Text.Equal(pos.name, name) THEN
			RETURN pos.value;
		END;
		pos := pos.next;

	END;

	(* Variable wasn't found in the current frame, so check the chain. *)
	IF self.chain = NIL THEN
		RAISE Abort(self.set_raised(Except.NoSuchVar, name));
	ELSE
		RETURN self.chain.get_var(name);
	END;

END GetVar;

PROCEDURE Raise(self : E; name : TEXT; info : TEXT) : ExceptList RAISES {Abort}=
VAR	pos : ExceptList;
BEGIN

	(* This function checks that an exception exists, then raises it. *)

	(* Search the exception list of the current frame. *)
	pos := self.exceptions;
	WHILE pos # NIL DO

		IF Text.Equal(pos.name, name) THEN
			(* Return an ExceptList node. *)
			RETURN NEW(ExceptList,
					name := name,
					info := info,
					next := NIL
				);
		END;
		pos := pos.next;

	END;

	(* Couldn't find the exception, so check chained frames. *)
	IF self.chain = NIL THEN

		(* There is no such exception. *)
		IF info # NIL THEN
			RAISE Abort(
				self.set_raised(Except.NoSuchExcept, name & "{" & info & "}")
				);
		ELSE
			RAISE Abort(
				self.set_raised(Except.NoSuchExcept, name)
				);
		END;

	ELSE
		(* Try the chain... *)
		RETURN self.chain.set_raised(name, info);
	END;

END Raise;

(* FUNCTIONS FOR DEALING WITH RETURN-VALUES. *)

(* To set the type of a return-value. *)

PROCEDURE AddReturnType(self : E; type : TYP) RAISES {Abort}=
BEGIN

	self.has_return := TRUE;
	IF type = NIL THEN self.return_value := NIL; RETURN; END;

	self.return_value := type.allocate_v(self);

END AddReturnType;

(* To set the return-value. *)

PROCEDURE SetReturnVal(self : E; val : V) RAISES {Abort}=
VAR	frame := self;
BEGIN

	WHILE frame # NIL DO

		IF frame.has_return THEN

			(* If return_value is NIL then no return value was expected. *)
			IF frame.return_value = NIL THEN
				RAISE Abort(self.set_raised(Except.BadType, "setreturnval"));
			END;
			frame.return_value.set_value(self, val);
			RETURN;

		END;

		(* Check the next frame. *)
		frame := frame.chain;

	END;

	(* No return value was expected. *)
	RAISE Abort(self.set_raised(Except.BadType, "noreturnval"));

END SetReturnVal;

(* To get the return-value. *)

PROCEDURE GetReturnVal(self : E) : V RAISES {Abort}=
VAR	frame := self;
BEGIN

	WHILE frame # NIL DO

		IF frame.has_return THEN
			RETURN frame.return_value;
		END;

		frame := frame.chain;
	END;

	(* No return value could be found. *)
	RAISE Abort(self.set_raised(Except.BadType, "getreturnval"));

END GetReturnVal;

(* FUNCTIONS FOR DEALING WITH USER-DEFINED TYPES: *)

(* To add a new user type. *)

PROCEDURE NewUserTypes(self : E; list : TypeList) : E RAISES {Abort}=

	PROCEDURE AddTypes(a, b : TypeList) : TypeList=
	BEGIN
		IF a = NIL THEN
				RETURN b;
		ELSE
				a.next := AddTypes(a.next, b);
				RETURN a;
		END;
	END AddTypes;

BEGIN

	(* Now add the new types. *)
	self.types := AddTypes(list, self.types);

(*
	(* Now be a sod and make sure all the types have been resolved. *)
	WHILE list # NIL DO
		list.resolve(self);
		list := list.getnext();
	END;
*)

	RETURN self;

END NewUserTypes;

(* To get a user type. *)

PROCEDURE GetUserType(self : E; name : TEXT) : TYP RAISES {Abort}=
VAR	list : TypeList;
BEGIN

	(* Search this frame's type list. *)
	list := self.types;
	WHILE list # NIL DO
		IF Text.Equal(list.name, name) THEN
			RETURN list.code;
		END;
		list := list.next;
	END;

	(* Not found.  Check the chain frame. *)
	IF self.chain # NIL THEN
		RETURN self.chain.get_type(name);
	END;

	(* The type was not found.  Raise an exception! *)
	RAISE Abort(self.set_raised(Except.NoSuchType, name));

END GetUserType;

(* FUNCTIONS FOR HANDLING THREAD BASE FRAMES. *)

(* Function to make the current frame a thread base. *)

PROCEDURE MakeThreadBase(self : E)=
BEGIN
	self.thread_base := TRUE;
END MakeThreadBase;

(* Function to return the base frame from the current environment. *)

PROCEDURE GetThreadBase(self : E) : E=
BEGIN
	IF self.thread_base THEN
		RETURN self;
	ELSE
		(* This assert should NEVER fail, because the top frame is a base. *)
		<* ASSERT self.parent # NIL *>

		(* Try the next frame down. *)
		RETURN self.parent.get_thread_base();
	END;
END GetThreadBase;

(* VarList METHOD IMPLEMENTATIONS. *)

(* Initialise a node with a name and value. *)

PROCEDURE InitVarList(self : VarList; name : TEXT; val : V) : VarList=
BEGIN
	self.name := name;
	self.value := val;
	RETURN self;
END InitVarList;

(* Set the next field to put this node into a chain. *)

PROCEDURE SetNextVar(self : VarList; next : VarList) : VarList=
BEGIN
	self.next := next;
	RETURN self;
END SetNextVar;

(* Get the name of a VarList node. *)

PROCEDURE GetVarName(self : VarList) : TEXT=
BEGIN
	RETURN self.name;
END GetVarName;

(* Get the Node.T value for a VarList node. *)

PROCEDURE GetVarValue(self : VarList) : V=
BEGIN
	RETURN self.value;
END GetVarValue;

(* GetNextVar gets the next variable in a varlist. *)

PROCEDURE GetNextVar(self : VarList) : VarList=
BEGIN
	RETURN self.next;
END GetNextVar;

(* ExceptList METHOD IMPLEMENTATIONS. *)

(* InitExceptList creates an exception list node with the desired name in it. *)

PROCEDURE InitExceptList(self : ExceptList; name : TEXT) : ExceptList=
BEGIN
	self.name := name;
	RETURN self;
END InitExceptList;

(* SetNextExceptList allows ExceptList nodes to be chained into lists. *)

PROCEDURE SetNextExceptList(self : ExceptList; next : ExceptList) : ExceptList=
BEGIN
	self.next := next;
	RETURN self;
END SetNextExceptList;

(* GetExceptName returns the name from an ExceptList node. *)

PROCEDURE GetExceptName(self : ExceptList) : TEXT=
BEGIN
	RETURN self.name;
END GetExceptName;

(* GetNextExcept returns the next exception. *)

PROCEDURE GetNextExcept(self : ExceptList) : ExceptList=
BEGIN
	RETURN self.next;
END GetNextExcept;

(* GetExceptInfo returns the information string for an exception. *)

PROCEDURE GetExceptInfo(self : ExceptList) : TEXT=
BEGIN
	RETURN self.info;
END GetExceptInfo;

(* GetExceptUnion returns the union of two exception lists. *)

PROCEDURE GetExceptUnion(self : ExceptList; with : ExceptList) : ExceptList=
VAR	pos : ExceptList;
BEGIN

	(* Work through, adding the exceptions. *)
	WHILE with # NIL DO

		(* Add this exception to the list, then move onto the next. *)
		self := SetExceptRaised(self, with.name, with.info);
		with := with.next;

	END;

	RETURN self;

END GetExceptUnion;

PROCEDURE SetExceptRaised(self : ExceptList; name, info : TEXT) : ExceptList=
BEGIN

	(* Did we reach the end of the list? *)
	IF self = NIL THEN
		RETURN NEW(ExceptList, name := name, info := info, next := NIL);
	ELSE

		IF Text.Equal(self.name, name) THEN

			(* Exceptions are the same => share them. *)
			IF info # NIL THEN
				IF self.info = NIL THEN
					self.info := info;
				ELSE
					self.info := self.info & ", " & info;
				END;
			END;
			RETURN self;

		ELSE

			(* Not the same exception, try the next one... *)
			self.next := SetExceptRaised(self.next, name, info);
			RETURN self;

		END;

	END;

END SetExceptRaised;

(* UnraiseExcept removes an exception from an ExceptList. *)

PROCEDURE UnraiseExcept(self : ExceptList; name : TEXT) : ExceptList=
BEGIN

	(* Is there any list to remove from? *)
	IF self = NIL THEN
		RETURN NIL;
	ELSE

		(* Is this the exception to unraise? *)
		IF Text.Equal(self.name, name) THEN
			RETURN self.next;
		ELSE
			self.next := UnraiseExcept(self.next, name);
			RETURN self;
		END;

	END;

END UnraiseExcept;

(* IsExceptRaised checks for an exception in an ExceptList. *)

PROCEDURE IsExceptRaised(self : ExceptList; name : TEXT) : BOOLEAN=
BEGIN
	RETURN FindExcept(self, name) # NIL;
END IsExceptRaised;

(* Find returns an exception from an ExceptList. *)

PROCEDURE FindExcept(self : ExceptList; name : TEXT) : ExceptList=
BEGIN

	IF self = NIL THEN
		RETURN NIL;
	ELSE
		IF Text.Equal(self.name, name) THEN
			RETURN self;
		ELSE
			RETURN FindExcept(self.next, name);
		END;
	END;

END FindExcept;

(* TYPE-LIST METHODS. *)

(* Initialisation procedure for TypeList nodes. *)

PROCEDURE InitTypeList(self : TypeList; name : TEXT; type : TYP) : TypeList=
BEGIN
	self.name := name;
	self.code := type;
	RETURN self;
END InitTypeList;

PROCEDURE SetNextTypeList(self : TypeList; next : TypeList) : TypeList=
BEGIN
	self.next := next;
	RETURN self;
END SetNextTypeList;

PROCEDURE GetTypeName(self : TypeList) : TEXT=
BEGIN
	RETURN self.name;
END GetTypeName;

PROCEDURE GetTypeNext(self : TypeList) : TypeList=
BEGIN
	RETURN self.next;
END GetTypeNext;

PROCEDURE GetTypeType(self : TypeList) : TYP=
BEGIN
	RETURN self.code;
END GetTypeType;

PROCEDURE ResolveType(self : TypeList; env : E) RAISES {Abort}=
BEGIN
	self.code := self.code.resolve(env);
END ResolveType;

(* EXTERNALLY AVAILABLE MODULE PROCEDURES. *)

(* Function to create a base environment. *)

PROCEDURE NewBaseEnv() : E=
VAR	env : E;
BEGIN

	env := NEW(E);

	(* Add in all the system exceptions. *)
	env.exceptions := Except.System();

	(* Set the base frame to be a turtle-holding frame. *)
	env.add_turtle();

	(* Make it a thread-base frame. *)
	env.make_thread_base();

	RETURN env;

END NewBaseEnv;

(* Function to print out a tree of commands. *)

PROCEDURE ToText(tree : T; indent : TEXT) : TEXT=
VAR text : TEXT;
BEGIN

	text := NIL;
	WHILE tree # NIL DO
		IF text = NIL THEN
			text := "";
		ELSE
			text := text & ";\n" & indent;
		END;
		text := text & tree.print(indent);
		tree := tree.next;
	END;

	RETURN text;

END ToText;

BEGIN
END Node.