(* Number module.
	Implements the numeric data types for Armadillo.
*)

(*
	12/12/96	Created.
	13/12/96	Added the NumberT allocator object.
	10/01/97	Added the get_typecode method.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
*)

MODULE Number;

(* ** DEBUGGING ** *)
(* IMPORT IO; *)

IMPORT Fmt, LongrealType;

IMPORT Node, TypeSys;

IMPORT Code, Tags;

(* The Real constant type. *)
REVEAL	Real = Node.K BRANDED OBJECT
			value : LONGREAL;
		OVERRIDES
			get_real := Real_GetReal;
			get_integer := Real_GetInt;
			print := Real_Print;
			printval := Real_PrintVal;
			compare := Real_Compare;
			equal := Real_Equal;
		END;

PROCEDURE Real_GetReal(self : Real; env : Node.E) : LONGREAL=
BEGIN
	RETURN self.value;
END Real_GetReal;

PROCEDURE Real_GetInt(self : Real; env : Node.E) : INTEGER=
BEGIN
	RETURN ROUND(self.value);
END Real_GetInt;

PROCEDURE Real_Print(self : Real; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN Fmt.LongReal(self.value);
END Real_Print;

PROCEDURE Real_PrintVal(self : Real; env : Node.E) : TEXT=
BEGIN
	RETURN Fmt.LongReal(self.value);
END Real_PrintVal;

PROCEDURE Real_Compare(self : Real; env : Node.E; val : Node.K) : Node.Result RAISES {Node.Abort}=
BEGIN
	RETURN LongrealType.Compare(self.value, val.get_real(env));
END Real_Compare;

PROCEDURE Real_Equal(self : Real; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN LongrealType.Compare(self.value, val.get_real(env)) = 0;
END Real_Equal;

(* The NumberV data type is a sub-type of Real, with methods
	implemented for setting as well as reading the value. *)
REVEAL	NumberV = Real BRANDED OBJECT
		OVERRIDES
			set_value := NumV_SetValue;
		END;

PROCEDURE NumV_SetValue(self : NumberV; env : Node.E; val : Node.K) RAISES {Node.Abort}=
BEGIN
	self.value := val.get_real(env);
END NumV_SetValue;

(* The Int constant type. *)
REVEAL	Int = Node.K BRANDED OBJECT
			value : INTEGER;
		OVERRIDES
			get_real := Int_GetReal;
			get_integer := Int_GetInt;
			print := Int_Print;
			printval := Int_PrintVal;
			compare := Int_Compare;
			equal := Int_Equal;
		END;

PROCEDURE Int_GetReal(self : Int; env : Node.E) : LONGREAL=
BEGIN
	RETURN FLOAT(self.value, LONGREAL);
END Int_GetReal;

PROCEDURE Int_GetInt(self : Int; env : Node.E) : INTEGER=
BEGIN
	RETURN self.value;
END Int_GetInt;

PROCEDURE Int_Print(self : Int; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN Fmt.Int(self.value);
END Int_Print;

PROCEDURE Int_PrintVal(self : Int; env : Node.E) : TEXT=
BEGIN
	RETURN Fmt.Int(self.value);
END Int_PrintVal;

PROCEDURE Int_Compare(self : Int; env : Node.E; val : Node.K) : Node.Result RAISES {Node.Abort}=
BEGIN
	RETURN LongrealType.Compare(FLOAT(self.value, LONGREAL), val.get_real(env));
END Int_Compare;

PROCEDURE Int_Equal(self : Int; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN LongrealType.Compare(FLOAT(self.value, LONGREAL), val.get_real(env)) = 0;
END Int_Equal;

(* NumberT is the number type-object, providing a method for allocating number storage. *)
REVEAL	NumberT = Node.TYP BRANDED OBJECT
		OVERRIDES
			allocate := NumberT_Allocate;
			check_type := NumberT_CheckType;
			print := NumberT_Print;
		END;

PROCEDURE NumberT_Allocate(self : NumberT; env : Node.E) : Node.K=
BEGIN
	RETURN NEW(NumberV);
END NumberT_Allocate;

PROCEDURE NumberT_CheckType(<* UNUSED *> s : NumberT; <* UNUSED *> env : Node.E; i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Numbers cannot be recursive. *)
	RETURN TRUE;
END NumberT_CheckType;

PROCEDURE NumberT_Print(self : NumberT; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "number";
END NumberT_Print;

(* Routine to read in the above types. *)
PROCEDURE Convert(tree : Code.T) : Node.T=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TAG_Real=>
			RETURN NEW(Node.V).init(NumberType,
					NEW(Real,
						value := NARROW(tree.extra, Code.RealR).value
					)
				);
	|	Tags.T.TAG_Int=>
			RETURN NEW(Node.V).init(NumberType,
					NEW(Int,
						value := NARROW(tree.extra, Code.IntR).value
					)
				);
	|	Tags.T.TYP_Number=>
			RETURN NumberType;
	ELSE
		RETURN NIL;
	END;

END Convert;

(* Routine to allocate a new number. *)
PROCEDURE NewNumber(value : LONGREAL) : Node.V=
BEGIN
	RETURN NEW(Node.V).init(NumberType, NEW(Real, value := value));
END NewNumber;

BEGIN
	NumberType := NEW(NumberT, code := TypeSys.NewTypeCode());
END Number.