(* Ports module.
	Implements the communication port data-type.
*)

(*
	03/02/96	Created.
*)

MODULE Ports;

(* IMPORT IO; *)

IMPORT Thread;

IMPORT Code, Tags;

IMPORT Node, Loader, TypeSys, Except;

(* The port data-type. *)
REVEAL	Port = Node.K BRANDED OBJECT
			empty, full, ack : Thread.Condition;
			data : Node.K := NIL;
			type : Node.TYP := NIL;
		OVERRIDES
			send := Port_Send;
			recieve := Port_Recieve;
			print := Port_Print;
		END;

PROCEDURE Port_Send(self : Port; env : Node.E; val : Node.V; lock : MUTEX) RAISES {Node.Abort}=
BEGIN

	(* ** NOTE : "lock" MUST BE SUCH THAT IT ENSURES THE SECURITY OF THIS
		DATA-STRUCTURE! *)

	(* Check that the type of the port is the same as the type of the data being sent. *)
	IF NOT(TypeSys.Equivalent(val.get_type(), self.type, env)) THEN
		RAISE Node.Abort(
				env.set_raised(Except.BadType, val.get_type().print("\t\t") &
						" # " & self.type.print("\t\t")
				)
			);
	END;

	(* Is the port already full? *)
	WHILE self.data # NIL DO
		(* Wait for it to be signalled empty. *)
		Thread.Wait(lock, self.empty);
	END;

	(* Here, we should be guaranteed that the port is empty! *)

	(* Leave the data in the port. *)
	self.data := val.get_data();

	(* Signal to wake up any threads that are waiting to recieve from this port. *)
	Thread.Signal(self.full);

	(* Wait for the data to be recieved. *)
	Thread.Wait(lock, self.ack);

	(* The recieving thread will clear the port. *)

END Port_Send;

PROCEDURE Port_Recieve(self : Port; env : Node.E; lock : Thread.Mutex) : Node.V RAISES {Node.Abort}=
VAR	retval : Node.V;
BEGIN

	(* ** NOTE : "lock" MUST BE SUCH THAT IT ENSURES THE SECURITY OF THIS
		DATA-STRUCTURE! *)

	(* Is the port empty? *)
	WHILE self.data = NIL DO
		(* Wait for it to be signalled full. *)
		Thread.Wait(lock, self.full);
	END;

	(* Here, we should be guaranteed that the port is full! *)

	(* Make a new return value. *)
	retval := NEW(Node.V).init(self.type, self.data);
	(* Clear out the port. *)
	self.data := NIL;
	(* Signal to acknowledge that the data sent to this port was recieved. *)
	Thread.Signal(self.ack);
	(* Signal to wake up any threads that are waiting to send to this port. *)
	Thread.Signal(self.empty);

	(* Return the data value created. *)
	RETURN retval;

END Port_Recieve;

PROCEDURE Port_Print(<* UNUSED *> s : Port; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN "port";
END Port_Print;

(* Port datatype allocator object. *)
REVEAL	PortT = Node.TYP BRANDED OBJECT
			type : Node.TYP;
			resolved := TRUE;
		OVERRIDES
			allocate := PortT_Allocate;
			check_type := PortT_CheckType;
			resolve := PortT_Resolve;
			print := PortT_Print;
		END;

PROCEDURE PortT_Allocate(self : PortT; <* UNUSED *> env : Node.E) : Node.K=
BEGIN
	RETURN NEW(Port, type := self.type,
			full := NEW(Thread.Condition),
			empty := NEW(Thread.Condition),
			ack := NEW(Thread.Condition)
		);
END PortT_Allocate;

PROCEDURE PortT_CheckType(<* UNUSED *> s : PortT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* Condition variables cannot be recursive. *)
	RETURN TRUE;
END PortT_CheckType;

PROCEDURE PortT_Resolve(s : PortT; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	IF NOT(s.resolved) THEN
		s.resolved := TRUE;
		s.type := s.type.resolve(env);
	END;
	RETURN s;
END PortT_Resolve;

PROCEDURE PortT_Print(self : PortT; i : TEXT) : TEXT=
BEGIN
	RETURN "... port";
END PortT_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Port=>
			RETURN NEW(PortT,
					type := Loader.Convert(tree.params.param),
					code := TypeSys.NewTypeCode()
				);
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Ports.