(* Record module.
	Implements methods for handling the record data-type.
*)

(*
	19/12/96	Created.
	10/01/97	Added the get_typecode method.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
*)

MODULE Record;

(* ** DEBUGGING ** *)
(* IMPORT IO, Fmt; *)

IMPORT Code, Tags;

IMPORT Node, Loader, Except, TypeSys;

(* The main record data-type. *)
REVEAL	Record = Node.K BRANDED OBJECT
			fields : Node.Fields := NIL;
		OVERRIDES
			get_field := Record_GetField;
			set_value := Record_SetValue;
			printval := Record_PrintVal;
			equal := Record_Equal;
		END;

PROCEDURE Record_GetField(self : Record; env : Node.E; n : CARDINAL) : Node.K RAISES {Node.Abort}=
VAR	ptr := self.fields;
	count : CARDINAL := 0;
BEGIN

	(* Search through the fields, looking for a match. *)
	WHILE (ptr # NIL) DO

		IF count = n THEN
			RETURN ptr.value;
		END;
		ptr := ptr.next;
		INC(count);

	END;

	(* Couldn't find a match, so fail! *)
	RAISE Node.Abort(env.set_raised(Except.NoSuchField, NIL));

END Record_GetField;

PROCEDURE Record_SetValue(self : Record; env : Node.E; val : Node.K) RAISES {Node.Abort}=
VAR	list1, list2 : Node.Fields;
BEGIN

	IF NOT(ISTYPE(val, Record)) THEN
		RAISE Node.Abort(env.set_raised(Except.BadType, NIL));
	END;

	(* Get the list of fields for the record value. *)
	list1 := NARROW(val, Record).fields;

	(* Work through the fields, assigning them. *)
	list2 := self.fields;	

	WHILE list1 # NIL DO
		IF list2 = NIL THEN
			(* This record has too few fields to have the
				value assigned to it! *)
			RAISE Node.Abort(env.set_raised(Except.BadType, NIL));
		END;

		(* Perform the field assignment. *)
		list2.value.set_value(env, list1.value);

		list2 := list2.next;
		list1 := list1.next;
	END;

	IF list2 # NIL THEN
		(* There are still more fields to be filled in, but
			the value given doesn't have any more! *)
		RAISE Node.Abort(env.set_raised(Except.BadType, NIL));
	END;

END Record_SetValue;

PROCEDURE Record_PrintVal(self : Record; env : Node.E) : TEXT RAISES {Node.Abort}=
VAR	text : TEXT := NIL;
	list := self.fields;
BEGIN

	WHILE list # NIL DO
		IF text = NIL THEN
			text := "{";
		ELSE
			text := text & ", ";
		END;

		text := text & list.value.printval(env);
		list := list.next;
	END;
	text := text & "}";

	RETURN text;

END Record_PrintVal;

PROCEDURE Record_Equal(self : Record; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	(* ** NOT YET IMPLEMENTED ** *)
	RAISE Node.Abort(env.set_raised(Except.Unsupported, "record.equal"));
END Record_Equal;

(* Record datatype allocator object. *)
REVEAL	RecordT = Node.TYP BRANDED OBJECT
			finite := FALSE;
			fields : Node.TYP;
			resolved := FALSE;
		OVERRIDES
			allocate := RecordT_Allocate;
			get_field_no := RecordT_GetFieldNo;
			get_field_type := RecordT_GetFieldType;
			get_varlist := RecordT_GetVarList;
			check_type := RecordT_CheckType;
			resolve := RecordT_Resolve;
			print := RecordT_Print;
		END;

PROCEDURE RecordT_Allocate(self : RecordT; env : Node.E) : Node.K RAISES {Node.Abort}=
BEGIN
	RETURN NEW(Record, fields := self.fields.get_fields(env));
END RecordT_Allocate;

PROCEDURE RecordT_GetFieldNo(self : RecordT; env : Node.E; name : TEXT) : INTEGER RAISES {Node.Abort}=
BEGIN
	RETURN self.fields.get_field_no(env, name);
END RecordT_GetFieldNo;

PROCEDURE RecordT_GetFieldType(self : RecordT; env : Node.E; n : CARDINAL) : Node.TYP RAISES {Node.Abort}=
BEGIN
	RETURN self.fields.get_field_type(env, n);
END RecordT_GetFieldType;

PROCEDURE RecordT_GetVarList(self : RecordT; env : Node.E; list : Node.VarList) : Node.VarList RAISES {Node.Abort}=
BEGIN
	(* This routine is used when checking structural equivalence by TypeSys. *)
	RETURN self.fields.get_varlist(env, list);
END RecordT_GetVarList;

PROCEDURE RecordT_CheckType(self : RecordT; env : Node.E; i : Node.InvalidTypes) : BOOLEAN RAISES {Node.Abort}=
VAR	list : Node.InvalidTypes;
BEGIN

	(* If this type is finite then just return true. *)
	IF self.finite THEN RETURN TRUE; END;

	(* Check the invalid typecode list for this typecode. *)
	list := i;
	WHILE list # NIL DO
		IF list.typecode = self.code THEN
			RETURN FALSE;
		END;
		list := list.next;
	END;

	(* Otherwise, check the field types. *)
	RETURN self.fields.check_type(env,
				NEW(Node.InvalidTypes,
					typecode := self.code,
					next := i
					)
				);
END RecordT_CheckType;

PROCEDURE RecordT_Resolve(self : RecordT; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN

	(* If the type isn't resolved then do it! *)
	IF NOT(self.resolved) THEN
		self.resolved := TRUE;
		self.fields := self.fields.resolve(env);
	END;

	(* Return this type-node. *)
	RETURN self;

END RecordT_Resolve;

PROCEDURE RecordT_Print(self : RecordT; i : TEXT) : TEXT=
BEGIN
	RETURN "record ... end";
END RecordT_Print;

(* The record-field dereference operator. *)
REVEAL	Dot = Node.T BRANDED OBJECT
			record : Node.T;
			field : Node.T;
		OVERRIDES
			get_value := Dot_GetValue;
			print := Dot_Print;
		END;

PROCEDURE Dot_GetValue(self : Dot; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN

	(* Get the value of the thing to be dereferenced and
		find the specified field from it: *)
	RETURN self.record.get_value(env).get_field_byname(env, self.field.get_name(env)); 

END Dot_GetValue;

PROCEDURE Dot_Print(self : Dot; i : TEXT) : TEXT=
BEGIN
	RETURN self.record.print(i) & "." & self.field.print(i);
END Dot_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Record=>
			RETURN NEW(RecordT,
				fields := Loader.Convert(tree.params.param),
				code := TypeSys.NewTypeCode()
				);
	|	Tags.T.TAG_RecordRef=>
			RETURN NEW(Dot,
					record := Loader.Convert(tree.params.param),
					field := Loader.Convert(tree.params.next.param)
				);
	ELSE
		RETURN NIL;
	END;
END Convert;

BEGIN
END Record.