(* RefType module.
	Implements methods for handling the pointer data-type.
*)

(*
	19/12/96	Created.
	20/12/96	Added the "AddressOf" object for creating pointers
			to variables and fields etc.
	10/01/97	Added the get_typecode method.
	13/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
	18/01/97	Changed the "nil" constant to a command like "new".
			Added some record-related methods to the reference
			type so that the "." operator now has implicit
			dereferencing of pointers.
	20/01/97	Changed "nil" back to a constant, using type-coercion.
			Currently, nil is of the "null" type, which is a special
			case of the "ref" type.
*)

MODULE RefType;

(* IMPORT IO; *)

IMPORT Code, Tags;

IMPORT Node, Loader, Except, TypeSys;

(* The main reference data-type. *)
REVEAL	Ref = Node.K BRANDED OBJECT
			pointer : Node.K := NIL;
		OVERRIDES
			get_ref := Ref_GetRef;
			get_field := Ref_GetField;
			set_value := Ref_SetValue;
			print := Ref_Print;
			printval := Ref_PrintVal;
			equal := Ref_Equal;
			alloc_ref := Ref_AllocRef;
		END;

PROCEDURE Ref_GetRef(self : Ref; <* UNUSED *> env : Node.E) : Node.K=
BEGIN
	RETURN self.pointer;
END Ref_GetRef;

PROCEDURE Ref_GetField(self : Ref; env : Node.E; n : CARDINAL) : Node.K RAISES {Node.Abort}=
BEGIN

	(* Check that the pointer ain't NIL! *)
	IF self.pointer = NIL THEN
		RAISE Node.Abort(env.set_raised(Except.DerefNil, NIL));
	END;
	
	(* Get the field. *)
	RETURN self.pointer.get_field(env, n);

END Ref_GetField;

PROCEDURE Ref_SetValue(self : Ref; env : Node.E; val : Node.K) RAISES {Node.Abort}=
BEGIN

	(* Note that the parent Node.V will already have type-checked
		this field! *)
	self.pointer := val.get_ref(env);

END Ref_SetValue;

PROCEDURE Ref_Print(self : Ref; i : TEXT) : TEXT=
BEGIN
	IF self.pointer = NIL THEN
		RETURN "nil";
	ELSE
		RETURN "!" & self.pointer.print(i);
	END;
END Ref_Print;

PROCEDURE Ref_PrintVal(self : Ref; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	IF self.pointer = NIL THEN
		RETURN "nil";
	ELSE
		RETURN "&" & self.pointer.printval(env);
	END;
END Ref_PrintVal;

PROCEDURE Ref_Equal(self : Ref; env : Node.E; val : Node.K) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN self.pointer = val.get_ref(env);
END Ref_Equal;

PROCEDURE Ref_AllocRef(self : Ref; env : Node.E; type : Node.TYP) RAISES {Node.Abort}=
BEGIN

	(* Note that the type above is the type of this reference,
		NOT the type of the thing it points to, so first
		we have to get the type of the thing to which it
		points, before we can allocate it. *)
	self.pointer := type.type_of_ref(env).allocate(env);

END Ref_AllocRef;

(* Reference datatype allocator object. *)
REVEAL	RefT = Node.TYP BRANDED OBJECT
			type : Node.TYP;
			resolved := FALSE;
		OVERRIDES
			allocate := RefT_Allocate;
			type_of_ref := RefT_TypeOfRef;
			get_field_no := RefT_GetFieldNo;
			get_field_type := RefT_GetFieldType;
			check_type := RefT_CheckType;
			resolve := RefT_Resolve;
			print := RefT_Print;
		END;

PROCEDURE RefT_Allocate(<* UNUSED *> self : RefT; <* UNUSED *> env : Node.E) : Node.K=
BEGIN
	RETURN NEW(Ref);
END RefT_Allocate;

PROCEDURE RefT_TypeOfRef(self : RefT; <* UNUSED *> env : Node.E) : Node.TYP=
BEGIN
	RETURN self.type;
END RefT_TypeOfRef;

PROCEDURE RefT_GetFieldNo(self : RefT; env : Node.E; name : TEXT) : INTEGER RAISES {Node.Abort}=
BEGIN
	RETURN self.type.get_field_no(env, name);
END RefT_GetFieldNo;

PROCEDURE RefT_GetFieldType(self : RefT; env : Node.E; n : CARDINAL) : Node.TYP RAISES {Node.Abort}=
BEGIN
	RETURN self.type.get_field_type(env, n);
END RefT_GetFieldType;

PROCEDURE RefT_CheckType(<* UNUSED *> s : RefT; <* UNUSED *> env : Node.E; <* UNUSED *> i : Node.InvalidTypes) : BOOLEAN=
BEGIN
	(* References cannot be infinite. *)
	RETURN TRUE;
END RefT_CheckType;

PROCEDURE RefT_Resolve(s : RefT; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	IF NOT(s.resolved) THEN
		s.resolved := TRUE;
		IF s.type # NIL THEN s.type := s.type.resolve(env); END;
	END;
	RETURN s;
END RefT_Resolve;

PROCEDURE RefT_Print(self : RefT; i : TEXT) : TEXT=
BEGIN
	IF self.type # NIL THEN
		RETURN "ref ...";
	ELSE
		RETURN "null";
	END;
END RefT_Print;

(* The dereference operator. *)
REVEAL	Deref = Node.T BRANDED OBJECT
			ref : Node.T;
		OVERRIDES
			get_value := Deref_GetValue;
			print := Deref_Print;
		END;

PROCEDURE Deref_GetValue(self : Deref; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	val : Node.V;
BEGIN

	(* Perform the de-reference! *)
	val := self.ref.get_value(env); 
	IF val.get_ref(env) = NIL THEN
		RAISE Node.Abort(env.set_raised(Except.DerefNil, NIL));
	END;

	(* Construct the return-value. *)
	RETURN NEW(Node.V).init(
		val.get_type().type_of_ref(env),
		val.get_ref(env)
		);

END Deref_GetValue;

PROCEDURE Deref_Print(self : Deref; i : TEXT) : TEXT=
BEGIN
	RETURN "^" & self.ref.print(i);
END Deref_Print;

(* The "new" command is implemented here for simplicity. *)
REVEAL	New = Node.T BRANDED OBJECT
			variable : Node.T;
		OVERRIDES
			execute := New_Execute;
			print := New_Print;
		END;

PROCEDURE New_Execute(self : New; env : Node.E) RAISES {Node.Abort}=
BEGIN

	(* Reference type variables have the "alloc_ref" method implemented
		for them, to instantiate the pointer with a new data structure. *)
	self.variable.get_value(env).alloc_ref(env);

END New_Execute;

PROCEDURE New_Print(self : New; i : TEXT) : TEXT=
BEGIN
	RETURN "new " & self.variable.print(i & "\t");
END New_Print;

(* AddressOf effectively calculates the "address" of a structure. *)
REVEAL	AddressOf = Node.T BRANDED OBJECT
			varname : Node.T;
		OVERRIDES
			get_value := AddressOf_GetValue;
			print := AddressOf_Print;
		END;

PROCEDURE AddressOf_GetValue(self : AddressOf; env : Node.E) : Node.V RAISES {Node.Abort}=
VAR	value : Node.V;
BEGIN
	(* Get the value object for the varname field and construct a ref
		variable that points to it. *)
	value := self.varname.get_value(env);
	RETURN NEW(Node.V).init(
			NEW(RefT,
				type := value.get_type(),
				code := TypeSys.NewTypeCode()
			),
			NEW(Ref, pointer := value.get_data())
		);
END AddressOf_GetValue;

PROCEDURE AddressOf_Print(self : AddressOf; i : TEXT) : TEXT=
BEGIN
	RETURN "!" & self.varname.print(i);
END AddressOf_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TYP_Ref=>
			RETURN NEW(RefT,
				type := Loader.Convert(tree.params.param),
				code := TypeSys.NewTypeCode()
				);
	|	Tags.T.REF_Deref=>
			RETURN NEW(Deref, ref := Loader.Convert(tree.params.param));
	|	Tags.T.CMD_New=>
			RETURN NEW(New, variable := Loader.Convert(tree.params.param));
	|	Tags.T.TAG_Nil=>
			(* Note that EVERY instance of NIL MUST have it's own TypeCode!!! *)
			RETURN NEW(Node.V).init(
				NEW(RefT, type := NIL, code := TypeSys.NewTypeCode()),
				NEW(Ref, pointer := NIL)
				);
	|	Tags.T.REF_Pointer=>
			RETURN NEW(AddressOf, varname := Loader.Convert(tree.params.param));
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END RefType.