(* TypeSys module.
	Implements the user type definition routine.
*)

(*
	19/12/96	Created.
	10/01/97	Added the get_typecode method.
	18/01/97	Added the get_nil method to allow user reference types.
	20/01/97	Added the Equivalent procedure to test type equivalences
			with in-built type-coercion.
	25/01/97	Added the infinite-type checking routine.
	26/02/97	The type declaration command may now be used in a "block"
			form and may declare several functions at once.
*)

MODULE TypeSys;

(* ** DEBUGGING! ** *)
(* IMPORT IO; *)

IMPORT Text;

IMPORT Code, Tags;

IMPORT Node, Loader, ArmRT;

IMPORT Functions, RefType, Record, Except;

(* The "type" command. *)
REVEAL	TypeDef = Node.T BRANDED OBJECT
			typedecls : Node.T;
			code : Node.T;
			is_block : BOOLEAN;
		OVERRIDES
			execute := TypeDef_Execute;
			print := TypeDef_Print;
		END;

PROCEDURE TypeDef_Execute(self : TypeDef; env : Node.E) RAISES {Node.Abort}=
VAR	typelist : Node.TypeList;
	new_env : Node.E;
BEGIN

	(* Get the list of new types. *)
	typelist := self.typedecls.get_typelist(env);

	(* Is this the block form of the command? *)
	IF self.is_block THEN
		(* Yes.  Create a child frame. *)
		new_env := env.make_child();
	ELSE
		new_env := env;
	END;

	(* Add the new types. *)
	new_env := new_env.add_types(typelist);

	IF self.is_block THEN
		(* Execute the command in this block. *)
		ArmRT.Execute(self.code, new_env);
	END;

END TypeDef_Execute;

PROCEDURE TypeDef_Print(self : TypeDef; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := "type " & self.typedecls.print(i & "\t");
	IF self.is_block THEN
		text := text & " in " & Node.ToText(self.code, i & "\t");
	END;
	RETURN text;
END TypeDef_Print;

(* The type-name tag. *)
REVEAL	TypeName = Node.TYP BRANDED OBJECT
			name : TEXT;
		OVERRIDES
			get_name := TypeName_GetName;
			allocate := TypeName_Allocate;
			type_of_ref := TypeName_TypeOfRef;
			get_field_no := TypeName_GetFieldNo;
			get_field_type := TypeName_GetFieldType;
			get_typecode := TypeName_GetTypeCode;
			set_typecode := TypeName_SetTypeCode;
			resolve := TypeName_Resolve;
			check_type := TypeName_CheckType;
			print := TypeName_Print;
		END;

PROCEDURE TypeName_GetName(self : TypeName; <* UNUSED *> env : Node.E) : TEXT=
BEGIN
	RETURN self.name;
END TypeName_GetName;

PROCEDURE TypeName_Allocate(self : TypeName; env : Node.E) : Node.K RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).allocate(env);
END TypeName_Allocate;

PROCEDURE TypeName_TypeOfRef(self : TypeName; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).type_of_ref(env);
END TypeName_TypeOfRef;

PROCEDURE TypeName_GetFieldNo(self : TypeName; env : Node.E; name : TEXT) : INTEGER RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).get_field_no(env, name);
END TypeName_GetFieldNo;

PROCEDURE TypeName_GetFieldType(self : TypeName; env : Node.E; n : CARDINAL) : Node.TYP RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).get_field_type(env, n);
END TypeName_GetFieldType;

PROCEDURE TypeName_GetTypeCode(self : TypeName; env : Node.E) : CARDINAL RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).get_typecode(env);
END TypeName_GetTypeCode;

PROCEDURE TypeName_SetTypeCode(self : TypeName; env : Node.E; code : CARDINAL) RAISES {Node.Abort}=
BEGIN
	env.get_type(self.name).set_typecode(env, code);
END TypeName_SetTypeCode;

	(* This type is used as the nodes in the following routine. *)
TYPE Names = REF RECORD
			name : TEXT;
			next : Names := NIL;
		END;

PROCEDURE TypeName_Resolve(self : TypeName; env : Node.E) : Node.TYP RAISES {Node.Abort}=
VAR	struct : Node.TYP;
	type := self;
	new_type : TypeName;
	list, names : Names := NIL;
BEGIN

	(* If a type is defined such as Type = Type then we can't get a structure
		and must raise an exception instead. *)
	LOOP
		struct := env.get_type(type.name).resolve(env);

		IF ISTYPE(struct, TypeName) THEN

			(* Move on to this new node. *)
			new_type := NARROW(struct, TypeName);

			(* Check that the name isn't in the list. *)
			list := names;
			WHILE list # NIL DO
				IF Text.Equal(list.name, new_type.name) THEN

					(* Name was found in list, meaning an infinite type! *)
					RAISE Node.Abort(
						env.set_raised(Except.InfiniteType, self.print(""))
						);

				END;
				list := list.next;
			END;

			(* Add this type's name to the list and check the next. *)
			names := NEW(Names,
					name := type.name,
					next := names
				);
			type := new_type;

		ELSE
			RETURN struct;
		END;

	END;
	
END TypeName_Resolve;

PROCEDURE TypeName_CheckType(self : TypeName; env : Node.E; i : Node.InvalidTypes) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	RETURN env.get_type(self.name).check_type(env, i);
END TypeName_CheckType;

PROCEDURE TypeName_Print(self : TypeName; <* UNUSED *> i : TEXT) : TEXT=
BEGIN
	RETURN self.name;
END TypeName_Print;

REVEAL	TypeDecl = Node.T BRANDED OBJECT
					name : Node.T;
					type : Node.TYP;
				OVERRIDES
					get_typelist := TypeDecl_GetTypeList;
					print := TypeDecl_Print;
				END;

PROCEDURE TypeDecl_GetTypeList(self : TypeDecl; env : Node.E; list : Node.TypeList) : Node.TypeList RAISES {Node.Abort}=
VAR	node : Node.TypeList;
BEGIN

	(* Create the node for this TyepDecl. *)
	node := NEW(Node.TypeList).init(self.name.get_name(env), self.type);

	IF self.next # NIL THEN
		(* Do the next one as well. *)
		node := node.setnext(self.next.get_typelist(env));
	ELSE
		(* Add on the specified list. *)
		node := node.setnext(list);
	END;

	RETURN node;

END TypeDecl_GetTypeList;

PROCEDURE TypeDecl_Print(self : TypeDecl; i : TEXT) : TEXT=
BEGIN
	RETURN self.name.print(i) & " = " & self.type.print(i & "\t");
END TypeDecl_Print;

(* Routine to load in the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.CMD_Type=>
			token := NEW(TypeDef,
					typedecls := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, TypeDef).is_block := TRUE;
				NARROW(token, TypeDef).code := Loader.ConvertCommand(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.TAG_TypeDecl=>
			token := NEW(TypeDecl,
					name := Loader.Convert(tree.params.param),
					type := Loader.Convert(tree.params.next.param)
				);
			IF tree.params.next.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.next.param);
			END;
			RETURN token;
	|	Tags.T.TAG_Type=>
			RETURN NEW(TypeName, name := tree.extra);
	ELSE
		RETURN NIL;
	END;

END Convert;

(* Routine to check for equivalence of two types. *)
PROCEDURE Equivalent(code1, code2 : Node.TYP; env : Node.E) : BOOLEAN RAISES {Node.Abort}=
VAR	result : BOOLEAN;
BEGIN

	result := SubEquiv(code1, code2, env, NIL);
	RETURN result;

END Equivalent;

PROCEDURE SubEquiv(code1, code2 : Node.TYP; env : Node.E; assume : Assumptions) : BOOLEAN RAISES {Node.Abort}=
VAR	list : Assumptions;
	type1, type2 : CARDINAL;
	result := FALSE;
	new_assume : Assumptions;
BEGIN

	type1 := code1.get_typecode(env);
	type2 := code2.get_typecode(env);
	(* If the type-codes are the same then match them trivially. *)
	IF type1 = type2 THEN
		RETURN TRUE;
	END;

(*
IO.Put("Structure Check : " & code1.print("\t\t") & ", " & code2.print("\t\t") & "\n");
*)

	(* Check that this isn't covered by one of the assumptions. *)
	list := assume;
	WHILE list # NIL DO
		IF (list.code1 = type1) AND (list.code2 = type2) THEN
			RETURN TRUE;
		END;
		list := list.next;
	END;

	(* Otherwise, we'll have to check it out structurally. *)

	(* Ensure that the first items in the structures are in-built
		types, not type-names. *)
	code1 := code1.resolve(env);
	code2 := code2.resolve(env);

	(* Before using structural checking, we assume that the types are the same: *)
	new_assume := NEW(Assumptions,
			next := assume,
			code1 := type1,
			code2 := type2
			);

	(* See whether or not we can test the sub-structure of the type. *)
	TYPECASE code1 OF
		RefType.RefT =>

			(* The second argument must also be a reference *)
			IF NOT(ISTYPE(code2, RefType.RefT)) THEN
				RETURN FALSE;
			END;

			(* Nasty type-kludge to make nil look like a reference type. *)
			IF (code1.type_of_ref(env) = NIL) OR
				(code2.type_of_ref(env) = NIL) THEN
				result := TRUE;
			ELSE
				(* Test that the types being referenced match. *)
				result := SubEquiv(code1.type_of_ref(env), code2.type_of_ref(env), env, new_assume);
			END;

	|	Functions.ClosureT =>

			(* If the second type is not a function then fail. *)
			IF NOT(ISTYPE(code2, Functions.ClosureT)) THEN
				RETURN FALSE;
			END;

			(* To check equivalence, get the lists of fields and check
				their types! *)
			VAR	params1, params2 : Node.TYP;
				return1, return2 : Node.TYP;
				fields1, fields2 : Node.VarList;
			BEGIN

				(* Get the parameter and return types. *)
				code1.type_of_fun(env, params1, return1);
				code2.type_of_fun(env, params2, return2);

				(* Check the return values are the same. *)
				IF (return1 = NIL) AND (return2 = NIL) THEN
					result := TRUE;
				ELSE
					IF (return1 = NIL) OR (return2 = NIL) THEN
						RETURN FALSE;
					END;
					IF NOT(SubEquiv(return1, return2, env, assume)) THEN
						RETURN FALSE;
					END;
				END;

				(* If both sets of params are NIL then we're ok! *)
				IF (params1 = NIL) AND (params2 = NIL) THEN
					result := TRUE;
				ELSE
					IF (params1 = NIL) OR (params2 = NIL) THEN
						RETURN FALSE;
					END;

					(* Check that the params are the same. *)
					fields1 := params1.get_varlist(env);
					fields2 := params2.get_varlist(env);

					(* Now work through the fields, checking them. *)
					WHILE fields1 # NIL DO
						IF fields2 = NIL THEN RETURN FALSE; END;

						(* Check type of this field. *)
						IF NOT(SubEquiv(
							fields1.getvalue().get_type(),
							fields2.getvalue().get_type(),
							env,
							new_assume
							)) THEN
							RETURN FALSE;
						END;
						fields1 := fields1.getnext();
						fields2 := fields2.getnext();

					END;

					IF fields2 # NIL THEN
						RETURN FALSE;
					END;

				END;
			END;

			(* If we reach here then they were the same! *)
			result := TRUE;

	|	Record.RecordT =>

			(* If the second type is not a record then fail. *)
			IF NOT(ISTYPE(code2, Record.RecordT)) THEN
				RETURN FALSE;
			END;

			(* To check equivalence, get the lists of fields and check
				their types! *)
			VAR	fields1, fields2 : Node.VarList;
			BEGIN
				fields1 := code1.get_varlist(env);
				fields2 := code2.get_varlist(env);

				(* Now work through the fields, checking them. *)
				WHILE fields1 # NIL DO
					IF fields2 = NIL THEN RETURN FALSE; END;

					(* Check type of this field. *)
					IF NOT(SubEquiv(
						fields1.getvalue().get_type(),
						fields2.getvalue().get_type(),
						env,
						new_assume
						)) THEN
						RETURN FALSE;
					END;
					fields1 := fields1.getnext();
					fields2 := fields2.getnext();

				END;

				IF fields2 # NIL THEN
					RETURN FALSE;
				END;

			END;

			(* If we reach here then they were the same! *)
			result := TRUE;

	ELSE
		RETURN FALSE;
	END;

	(* If the types turned out to be the same by structural equivalence then
		make their type-codes the same to prevent the need for future
		confusion! *)

	IF result THEN
		(* Only change typecodes if no assumptions were made. *)
		IF assume = NIL THEN

			(* Set typecodes in both directions - the set_typecode routine
				will only let typecodes get lower, like a valve. *)
			code2.set_typecode(env, type1);
			code1.set_typecode(env, type2);

		END;
	END;

	RETURN result;

END SubEquiv;

TYPE	Assumptions = OBJECT
			next : Assumptions := NIL;
			code1, code2 : CARDINAL;
		END;

(* Routine to generate new type-codes. *)
PROCEDURE NewTypeCode() : CARDINAL=
BEGIN
	INC(TypeCode);
	RETURN TypeCode;
END NewTypeCode;

VAR	TypeCode : CARDINAL := 0;
BEGIN
END TypeSys.