(* Variables module.
	This module implements methods for variable related
	commands and stuff.  Also provides a loader for
	these object types.
*)

(*
	11/12/96	Skeleton created.
	12/12/96	Added methods for the Variable type.
			Added various types used in variable lists etc.
	03/01/96	Added the LetAll object.
	10/01/96	Added the get_typecode method to the type-list
			objects.
	14/01/97	MAJOR RE-WRITE.  SEE NODE MODULE.
*)

MODULE Variables;

(* ** DEBUGGING. ** *)
(* IMPORT IO; *)

IMPORT Text;

IMPORT Code, Tags;

IMPORT Loader, ArmRT;
IMPORT Node;

(* The basic variable type. *)
REVEAL	Variable = Node.T BRANDED OBJECT
			name : TEXT;
		OVERRIDES

			get_value := Var_GetValue;

			print := Var_Print;

			get_name := Var_GetName;

		END;

PROCEDURE Var_GetValue(self : Variable; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN env.get_var(self.name).get_value(env);
END Var_GetValue;

PROCEDURE Var_Print(self : Variable; i : TEXT) : TEXT=
BEGIN
	RETURN self.name;
END Var_Print;

PROCEDURE Var_GetName(self : Variable; env : Node.E) : TEXT=
BEGIN
	RETURN self.name;
END Var_GetName;

(* The variable declaration block. *)
REVEAL	VarBlk = Node.T BRANDED OBJECT
			varnames : Node.TYP;
			command : Node.T;
			is_block := FALSE;
		OVERRIDES
			execute := VarBlk_Execute;
			print := VarBlk_Print;
		END;

(* This routine adds a new stack frame containing some variables. *)
PROCEDURE VarBlk_Execute(self : VarBlk; env : Node.E) RAISES {Node.Abort}=
VAR	new_env : Node.E;
BEGIN

	IF self.is_block THEN
		(* Make a new environment to add to. *)
		new_env := env.make_child();
	ELSE
		(* Add to the current environment. *)
		new_env := env;
	END;

	(* Check that all types are resolved. *)
	self.varnames := self.varnames.resolve(env);

	(* Add the variables to the specified environment frame. *)
	new_env := new_env.add_variables(self.varnames.get_varlist(new_env));

	IF self.is_block THEN
		(* Execute the specified commands in this frame. *)
		ArmRT.Execute(self.command, new_env);
	END;

END VarBlk_Execute;

PROCEDURE VarBlk_Print(self : VarBlk; i : TEXT) : TEXT=
VAR	text : TEXT;
BEGIN
	text := "var " & self.varnames.print(i & "\t");
	IF self.is_block THEN
		text := text & " in\n" & i & "\t" &
			Node.ToText(self.command, i & "\t");
	END;
	RETURN text;
END VarBlk_Print;

(* Variable assignment - the "let" block. *)
REVEAL	Let = Node.T BRANDED OBJECT
			varnames : Node.T;
			expressions : Node.T;
		OVERRIDES
			execute := Let_Execute;
			print := Let_Print;
		END;

PROCEDURE Let_Execute(self : Let; env : Node.E) RAISES {Node.Abort}=
VAR	var := self.varnames;
	expr := self.expressions;
BEGIN

	(* Work through the lists of variables and expressions, assigning. *)
	WHILE var # NIL DO

		(* The assignment can be done in a single, complex-looking command! *)
		var.get_value(env).set_value(env, expr.get_expr(env).get_value(env));

		(* Now move on to the next of each. *)
		var := var.next;
		expr := expr.next;

	END;

END Let_Execute;

PROCEDURE Let_Print(self : Let; i : TEXT) : TEXT=
BEGIN
	RETURN "let " & self.varnames.print(i) & " = " & self.expressions.print(i);
END Let_Print;

(* Multiple variable assignment - the "letall" block. *)
REVEAL	LetAll = Node.T BRANDED OBJECT
			varnames : Node.T;
			expression : Node.T;
		OVERRIDES
			execute := LetAll_Execute;
			print := LetAll_Print;
		END;

PROCEDURE LetAll_Execute(self : LetAll; env : Node.E) RAISES {Node.Abort}=
VAR	var := self.varnames;
	value : Node.V;
BEGIN

	(* Evaluate the expression. *)
	value := self.expression.get_expr(env).get_value(env);

	(* Work through the lists of variables, assigning. *)
	WHILE var # NIL DO

		(* The assignment can be done in a single command! *)
		var.get_value(env).set_value(env, value);

		(* Now move on to the next variable. *)
		var := var.next;

	END;

END LetAll_Execute;

PROCEDURE LetAll_Print(self : LetAll; i : TEXT) : TEXT=
BEGIN
	RETURN "let " & self.varnames.print(i) & " = " & self.expression.print(i);
END LetAll_Print;

(* Variable declaration list. *)
REVEAL VarDeclList = Node.TYP BRANDED OBJECT
				vardecl : Node.TYP;
				resolved := FALSE;
			OVERRIDES
				get_varlist := VarDeclList_GetVarList;
				get_field_no := VarDeclList_GetFieldNo;
				get_field_type := VarDeclList_GetFieldType;
				check_type := VarDeclList_CheckType;
				resolve := VarDeclList_Resolve;
				print := VarDeclList_Print;
			END;

PROCEDURE VarDeclList_GetVarList(self : VarDeclList;
				env : Node.E;
				list : Node.VarList) : Node.VarList RAISES {Node.Abort}=
BEGIN

	IF self.next # NIL THEN
		list := self.next.get_varlist(env, list);
	END;

	list := self.vardecl.get_varlist(env, list);

	RETURN list;

END VarDeclList_GetVarList;

PROCEDURE VarDeclList_GetFieldNo(self : VarDeclList;
				env : Node.E;
				name : TEXT) : INTEGER RAISES {Node.Abort}=
VAR	pos := 0;
	found_at : INTEGER;
BEGIN

	(* Check the first variable declaration. *)
	found_at := VarDecl_GetFieldNo(self.vardecl, name, pos); 

	IF found_at < 0 THEN
		IF self.next = NIL THEN
			RETURN -1;
		ELSE
			(* Check the next vardecl. *)
			RETURN pos + NARROW(self.next, VarDeclList).get_field_no(env, name);
		END;
	ELSE
		RETURN found_at;
	END;

END VarDeclList_GetFieldNo;

PROCEDURE VarDeclList_GetFieldType(self : VarDeclList;
				env : Node.E;
				n : CARDINAL) : Node.TYP RAISES {Node.Abort}=
VAR	pos := n;
	type : Node.T;
BEGIN

	(* Check this vardecl. *)
	type := VarDecl_GetFieldType(self.vardecl, pos);

	IF type = NIL THEN
		IF self.next = NIL THEN
			RETURN NIL;
		ELSE
			(* Try the next one instead. *)
			RETURN NARROW(self.next, VarDeclList).get_field_type(env, pos);
		END;
	ELSE
		RETURN type;
	END;

END VarDeclList_GetFieldType;

PROCEDURE VarDeclList_CheckType( s : VarDeclList; env : Node.E; i : Node.InvalidTypes) : BOOLEAN RAISES {Node.Abort}=
VAR	pos : VarDeclList;
BEGIN
	(* Run through the variable declaration list, checking the types. *)
	pos := s;
	WHILE pos # NIL DO
		IF NOT(pos.vardecl.check_type(env, i)) THEN
			RETURN FALSE;
		END;
		pos := pos.next;
	END;

	RETURN TRUE;
END VarDeclList_CheckType;

PROCEDURE VarDeclList_Resolve(self : VarDeclList; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	IF NOT(self.resolved) THEN
		self.resolved := TRUE;
		self.vardecl := self.vardecl.resolve(env);
		IF self.next # NIL THEN
			self.next := NARROW(self.next, VarDeclList).resolve(env);
		END;
	END;
	RETURN self;
END VarDeclList_Resolve;

PROCEDURE VarDeclList_Print(self : VarDeclList; i : TEXT) : TEXT=
VAR	text : TEXT;
	pos := self;
BEGIN

	text := NIL;
	WHILE pos # NIL DO

		IF text = NIL THEN
			text := "";
		ELSE
			text := text & ", ";
		END;

		text := text & pos.vardecl.print(i);
		pos := pos.next;

	END;
	IF text = NIL THEN RETURN ""; END;

	RETURN text;

END VarDeclList_Print;

REVEAL	VarDecl = Node.TYP BRANDED OBJECT
			varlist : Node.T;
			type : Node.TYP;
			resolved := FALSE;
		OVERRIDES
			get_varlist := VarDecl_GetVarList;
			check_type := VarDecl_CheckType;
			resolve := VarDecl_Resolve;
			print := VarDecl_Print;
		END;

PROCEDURE VarDecl_GetVarList(self : VarDecl;
			env : Node.E;
			list : Node.VarList) : Node.VarList RAISES {Node.Abort}=
BEGIN

	(* Create a varlist of the specified names, with appropriate type nodes. *)
	RETURN MakeVarList(self.varlist, env, list, self.type);

END VarDecl_GetVarList;

(* Sub-routine to get the variable list in the right order. *)
PROCEDURE MakeVarList(varname : Node.T; env : Node.E; list : Node.VarList; type : Node.TYP) : Node.VarList RAISES {Node.Abort}=
BEGIN

	IF varname = NIL THEN RETURN list; END;

	RETURN NEW(Node.VarList).init(varname.get_name(env),
					type.allocate_v(env)
				).setnext(MakeVarList(varname.next, env, list, type));

END MakeVarList;

PROCEDURE VarDecl_CheckType(s : VarDecl; env : Node.E; i : Node.InvalidTypes) : BOOLEAN RAISES {Node.Abort}=
BEGIN
	(* Check the type of the declarator. *)
	RETURN s.type.check_type(env, i);
END VarDecl_CheckType;

PROCEDURE VarDecl_Resolve(s : VarDecl; env : Node.E) : Node.TYP RAISES {Node.Abort}=
BEGIN
	IF NOT(s.resolved) THEN
		s.resolved := TRUE;
		s.type := s.type.resolve(env);
	END;
	RETURN s;
END VarDecl_Resolve;

PROCEDURE VarDecl_Print(self : VarDecl; i : TEXT) : TEXT=
BEGIN
	RETURN self.varlist.print(i) & " : " & self.type.print(i);
END VarDecl_Print;

(* Internal procedure for locating a field. *)
PROCEDURE VarDecl_GetFieldNo(self : VarDecl; name : TEXT; VAR pos : INTEGER) : INTEGER RAISES {Node.Abort}=
VAR	listpos := self.varlist;
BEGIN

	(* Search the list for the required element. *)
	WHILE listpos # NIL DO
		IF Text.Equal(listpos.get_name(NIL), name) THEN
			RETURN pos;
		END;
		listpos := listpos.next;
		INC(pos);
	END;

	(* Failed - return an invalid value! *)
	RETURN -1;

END VarDecl_GetFieldNo;

(* Internal procedure for getting the type of a field. *)
PROCEDURE VarDecl_GetFieldType(self : VarDecl; VAR pos : CARDINAL) : Node.T=
VAR	listpos := self.varlist;
BEGIN

	(* Search the list for the required element. *)
	WHILE listpos # NIL DO
		IF pos = 0 THEN
			RETURN self.type;
		END;
		listpos := listpos.next;
		DEC(pos);
	END;

	(* Failed - return an invalid value! *)
	RETURN NIL;

END VarDecl_GetFieldType;

(* List of variable names. *)
REVEAL	VarNameList = Node.T BRANDED OBJECT
			name : Node.T;
		OVERRIDES
			get_name := VarNameList_GetName;
			get_value := VarNameList_GetValue;
			print := VarNameList_Print;
		END;

PROCEDURE VarNameList_GetName(self : VarNameList; env : Node.E) : TEXT RAISES {Node.Abort}=
BEGIN
	RETURN self.name.get_name(env);
END VarNameList_GetName;

PROCEDURE VarNameList_GetValue(self : VarNameList; env : Node.E) : Node.V RAISES {Node.Abort}=
BEGIN
	RETURN self.name.get_value(env);
END VarNameList_GetValue;

PROCEDURE VarNameList_Print(self : VarNameList; i : TEXT) : TEXT=
VAR	pos : VarNameList;
	text : TEXT := NIL;
BEGIN

	pos := self;
	WHILE pos # NIL DO

		IF text = NIL THEN
			text := "";
		ELSE
			text := text & ", ";
		END;

		text := text & pos.name.print(i);
		pos := pos.next;

	END;
	IF text = NIL THEN RETURN ""; END; 

	RETURN text;

END VarNameList_Print;

(* Routine to produce the above objects. *)
PROCEDURE Convert(tree : Code.T) : Node.T RAISES {Loader.Failure}=
VAR	token : Node.T;
BEGIN

	IF tree = NIL THEN RETURN NIL; END;

	CASE tree.type OF
		Tags.T.TAG_Variable=>
			RETURN NEW(Variable, name := tree.extra);
	|	Tags.T.BLK_Var=>
			token := NEW(VarBlk,
					varnames := Loader.Convert(tree.params.param)
				);
			IF tree.params.next # NIL THEN
				NARROW(token, VarBlk).command := Loader.ConvertCommand(tree.params.next.param);
				NARROW(token, VarBlk).is_block := TRUE;
			END;
			RETURN token;
	|	Tags.T.CMD_Let=>
			RETURN NEW(Let,
					varnames := Loader.Convert(tree.params.param),
					expressions := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.CMD_LetAll=>
			RETURN NEW(LetAll,
					varnames := Loader.Convert(tree.params.param),
					expression := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.TAG_VarDeclList=>
			token := NEW(VarDeclList, vardecl := Loader.Convert(tree.params.param));
			IF tree.params.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	|	Tags.T.TAG_VarDecl=>
			RETURN NEW(VarDecl,
					varlist := Loader.Convert(tree.params.param),
					type := Loader.Convert(tree.params.next.param)
				);
	|	Tags.T.TAG_VarList=>
			token := NEW(VarNameList, name := Loader.Convert(tree.params.param));
			IF tree.params.next # NIL THEN
				token.next := Loader.Convert(tree.params.next.param);
			END;
			RETURN token;
	ELSE
		RETURN NIL;
	END;

END Convert;

BEGIN
END Variables.